/*
 * QLogic iSCSI HBA Driver
 * Copyright (c)   2003-2012 QLogic Corporation
 *
 * See LICENSE.qla4xxx for copyright and licensing details.
 */

#include "ql4_def.h"
#include "ql4_version.h"
#include "ql4_glbl.h"
#include "ql4_dbg.h"

/* Flash Read/Write functions */

uint32_t qla4_83xx_rd_reg(struct scsi_qla_host *ha, ulong addr)
{
	return readl((void __iomem *) (ha->nx_pcibase + addr));
}

void qla4_83xx_wr_reg(struct scsi_qla_host *ha, ulong addr, uint32_t val)
{
	writel(val, (void __iomem *)((ha)->nx_pcibase + addr));
}

static int qla4_83xx_set_win_base(struct scsi_qla_host *ha, uint32_t addr)
{
	uint32_t val;
	int ret_val = QLA_SUCCESS;

	qla4_83xx_wr_reg(ha, QLA83XX_CRB_WIN_FUNC(ha->func_num), addr);
	val = qla4_83xx_rd_reg(ha, QLA83XX_CRB_WIN_FUNC(ha->func_num));
	if (val != addr) {
		ql4_err(ha, "%s: Failed to set register window : "
			"addr written 0x%x, read 0x%x!\n",
			__func__, addr, val);
		ret_val = QLA_ERROR;
	}

	return ret_val;
}

int qla4_83xx_rd_reg_indirect(struct scsi_qla_host *ha, uint32_t addr,
				   uint32_t *data)
{
	int ret_val = QLA_SUCCESS;

	ret_val = qla4_83xx_set_win_base(ha, addr);
	if (!ret_val)
		*data = qla4_83xx_rd_reg(ha, QLA83XX_WILDCARD);
	else 
		ql4_err(ha, "%s: failed read of addr 0x%x!\n", __func__, addr);

	return ret_val;
}

int qla4_83xx_wr_reg_indirect(struct scsi_qla_host *ha, uint32_t addr, 
    			     uint32_t data)
{
	int ret_val = QLA_SUCCESS;
	
	ret_val = qla4_83xx_set_win_base(ha, addr);
	if (!ret_val)
		qla4_83xx_wr_reg(ha, QLA83XX_WILDCARD, data);
	else
		ql4_err(ha, "%s: failed wrt to addr 0x%x, data 0x%x\n", 
			__func__, addr, data);

	return ret_val;
}

void qla4_83xx_read_crb(struct scsi_qla_host *ha, char *buf, 
		        uint32_t offset, size_t size)
{
	uint32_t data;
	int status;

	status = qla4_83xx_rd_reg_indirect(ha, offset, &data);
	if (status == QLA_SUCCESS)
		memcpy(buf, &data, size);
	else
		ql4_err(ha, "%s: failed read of offset 0x%x!\n", 
			__func__, offset);
}

int qla4_83xx_write_crb(struct scsi_qla_host *ha, char *buf, 
		         uint32_t offset, size_t size)
{
	uint32_t data;
	int status;

	memcpy(&data, buf, size);
	status = qla4_83xx_wr_reg_indirect(ha, offset, data);
	if (status == QLA_ERROR) {
		ql4_err(ha, "%s: failed wrt to offset 0x%x!\n", 
			__func__, offset);
	}

	return status;
}

int qla4_83xx_flash_lock(struct scsi_qla_host *ha)
{
	int lock_owner;
	int timeout = 0;
	uint32_t lock_status = 0;
	int ret_val = QLA_SUCCESS;

	while (lock_status == 0) {
		lock_status = qla4_83xx_rd_reg(ha, QLA83XX_FLASH_LOCK);
		if (lock_status)
			break;

		if (++timeout >= QLA83XX_FLASH_LOCK_TIMEOUT / 20) {
			lock_owner = qla4_83xx_rd_reg(ha, QLA83XX_FLASH_LOCK_ID);
			ql4_err(ha, "%s: flash lock by %d failed, held by %d\n",
				__func__, ha->func_num, lock_owner);
			ret_val = QLA_ERROR;
			break;
		}
		msleep(20);
	}

	qla4_83xx_wr_reg(ha, QLA83XX_FLASH_LOCK_ID, ha->func_num);

	return ret_val;
}

void qla4_83xx_flash_unlock(struct scsi_qla_host *ha)
{
	int ret_val;

	/* Reading FLASH_UNLOCK register unlocks the Flash */
	qla4_83xx_wr_reg(ha, QLA83XX_FLASH_LOCK_ID, 0xFF);
	ret_val = qla4_83xx_rd_reg(ha, QLA83XX_FLASH_UNLOCK);
}

int qla4_83xx_flash_read_u32(struct scsi_qla_host *ha, uint32_t flash_addr,
			     uint8_t *p_data, int u32_word_count)
{
	int i;
	uint32_t u32_word;
	uint32_t addr = flash_addr;
	int ret_val = QLA_SUCCESS;

	if (qla4_83xx_flash_lock(ha) != 0) {
		ret_val = QLA_ERROR;
		goto exit_lock_error;
	}

	if (addr & 0x03) {
		ql4_err(ha, "%s: Illegal addr = 0x%x\n", __func__, addr);
		ret_val = QLA_ERROR;
		goto exit_flash_read;
	}

	for (i = 0; i < u32_word_count; i++) {
		if (qla4_83xx_wr_reg_indirect(ha, 
				QLA83XX_FLASH_DIRECT_WINDOW, 
				(addr & 0xFFFF0000))) {
			ql4_err(ha, "%s: failed to write addr 0x%x to "
				"FLASH_DIRECT_WINDOW\n! ", __func__, addr);
			ret_val = QLA_ERROR;
			goto exit_flash_read;
		}
	
		ret_val = qla4_83xx_rd_reg_indirect(ha,
				QLA83XX_FLASH_DIRECT_DATA(addr),
				&u32_word);
		if (ret_val != QLA_SUCCESS) {
			ql4_err(ha, "%s: failed to read addr 0x%x!\n", 
				__func__, addr);
			goto exit_flash_read;
		}	

		*(__le32 *)p_data = le32_to_cpu(u32_word);
		p_data = p_data + 4;
		addr = addr + 4;
	}

exit_flash_read:
	qla4_83xx_flash_unlock(ha);

exit_lock_error:
	return ret_val;
}

int qla4_83xx_lockless_flash_read_u32(struct scsi_qla_host *ha, 
				      uint32_t flash_addr, uint8_t *p_data, 
				      int u32_word_count)
{
	uint32_t i;
	uint32_t u32_word; 
	uint32_t flash_offset;
        uint32_t addr = flash_addr;
	int ret_val = QLA_SUCCESS;

        flash_offset = addr & (QLA83XX_FLASH_SECTOR_SIZE - 1);

        if (addr & 0x3) {
                ql4_err(ha, "%s: Illegal addr = 0x%x\n", 
			__func__, addr);
		ret_val = QLA_ERROR;
		goto exit_lockless_read;
        }

        ret_val = qla4_83xx_wr_reg_indirect(ha,
                        QLA83XX_FLASH_DIRECT_WINDOW, (addr));
	if (ret_val != QLA_SUCCESS) {
		ql4_err(ha, "%s: failed to write addr 0x%x to "
			"FLASH_DIRECT_WINDOW!\n", __func__, addr);
		goto exit_lockless_read;
	}	

        /* Check if data is spread across multiple sectors  */
        if ((flash_offset + (u32_word_count * sizeof(uint32_t))) >
                                (QLA83XX_FLASH_SECTOR_SIZE - 1)) {

                /* Multi sector read */
                for (i = 0; i < u32_word_count; i++) {
                        ret_val = qla4_83xx_rd_reg_indirect(ha,
                                        QLA83XX_FLASH_DIRECT_DATA(addr),
					&u32_word);
			if (ret_val != QLA_SUCCESS) {
				ql4_err(ha, "%s: failed to read addr 0x%x!\n", 
					__func__, addr);
				goto exit_lockless_read;
			}	

                        *(__le32 *)p_data  = le32_to_cpu(u32_word);
                        p_data = p_data + 4;
                        addr = addr + 4;
                        flash_offset = flash_offset + 4;

                        if (flash_offset > (QLA83XX_FLASH_SECTOR_SIZE - 1)) {
                                /* This write is needed once for each sector */
                                ret_val = qla4_83xx_wr_reg_indirect(ha,
                                        QLA83XX_FLASH_DIRECT_WINDOW, (addr));
				if (ret_val != QLA_SUCCESS) {
					ql4_err(ha, "%s: failed to write addr "
						"0x%x to FLASH_DIRECT_WINDOW!\n",
						__func__, addr);
					goto exit_lockless_read;
				}	
                                flash_offset = 0;
                        }
                }
        } else {
                /* Single sector read */
                for (i = 0; i < u32_word_count; i++) {
                        ret_val = qla4_83xx_rd_reg_indirect(ha,
                                        QLA83XX_FLASH_DIRECT_DATA(addr),
					&u32_word);
			if (ret_val != QLA_SUCCESS) {
				ql4_err(ha, "%s: failed to read addr 0x%x!\n", 
					__func__, addr);
				goto exit_lockless_read;
			}	

                        *(__le32 *)p_data = le32_to_cpu(u32_word);
                        p_data = p_data + 4;
                        addr = addr + 4;
                }
        }

exit_lockless_read:
        return ret_val;
}

static int qla4_83xx_poll_flash_status_reg(struct scsi_qla_host *ha)
{
	uint32_t flash_status;
	int retries = QLA83XX_FLASH_READ_RETRY_COUNT;
	int ret_val = QLA_SUCCESS;

	do {
		ret_val = qla4_83xx_rd_reg_indirect(ha, QLA83XX_FLASH_STATUS,
						    &flash_status);
		if (ret_val == QLA_ERROR) {
			ql4_err(ha, "%s: failed to read FLASH_STATUS reg!\n", 
				__func__);
			break;	
		}

		if ((flash_status & QLA83XX_FLASH_STATUS_READY) ==
				QLA83XX_FLASH_STATUS_READY)	
			break;

		msleep(QLA83XX_FLASH_STATUS_REG_POLL_DELAY);
	} while (--retries);

	if (!retries)
		ret_val = QLA_ERROR;

	return QLA_SUCCESS;
}

#ifdef FLASH_UPDATE
static int qla4_83xx_read_flash_status_reg(struct scsi_qla_host *ha)
{
	int flash_status;
	int ret_val = QLA_SUCCESS;

	ret_val = qla4_83xx_wr_reg_indirect(ha, QLA83XX_FLASH_ADDR, 0xFD0005);
	ret_val += qla4_83xx_wr_reg_indirect(ha, QLA83XX_FLASH_CONTROL, 0x3F);
	if (ret_val != QLA_SUCCESS) {
		ql4_err(ha, "%s: failed to write to FLASH_ADDR/CONTROL!\n", __func__);
		goto exit_read_status;
	}

	ret_val = qla4_83xx_poll_flash_status_reg(ha);
	if (ret_val) {
		ql4_err(ha, "%s: failed!\n", __func__);
		goto exit_read_status;
	}

	ret_val = qla4_83xx_rd_reg_indirect(ha, QLA83XX_FLASH_RDDATA, 
					    &flash_status);
	if (ret_val == QLA_ERROR) {
		ql4_err(ha, "%s: failed to read FLASH_STATUS reg!\n", 
			__func__);
		goto exit_read_status;
	}
	ret_val = (flash_status & 0xFF);

exit_read_status:
	return ret_val;
}

static int qla4_83xx_write_flash_status_reg(struct scsi_qla_host *ha, 
					    uint32_t data)
{
	int ret_val = QLA_SUCCESS;

	ret_val = qla4_83xx_wr_reg_indirect(ha, QLA83XX_FLASH_ADDR, 0xFD0101);
	ret_val += qla4_83xx_wr_reg_indirect(ha, QLA83XX_FLASH_WRDATA, data);
	ret_val += qla4_83xx_wr_reg_indirect(ha, QLA83XX_FLASH_CONTROL, 0x5);
	if (ret_val != QLA_SUCCESS) {
		ql4_err(ha, "%s: failed write to FLASH_ADDR/WRDATA/CONTROL!\n", 
			__func__);
		goto exit_wrt_flash_status;
	}

	ret_val = qla4_83xx_poll_flash_status_reg(ha);
	if (ret_val)
		ql4_err(ha, "%s: Error polling flash status reg!\n",
			__func__);

exit_wrt_flash_status:
	return ret_val;
}

static int qla4_83xx_save_flash_status(struct scsi_qla_host *ha)
{
	int ret_val = QLA_SUCCESS;

	ret_val = qla4_83xx_read_flash_status_reg(ha);
	if (ret_val == QLA_ERROR) {
		ql4_err(ha, "%s: read_flash_status failed!\n", __func__);
		goto exit_save_status;
	}

	ret_val = qla4_83xx_write_flash_status_reg(ha, 0);
	if (ret_val == QLA_ERROR) {
		ql4_err(ha, "%s: write_flash_status failed!\n", __func__);
	}

exit_save_status:
	return ret_val;
}

static int qla4_83xx_restore_flash_status(struct scsi_qla_host *ha, int status)
{
	int ret_val = QLA_SUCCESS;

	ret_val = qla4_83xx_write_flash_status_reg(ha, status);
	if (ret_val == QLA_ERROR) {
		ql4_err(ha, "%s: write_flash_status failed!\n", __func__);
	}

	return ret_val;
}

static int qla4_83xx_erase_flash_sector(struct scsi_qla_host *ha, 
					uint32_t sector_start_addr)
{
	uint32_t status;
	uint32_t reversed_addr;
	int ret_val = QLA_SUCCESS;

	if (qla4_83xx_flash_lock(ha) != 0) {
		ret_val = QLA_ERROR;
		goto exit_lock_error;
	}

	/* save status register */
	ret_val = qla4_83xx_read_flash_status_reg(ha);
	if (ret_val == QLA_ERROR) {
		ql4_err(ha, "%s: Read Flash Status failed!\n", 
			 __func__);
		goto exit_erase_flash;
	}

	ret_val = qla4_83xx_write_flash_status_reg(ha, 0);
	if (ret_val == QLA_ERROR) {
		ql4_err(ha, "%s: Write Flash Status failed!\n", 
			 __func__);
		goto exit_erase_flash;
	}

	ret_val = qla4_83xx_wr_reg_indirect(ha, QLA83XX_FLASH_WRDATA, 0x0);
	ret_val += qla4_83xx_wr_reg_indirect(ha, QLA83XX_FLASH_ADDR, 0xFD0101);
	ret_val += qla4_83xx_wr_reg_indirect(ha, QLA83XX_FLASH_CONTROL, 0x5);
	if (ret_val != QLA_SUCCESS) {
		ql4_err(ha, "%s: failed write to FLASH_WRDATA/ADDR/CONTROL!\n", 
			__func__);
		goto exit_erase_flash;
	}

	ret_val = qla4_83xx_poll_flash_status_reg(ha);
	if (ret_val == QLA_ERROR) {
		ql4_err(ha, "%s: Poll Flash Status after erase failed!\n", 
			 __func__);
		goto exit_erase_flash;
	}

	reversed_addr = (((sector_start_addr & 0xFF) << 16) |
				((sector_start_addr & 0xFF0000) >> 16));

	ret_val = qla4_83xx_wr_reg_indirect(ha,
				QLA83XX_FLASH_WRDATA, reversed_addr);
	ret_val += qla4_83xx_wr_reg_indirect(ha, QLA83XX_FLASH_ADDR, 0xFD03D8);
	ret_val += qla4_83xx_wr_reg_indirect(ha, QLA83XX_FLASH_CONTROL, 0x3D);
	if (ret_val != QLA_SUCCESS) {
		ql4_err(ha, "%s: failed write to FLASH_WRDATA/ADDR/CONTROL!\n", 
			__func__);
		goto exit_erase_flash;
	}

	ret_val = qla4_83xx_poll_flash_status_reg(ha);
	if (ret_val == QLA_ERROR) {
		ql4_err(ha, "%s: Poll Flash Status failed!\n", 
			 __func__);
		goto exit_erase_flash;
	}

	/* restore status register */
	ret_val = qla4_83xx_write_flash_status_reg(ha, status);
	if (ret_val)
		ql4_err(ha, "%s: Write to restore Flash Status failed!\n", 
			 __func__);

exit_erase_flash:
	qla4_83xx_flash_unlock(ha);

exit_lock_error:
	return ret_val;
}
#endif

/* 
 * qla4_83xx_flash_write_u32 - Write data to flash
 *
 * @ha : Pointer to adapter structure
 * addr : Flash address to write to
 * p_data : Data to be written
 *
 * Return Value - QLA_SUCCESS/QLA_ERROR
 *
 * NOTE: Lock should be held on entry 
 */ 
int qla4_83xx_flash_write_u32(struct scsi_qla_host *ha, uint32_t addr,
			      uint32_t *p_data)
{
	int ret_val = QLA_SUCCESS;
	
	ret_val = qla4_83xx_wr_reg_indirect(ha, QLA83XX_FLASH_ADDR,
				 0x00800000 | (addr >> 2));
	ret_val += qla4_83xx_wr_reg_indirect(ha, QLA83XX_FLASH_WRDATA, 
					      *p_data);
	ret_val += qla4_83xx_wr_reg_indirect(ha, QLA83XX_FLASH_CONTROL, 0x3D);
	if (ret_val != QLA_SUCCESS) {
		ql4_err(ha, "%s: failed write to FLASH_ADDR/WRDATA/CONTROL!\n", 
			__func__);
		goto exit_flash_wrt32;
	}

	ret_val = qla4_83xx_poll_flash_status_reg(ha);
	if (ret_val == QLA_ERROR)
		ql4_err(ha, "%s: poll_flash_status failed\n", __func__);

exit_flash_wrt32:
	return ret_val;
}

/* 
 * qla4_83xx_flash_buffer_write - Write data to in buffer mode
 * min size = 2 dwords, max size = 64 dwords
 *
 * @ha : Pointer to adapter structure
 * addr : Flash address to write to
 * p_data : Data to be written
 *
 * Return Value - QLA_SUCCESS/QLA_ERROR
 *
 * NOTE: Lock should be held on entry 
 */ 
int qla4_83xx_flash_buffer_write(struct scsi_qla_host *ha, uint32_t addr,
			         uint32_t *p_data, int u32_word_count)
{
	uint32_t spi_ctrl;
	int ret_val = QLA_SUCCESS;
	int spi_status;

	if ((u32_word_count <  QLA83XX_FLASH_BUFFER_WRITE_MIN) ||
			(u32_word_count >  QLA83XX_FLASH_BUFFER_WRITE_MAX)) {
		ql4_err(ha, "%s: Invalid word count!\n", __func__);
		ret_val = QLA_ERROR;
		goto exit_flash_buffer_write;
	}

	ret_val = qla4_83xx_rd_reg_indirect(ha, QLA83XX_FLASH_SPI_CONTROL,
					    &spi_ctrl);
	if (ret_val == QLA_ERROR) {
		ql4_err(ha, "%s: FLASH_SPI_CONTROL read failed!\n", __func__);
		goto exit_flash_buffer_write;
	}

	ret_val = qla4_83xx_wr_reg_indirect(ha,
				 QLA83XX_FLASH_SPI_CONTROL, (spi_ctrl | 0x4));
	ret_val += qla4_83xx_wr_reg_indirect(ha, QLA83XX_FLASH_ADDR, 
					      0x00800000);

	/* First DWORD write */
	ret_val += qla4_83xx_wr_reg_indirect(ha, QLA83XX_FLASH_WRDATA, 
					      *p_data++);
	ret_val += qla4_83xx_wr_reg_indirect(ha, QLA83XX_FLASH_CONTROL, 0x43);
	if (ret_val != QLA_SUCCESS) {
		ql4_err(ha, "%s: failed write to FLASH_ADDR/WRDATA/CONTROL!\n", 
			__func__);
		goto exit_flash_buffer_write;
	}
	
	ret_val = qla4_83xx_poll_flash_status_reg(ha);
	if (ret_val == QLA_ERROR) {
		ql4_err(ha, "%s: poll_flash_status failed!\n", __func__);
		goto exit_flash_buffer_write;
	}

	u32_word_count--;

	ret_val = qla4_83xx_wr_reg_indirect(ha, QLA83XX_FLASH_ADDR, 
					     0x00800001);
	if (ret_val != QLA_SUCCESS) {
		ql4_err(ha, "%s: failed write to FLASH_ADDR!\n", __func__);
		goto exit_flash_buffer_write;
	}

	/* Second to N-1 DWORD writes */
	while (u32_word_count != 1) {
		ret_val = qla4_83xx_wr_reg_indirect(ha,
					QLA83XX_FLASH_WRDATA, *p_data++);
		ret_val += qla4_83xx_wr_reg_indirect(ha,
					QLA83XX_FLASH_CONTROL, 0x7F);
		if (ret_val != QLA_SUCCESS) {
			ql4_err(ha, "%s: failed write to "
				"FLASH_WRDATA/CONTROL!\n", __func__);
			goto exit_flash_buffer_write;
		}

		ret_val = qla4_83xx_poll_flash_status_reg(ha);
		if (ret_val == QLA_ERROR) {
			ql4_err(ha, "%s: poll_flash_status failed!\n", 
				__func__);
			goto exit_flash_buffer_write;
		}
		u32_word_count--;
	}

	ret_val = qla4_83xx_wr_reg_indirect(ha, QLA83XX_FLASH_ADDR,
				 0x00800000 | (addr >> 2));
	/* Last DWORD write */
	ret_val += qla4_83xx_wr_reg_indirect(ha, QLA83XX_FLASH_WRDATA, 
					      *p_data++);
	ret_val += qla4_83xx_wr_reg_indirect(ha, QLA83XX_FLASH_CONTROL, 0x7D);
	if (ret_val != QLA_SUCCESS) {
		ql4_err(ha, "%s: failed write to FLASH_ADDR/WRDATA/CONTROL!\n", 
			__func__);
		goto exit_flash_buffer_write;
	}

	ret_val = qla4_83xx_poll_flash_status_reg(ha);
	if (ret_val == QLA_ERROR) {
		ql4_err(ha, "%s: poll_flash_status failed!\n", __func__);
		goto exit_flash_buffer_write;
	}

	ret_val = qla4_83xx_rd_reg_indirect(ha, QLA83XX_FLASH_SPI_STATUS,
					    &spi_status);
	if ((ret_val == QLA_ERROR) || (spi_status & 0x4)) {
		ql4_err(ha, "%s: Read of Flash SPI status failed!\n", 
			 __func__);
		/* Operation failed, clear error bit */
		ret_val = qla4_83xx_rd_reg_indirect(ha,
					QLA83XX_FLASH_SPI_CONTROL,
					&spi_ctrl);
		if (ret_val == QLA_ERROR) {
			ql4_err(ha, "%s: FLASH_SPI_CONTROL read failed!\n", 
				__func__);
			goto exit_flash_buffer_write;
		}

		ret_val = qla4_83xx_wr_reg_indirect(ha,
				QLA83XX_FLASH_SPI_CONTROL, (spi_ctrl | 0x4));
		if (ret_val == QLA_ERROR)
			ql4_err(ha, "%s: FLASH_SPI_CONTROL write failed!\n", 
				__func__);
		ret_val = QLA_ERROR;
	}

exit_flash_buffer_write:
	return ret_val;
}

void qla4_83xx_rom_lock_recovery(struct scsi_qla_host *ha)
{
	if (qla4_83xx_flash_lock(ha)) {
		ql4_info(ha, "%s: Resetting rom lock\n", __func__);
	}

	/* We got the lock, or someone else is holding the lock
 	 * since we are restting, forcefully unlock
 	 */
	qla4_83xx_flash_unlock(ha);
}

/*------------------Memory Write functions -------------------------------*/

/* 
 * qla4_83xx_ms_mem_write_128b - Writes data to MS/off-chip memory
 *
 * @ha : Pointer to adapter structure
 * addr : Flash address to write to
 * data : Data to be written
 * count : word_count to be written
 *
 * Return Value - QLA_SUCCESS/QLA_ERROR
 */ 
int qla4_83xx_ms_mem_write_128b(struct scsi_qla_host *ha,
			        uint64_t addr, uint32_t *data, uint32_t count)
{
	int i, j, ret_val = QLA_SUCCESS;
	uint32_t agt_ctrl;
	unsigned long flags;

	/* Only 128-bit aligned access */
	if (addr & 0xF) {
		ret_val = QLA_ERROR;
		goto exit_ms_mem_write;
	}

	write_lock_irqsave(&ha->hw_lock, flags);

	/* Write address */
	ret_val = qla4_83xx_wr_reg_indirect(ha, MD_MIU_TEST_AGT_ADDR_HI, 0);
	if (ret_val == QLA_ERROR) {
		ql4_err(ha, "%s: write to AGT_ADDR_HI failed!\n", __func__);
		goto exit_ms_mem_write_unlock;
	}

	for (i = 0 ; i < count ; i++, addr += 16) {
		if (!((QLA8XXX_ADDR_IN_RANGE(addr, QLA8XXX_ADDR_QDR_NET,
			QLA8XXX_ADDR_QDR_NET_MAX)) ||
		     (QLA8XXX_ADDR_IN_RANGE(addr, QLA8XXX_ADDR_DDR_NET,
			QLA8XXX_ADDR_DDR_NET_MAX)))) {
			ret_val = QLA_ERROR;
			goto exit_ms_mem_write_unlock;
		} 

		ret_val = qla4_83xx_wr_reg_indirect(ha,
				MD_MIU_TEST_AGT_ADDR_LO, addr);

		/* Write data */
		ret_val += qla4_83xx_wr_reg_indirect(ha,
				MD_MIU_TEST_AGT_WRDATA_LO, *data++);
		ret_val += qla4_83xx_wr_reg_indirect(ha,
				MD_MIU_TEST_AGT_WRDATA_HI, *data++);
		ret_val += qla4_83xx_wr_reg_indirect(ha,
				MD_MIU_TEST_AGT_WRDATA_ULO, *data++);
		ret_val += qla4_83xx_wr_reg_indirect(ha,
				MD_MIU_TEST_AGT_WRDATA_UHI, *data++);
		if (ret_val == QLA_ERROR) {
			ql4_err(ha, "%s: write to AGT_WRDATA failed!\n", 
				__func__);
			goto exit_ms_mem_write_unlock;
		}

		/* Check write status */
		ret_val = qla4_83xx_wr_reg_indirect(ha, MD_MIU_TEST_AGT_CTRL,
					 MIU_TA_CTL_WRITE_ENABLE);
		ret_val += qla4_83xx_wr_reg_indirect(ha, MD_MIU_TEST_AGT_CTRL,
					 MIU_TA_CTL_WRITE_START);
		if (ret_val == QLA_ERROR) {
			ql4_err(ha, "%s: write to AGT_CTRL failed!\n", __func__);
			goto exit_ms_mem_write_unlock;
		}

		for (j = 0; j < MAX_CTL_CHECK; j++) {
			ret_val = qla4_83xx_rd_reg_indirect(ha,
					     MD_MIU_TEST_AGT_CTRL,
					     &agt_ctrl);
			if (ret_val == QLA_ERROR) {
				ql4_err(ha, "%s: failed to read "
					"MD_MIU_TEST_AGT_CTRL!\n", __func__);
				goto exit_ms_mem_write_unlock;
			}
			if ((agt_ctrl & MIU_TA_CTL_BUSY) == 0)
				break;
		}

		/* Status check failed */
		if (j >= MAX_CTL_CHECK) {
			if (printk_ratelimit()) {
				ql4_err(ha, "%s: MS memory write failed!\n",
					__func__);
				ret_val = QLA_ERROR;
				goto exit_ms_mem_write_unlock;
			}
		}
	}

exit_ms_mem_write_unlock:
	write_unlock_irqrestore(&ha->hw_lock, flags);

exit_ms_mem_write:
	return ret_val;
}

/*-------------------- ISP83xx IDC Functions ------------------*/

#define INTENT_TO_RECOVER	0x01
#define PROCEED_TO_RECOVER	0x02

/*
 * qla4_83xx_lock_recovery - Recovers the idc_lock.
 *
 * @ha : Pointer to adapter structure
 *
 * Lock Recovery Register
 * 5-2	Lock recovery owner: Function ID of driver doing lock recovery,
 * 	valid if bits 1..0 are set by driver doing lock recovery.
 * 1-0  1 - Driver intends to force unlock the IDC lock.
 * 	2 - Driver is moving forward to unlock the IDC lock. Driver clears
 * 	    this field after force unlocking the IDC lock.
 *
 * Lock Recovery process
 * a. Read the IDC_LOCK_RECOVERY register. If the value in bits 1..0 is
 *    greater than 0, then wait for the other driver to unlock otherwise
 *    move to the next step.
 * b. Indicate intent to force-unlock by writing 1h to the IDC_LOCK_RECOVERY
 *    register bits 1..0 and also set the function# in bits 5..2.
 * c. Read the IDC_LOCK_RECOVERY register again after a delay of 200ms.
 *    Wait for the other driver to perform lock recovery if the function
 *    number in bits 5..2 has changed, otherwise move to the next step.
 * d. Write a value of 2h to the IDC_LOCK_RECOVERY register bits 1..0
 *    leaving your function# in bits 5..2.
 * e. Force unlock using the DRIVER_UNLOCK register and immediately clear
 *    the IDC_LOCK_RECOVERY bits 5..0 by writing 0.
 *
 * Return Value - QLA_SUCCESS/QLA_ERROR
 */
static int qla4_83xx_lock_recovery(struct scsi_qla_host *ha) {

	uint32_t lock = 0, lockid;

        lockid = ha->isp_ops->rd_reg_direct(ha, QLA83XX_DRV_LOCKRECOVERY);

	/* Check for other Recovery in progress, go wait */
	if((lockid & 0x3) != 0)
		return QLA_ERROR;

	/* Intent to Recover */
	ha->isp_ops->wr_reg_direct(ha,
		QLA83XX_DRV_LOCKRECOVERY,
		(ha->func_num << 2) | INTENT_TO_RECOVER);

	msleep(200);

	/* Check Intent to Recover is advertised */
        lockid = ha->isp_ops->rd_reg_direct(ha, QLA83XX_DRV_LOCKRECOVERY);
	if ((lockid & 0x3C) != (ha->func_num << 2))
		return QLA_ERROR;

	ql4_info(ha, "%s:%d: IDC Lock recovery initiated\n", __func__,
		 ha->func_num);

	/* Proceed to Recover */
	ha->isp_ops->wr_reg_direct(ha,
				   QLA83XX_DRV_LOCKRECOVERY,
				   (ha->func_num << 2) | PROCEED_TO_RECOVER);

	/* Force Unlock() */
	ha->isp_ops->wr_reg_direct(ha, QLA83XX_DRV_LOCK_ID, 0xFF);
	ha->isp_ops->rd_reg_direct(ha, QLA83XX_DRV_UNLOCK);

	/* Clear bits 0-5 in IDC_RECOVERY register*/
	ha->isp_ops->wr_reg_direct(ha, QLA83XX_DRV_LOCKRECOVERY, 0);

	/* Get lock() */
	lock = ha->isp_ops->rd_reg_direct(ha, QLA83XX_DRV_LOCK);
	if (lock) {
		lockid = ha->isp_ops->rd_reg_direct(ha, QLA83XX_DRV_LOCK_ID);
		lockid = ((lockid + (1 << 8)) & ~0xFF) | ha->func_num;
		ha->isp_ops->wr_reg_direct(ha, QLA83XX_DRV_LOCK_ID,
					   lockid);
		return QLA_SUCCESS;
	} else
		return QLA_ERROR;
}

#define QLA83XX_DRV_LOCK_MSLEEP 200

int qla4_83xx_drv_lock(struct scsi_qla_host *ha)
{
	int timeout = 0;
	uint32_t status = 0;
	int ret_val = QLA_SUCCESS;
	uint32_t first_owner = 0;
	uint32_t tmo_owner = 0;
	uint32_t lock_id;
	uint32_t func_num;
	uint32_t lock_cnt;

	while (status == 0) {
		status = qla4_83xx_rd_reg(ha, QLA83XX_DRV_LOCK);
		if (status) {
			/* Increment Counter (8-31) and update func_num (0-7) on
			 * getting a successful lock
			 */
			lock_id = qla4_83xx_rd_reg(ha, QLA83XX_DRV_LOCK_ID);
			lock_id = ((lock_id + (1 << 8)) & ~0xFF) | ha->func_num;
			qla4_83xx_wr_reg(ha, QLA83XX_DRV_LOCK_ID, lock_id);
			break;
		}

		if (timeout == 0) {
			/* Save counter + ID of function holding the lock for
			 * first failure
			 */
			first_owner = ha->isp_ops->rd_reg_direct(ha,
							QLA83XX_DRV_LOCK_ID);
		}

		if (++timeout >= (QLA83XX_DRV_LOCK_TIMEOUT/QLA83XX_DRV_LOCK_MSLEEP)) {
			tmo_owner = qla4_83xx_rd_reg(ha, QLA83XX_DRV_LOCK_ID);

			func_num = tmo_owner & 0xFF;
			lock_cnt = tmo_owner >> 8;
			ql4_info(ha, "%s: Lock by func %d failed after 2s, "
				"lock held by func %d, lock count %d, "
				"first_owner %d\n", __func__, ha->func_num,
				func_num, lock_cnt, (first_owner & 0xFF));

			if (first_owner != tmo_owner) {
				/* Some other driver got lock,
				 * OR same driver got lock again (counter
				 * value changed), when we were waiting for
				 * lock. Retry for another 2 sec */
				ql4_info(ha, "%s: %d: IDC lock failed\n",
					 __func__, ha->func_num);
				timeout = 0;
			} else {
				/* Same driver holding lock > 2sec.
				 * Force Recovery */
				if (qla4_83xx_lock_recovery(ha) ==
							QLA_SUCCESS) {
					/* Recovered and got lock */
					ret_val = QLA_SUCCESS;
					ql4_info(ha, "%s: "
						 "IDC lock Recovery by %d "
						 "successful...\n", __func__,
						 ha->func_num);
					break;
			       }

			       /* Recovery Failed, some other function
				* has the lock, wait for 2secs
				* and retry
				*/
			       ql4_info(ha, "%s: IDC lock Recovery by %d "
					"failed, Retrying timout\n", __func__,
					ha->func_num);
			       timeout = 0;
			}
		}
		msleep(QLA83XX_DRV_LOCK_MSLEEP);
	}

	return ret_val;
}

void qla4_83xx_drv_unlock(struct scsi_qla_host *ha)
{
	int id;

	id = qla4_83xx_rd_reg(ha, QLA83XX_DRV_LOCK_ID);
	if ((id & 0xFF) != ha->func_num) {
		ql4_err(ha, "%s: IDC Unlock by %d failed, lock owner is %d!\n",
			 __func__, ha->func_num, (id & 0xFF));
		return;
	}

	/* Keep lock counter value, update the ha->func_num to 0xFF */
	qla4_83xx_wr_reg(ha, QLA83XX_DRV_LOCK_ID, (id | 0xFF));
	qla4_83xx_rd_reg(ha, QLA83XX_DRV_UNLOCK);
}


void qla4_83xx_set_idc_dontreset(struct scsi_qla_host *ha)
{
	uint32_t idc_ctrl;

	idc_ctrl = qla4_83xx_rd_reg(ha, QLA83XX_IDC_DRV_CTRL);
	idc_ctrl |= DONTRESET_BIT0;
	DEBUG2(ql4_info(ha, "%s: idc_ctrl = %d\n", __func__,
			idc_ctrl));
	qla4_83xx_wr_reg(ha, QLA83XX_IDC_DRV_CTRL, idc_ctrl);
}

void qla4_83xx_clear_idc_dontreset(struct scsi_qla_host *ha)
{
	uint32_t idc_ctrl;

	idc_ctrl = qla4_83xx_rd_reg(ha, QLA83XX_IDC_DRV_CTRL);
	idc_ctrl &= ~DONTRESET_BIT0;
	DEBUG2(ql4_info(ha, "%s: idc_ctrl = %d\n", __func__,
			idc_ctrl));
	qla4_83xx_wr_reg(ha, QLA83XX_IDC_DRV_CTRL, idc_ctrl);
}

int qla4_83xx_idc_dontreset(struct scsi_qla_host *ha)
{
	uint32_t idc_ctrl;

	idc_ctrl = qla4_83xx_rd_reg(ha, QLA83XX_IDC_DRV_CTRL);
	return (idc_ctrl & DONTRESET_BIT0);
}

/*-------------------------IDC State Machine ---------------------*/

enum {
	UNKNOWN_CLASS = 0,
	NIC_CLASS,
	FCOE_CLASS,
	ISCSI_CLASS
};

struct device_info {
	int func_num;
	int device_type;
	int port_num;
};

int qla4_83xx_can_perform_reset(struct scsi_qla_host *ha)
{
	uint32_t drv_active;
	uint32_t dev_part, dev_part1, dev_part2;
	int i;
	struct device_info device_map[16];
	int func_nibble;
	int nibble;
	int nic_present = 0;
	int iscsi_present = 0;
	int iscsi_func_low = 0;

	/* Use the dev_partition register to determine the PCI function number 
 	 * and then check drv_active register to see which driver is loaded .
 	 */
	dev_part1 = qla4_83xx_rd_reg(ha, 
				     ha->reg_tbl[QLA8XXX_CRB_DEV_PART_INFO]);
	dev_part2 = qla4_83xx_rd_reg(ha, QLA83XX_CRB_DEV_PART_INFO2);
	drv_active = qla4_83xx_rd_reg(ha,
				      ha->reg_tbl[QLA8XXX_CRB_DRV_ACTIVE]);
	
	/* Each function has 4 bits in dev_partition Info register,
 	 * Lower 2 bits - device type, Upper 2 bits - physical port number
 	 */
	dev_part = dev_part1;
	for (i = nibble = 0; i <= 15; i++, nibble++) {
		func_nibble = dev_part & (0xF << (nibble * 4));
		func_nibble >>= (nibble * 4);
		device_map[i].func_num = i;
		device_map[i].device_type = func_nibble & 0x3; 
		device_map[i].port_num = func_nibble & 0xC;

		if (device_map[i].device_type == NIC_CLASS) {
			if (drv_active & (1 << device_map[i].func_num)) {
				nic_present++;	
				break;
			}
		} else if (device_map[i].device_type == ISCSI_CLASS) {
			if (drv_active & (1 << device_map[i].func_num)) {
				if (!iscsi_present ||
				    (iscsi_present && 
				     (iscsi_func_low > device_map[i].func_num)))
					iscsi_func_low = device_map[i].func_num;

				iscsi_present++;
			}
		}

		/* For function_num[8..15] get info from dev_part2 register */
		if (nibble == 7) {
			nibble = 0;
			dev_part = dev_part2;
		}
	}

 	/* NIC, iSCSI and FCOE are the Reset owners based on order, NIC gets 
 	 * predence over iSCSI and FCOE and iSCSI over FCOE, based on drivers present.
 	 */
	if (!nic_present &&
	    (ha->func_num == iscsi_func_low)) {
		DEBUG2(ql4_info(ha, "%s: can reset - NIC not present and "
			"is lower iSCSI function %d\n", __func__, ha->func_num));
		return 1;
	}

	return 0;
}

/**
 * qla4_83xx_need_reset_handler - Code to start reset sequence
 * @ha: pointer to adapter structure
 *
 * Note: IDC lock must be held upon entry
 **/
void qla4_83xx_need_reset_handler(struct scsi_qla_host *ha)
{
	uint32_t dev_state, drv_state, drv_active;
	unsigned long reset_timeout, dev_init_timeout;

	ql4_info(ha, "%s: Performing ISP error recovery\n", __func__);

	if (!test_bit(AF_8XXX_RST_OWNER, &ha->flags)) {
		DEBUG2(ql4_info(ha, "%s(%d): reset acknowledged\n",
					__func__, ha->host_no));
		qla4_8xxx_set_rst_ready(ha);

		/* Non-reset owners ACK Reset and wait for device INIT state
		 * as part of Reset Recovery by Reset Owner
		 */
		dev_init_timeout = jiffies + (ha->nx_dev_init_timeout * HZ);

		do {
			if (time_after_eq(jiffies, dev_init_timeout)) {
				ql4_info(ha, "%s: Non Reset owner DEV INIT "
						"TIMEOUT!\n", __func__);
				break;
			}

			ha->isp_ops->idc_unlock(ha);
			msleep(1000);
			ha->isp_ops->idc_lock(ha);

			dev_state = qla4_8xxx_rd_direct(ha,
					QLA8XXX_CRB_DEV_STATE);
		} while (dev_state == QLA8XXX_DEV_NEED_RESET);
	} else {
		qla4_8xxx_set_rst_ready(ha);

		/* wait for 10 seconds for reset ack from all functions */
		reset_timeout = jiffies + (ha->nx_reset_timeout * HZ);

		drv_state = qla4_8xxx_rd_direct(ha, QLA8XXX_CRB_DRV_STATE);
		drv_active = qla4_8xxx_rd_direct(ha, QLA8XXX_CRB_DRV_ACTIVE);

		ql4_info(ha, "%s(%d): drv_state = 0x%x, drv_active = 0x%x\n",
				__func__, ha->host_no, drv_state, drv_active);

		while (drv_state != drv_active) {
			if (time_after_eq(jiffies, reset_timeout)) {
				ql4_info(ha, "%s: RESET TIMEOUT!"
						"drv_state: 0x%08x, "
						"drv_active: 0x%08x\n",
						DRIVER_NAME, drv_state,
						drv_active);
				break;
			}

			ha->isp_ops->idc_unlock(ha);
			msleep(1000);
			ha->isp_ops->idc_lock(ha);

			drv_state = qla4_8xxx_rd_direct(ha,
							QLA8XXX_CRB_DRV_STATE);
			drv_active = qla4_8xxx_rd_direct(ha,
							QLA8XXX_CRB_DRV_ACTIVE);
		}

		if (drv_state != drv_active) {
			ql4_info(ha, "%s(%d): Reset_owner turning off "
					"drv_active of non-acking function "
					"0x%x\n", __func__, ha->host_no,
					(drv_active ^ drv_state));
			drv_active = drv_active & drv_state;
			qla4_8xxx_wr_direct(ha, QLA8XXX_CRB_DRV_ACTIVE,
					    drv_active);
		}

		/* Clear RESET OWNER, will be set at next reset by next
		 * RST_OWNER */
		clear_bit(AF_8XXX_RST_OWNER, &ha->flags);

		/* Start Reset Recovery */
		qla4_8xxx_device_bootstrap(ha);
	}
}

/*
 * qla4_83xx_isp_reset - Resets ISP and aborts all outstanding commands.
 * @ha: pointer to host adapter structure.
 **/
int
qla4_83xx_isp_reset(struct scsi_qla_host *ha)
{
	int rval;
	uint32_t dev_state;

	ha->isp_ops->idc_lock(ha);
	dev_state = qla4_8xxx_rd_direct(ha, QLA8XXX_CRB_DEV_STATE);

	if (ql4xdontresethba)
		qla4_83xx_set_idc_dontreset(ha);

	if (dev_state == QLA8XXX_DEV_READY) {
		/* If IDC_CTRL DONTRESETHBA_BIT0 is set dont do reset
		 * recovery
		 */
		if (qla4_83xx_idc_dontreset(ha) == DONTRESET_BIT0) {
			ql4_err(ha, "%s: RESET RECOVERY DISABLED!\n",
				__func__);
			rval = QLA_ERROR;
			ha->isp_ops->idc_unlock(ha);
			goto exit_isp_reset;
		}

		DEBUG2(ql4_info(ha, "HW State: NEED RESET\n"));
		qla4_8xxx_wr_direct(ha, QLA8XXX_CRB_DEV_STATE,
				    QLA8XXX_DEV_NEED_RESET);

	} else
		ql4_printk(KERN_INFO, ha, "HW State: 0x%x = %s\n", dev_state,
			   ((dev_state < MAX_STATES) ?
			   qdev_state[dev_state] : "Unknown"));

	/* For ISP83xx, Reset owner is NIC, iSCSI or FCOE based on priority
	 * and which drivers are present. Unlike P3P, the function setting
	 * NEED_RESET, may not be the Reset owner
	 */
	if (qla4_83xx_can_perform_reset(ha))
		set_bit(AF_8XXX_RST_OWNER, &ha->flags);

	ha->isp_ops->idc_unlock(ha);

	rval = qla4_8xxx_device_state_handler(ha);

	ha->isp_ops->idc_lock(ha);
	qla4_8xxx_clear_rst_ready(ha);
	ha->isp_ops->idc_unlock(ha);

	if (rval == QLA_SUCCESS)
		clear_bit(AF_FW_RECOVERY, &ha->flags);

exit_isp_reset:
	return rval;
}

/*-------------------------Reset Sequence Functions-----------------------*/

void qla4_83xx_dump_reset_seq_hdr(struct scsi_qla_host *ha)
{
	u8 *phdr;

	if (!ha->reset_tmplt.buff) {
		ql4_err(ha, "%s: Error Invalid reset_seq_template\n", __func__);
		return ;
	}

	phdr = ha->reset_tmplt.buff;

	DEBUG2(ql4_info(ha, "Reset Template :\n\t0x%X 0x%X 0x%X 0x%X"
			"0x%X 0x%X 0x%X 0x%X 0x%X 0x%X\n"
			"\t0x%X 0x%X 0x%X 0x%X 0x%X 0x%X\n\n",
			*phdr, *(phdr+1), *(phdr+2), *(phdr+3), *(phdr+4),
			*(phdr+5), *(phdr+6), *(phdr+7), *(phdr + 8),
			*(phdr+9), *(phdr+10), *(phdr+11), *(phdr+12),
			*(phdr+13), *(phdr+14), *(phdr+15)));
}

int qla4_83xx_copy_bootloader(struct scsi_qla_host *ha)
{
	uint8_t *p_cache;
	uint32_t src, count, size;
	uint64_t dest;
	int ret_val = QLA_SUCCESS;

	src = QLA83XX_BOOTLOADER_FLASH_ADDR;
	dest = qla4_83xx_rd_reg(ha, QLA83XX_BOOTLOADER_ADDR);
	size = qla4_83xx_rd_reg(ha, QLA83XX_BOOTLOADER_SIZE);

	/* 128 bit alignment check */
	if (size & 0xF)
		size = (size + 16) & ~0xF;

	/* 16 byte count */
	count = size/16;

	p_cache = vmalloc(size);
	if (p_cache == NULL) {
		ql4_err(ha, "%s: Failed to allocate memory for "
			"boot loader cache\n", __func__);
		ret_val = QLA_ERROR;
		goto exit_copy_bootloader;
	}

	ret_val = qla4_83xx_lockless_flash_read_u32(ha, src,
					p_cache, size/sizeof(uint32_t));
	if (ret_val == QLA_ERROR) {
		ql4_err(ha, "%s: Error reading F/W from flash!!!\n", __func__);
		goto exit_copy_error;
	}
	DEBUG2(ql4_info(ha, "%s: Read F/W from flash!\n", __func__));

	/* 128 bit/16 byte write to MS memory */
	ret_val = qla4_83xx_ms_mem_write_128b(ha, dest,
					(uint32_t *)p_cache, count);
	if (ret_val == QLA_ERROR) {
		ql4_err(ha, "%s: Error writing F/W to MS !!!\n", __func__);
		goto exit_copy_error;
	}

	DEBUG2(ql4_info(ha, "%s: Wrote F/W (size %d) to MS !!!\n", 
		__func__, size));

exit_copy_error:
	vfree(p_cache);

exit_copy_bootloader:
	return ret_val;
}

/* 
 * qla4_83xx_check_cmd_peg_status - Check peg status to see if Peg is 
 * initialized.
 *
 * @ha : Pointer to adapter structure
 *
 * Return Value - QLA_SUCCESS/QLA_ERROR
 */ 
static int qla4_83xx_check_cmd_peg_status(struct scsi_qla_host *ha)
{
	uint32_t val, ret_val = QLA_ERROR;
	int retries = CRB_CMDPEG_CHECK_RETRY_COUNT;

	do {
		val = qla4_83xx_rd_reg(ha, QLA83XX_CMDPEG_STATE);
		if (val == PHAN_INITIALIZE_COMPLETE) {
			DEBUG2(ql4_info(ha, "%s: Command Peg initialization "
				"complete! state=0x%x\n", __func__, val));
			ret_val = QLA_SUCCESS;	
			break;
		}
		msleep(CRB_CMDPEG_CHECK_DELAY);
	} while (--retries);

	return ret_val;
}

/* 
 * qla4_83xx_poll_reg - Poll the given CRB addr for duration msecs till
 * value read ANDed with test_mask is equal to test_result.
 *
 * @ha : Pointer to adapter structure
 * @addr : CRB register address
 * @duration : Poll for total of "duration" msecs
 * @test_mask : Mask value read with "test_mask"
 * @test_result : Compare (value&test_mask) with test_result.
 *
 * Return Value - QLA_SUCCESS/QLA_ERROR
 */ 
static int qla4_83xx_poll_reg(struct scsi_qla_host *ha, uint32_t addr, 
		 	      int duration, uint32_t test_mask, 
			      uint32_t test_result)
{
	uint32_t value;
	int timeout_error;
	uint8_t retries;
	int ret_val = QLA_SUCCESS;

	ret_val = qla4_83xx_rd_reg_indirect(ha, addr, &value);
	if (ret_val == QLA_ERROR) {
		timeout_error = 1;
		goto exit_poll_reg;
	}

	/* poll every 1/10 of the total duration */
	retries = duration/10;

	do {
		if ((value & test_mask) != test_result) {
			timeout_error = 1;
			msleep(duration/10);
			ret_val = qla4_83xx_rd_reg_indirect(ha, addr, &value);
			if (ret_val == QLA_ERROR) {
				timeout_error = 1;
				goto exit_poll_reg;
			}
		} else {
			timeout_error = 0;
			break;
		}
	} while (retries--);

exit_poll_reg:
	if (timeout_error) {
		ha->reset_tmplt.seq_error++ ;
		ql4_err(ha, "%s: Poll Failed:  0x%08x 0x%08x 0x%08x\n",
			 __func__, value, test_mask, test_result);
	}

	return timeout_error;
}

/* 
 * qla4_83xx_reset_seq_checksum_test - Validate Reset Sequence template.
 *
 * @ha : Pointer to adapter structure
 *
 * Return Value - QLA_SUCCESS/QLA_ERROR
 */ 
static int qla4_83xx_reset_seq_checksum_test(struct scsi_qla_host *ha)
{
	uint32_t sum =  0;
	uint16_t *buff = (uint16_t *)ha->reset_tmplt.buff;
	int u16_count =  ha->reset_tmplt.hdr->size / sizeof(uint16_t);

	while (u16_count-- > 0)
		sum += *buff++;

	while (sum >> 16)
		sum = (sum & 0xFFFF) +  (sum >> 16);

	/* checksum of 0 indicates a valid template */
	if (~sum) {
		return QLA_SUCCESS;
	} else {
		ql4_err(ha, "%s: Reset seq checksum failed\n", __func__);
		return QLA_ERROR;
	}
}

/* 
 * qla4_83xx_read_reset_template - Read Reset Template from Flash, validate
 * the template and store offsets of stop/start/init offsets in ha->reset_tmplt.
 *
 * @ha : Pointer to adapter structure
 *
 * Return Value - QLA_SUCCESS/QLA_ERROR
 */ 
int qla4_83xx_read_reset_template(struct scsi_qla_host *ha)
{
	uint8_t *p_buff;
	uint32_t addr, tmplt_hdr_def_size, tmplt_hdr_size;
	int ret_val = QLA_SUCCESS;

	ha->reset_tmplt.seq_error = 0;
	ha->reset_tmplt.buff = vmalloc(QLA83XX_RESTART_TEMPLATE_SIZE);
	if (ha->reset_tmplt.buff == NULL) {
		ql4_err(ha, "%s: Failed to allocate reset template resources\n",
			__func__);
		ret_val = QLA_ERROR;
		goto exit_read_reset_template;
	}

	p_buff = ha->reset_tmplt.buff;
	addr = QLA83XX_RESET_TEMPLATE_ADDR;

	tmplt_hdr_def_size = sizeof(struct qla4_83xx_reset_template_hdr) / 
							sizeof(uint32_t);

	DEBUG2(ql4_info(ha, "%s: Read template hdr size %d from Flash\n", 
		__func__, tmplt_hdr_def_size));

	/* Copy template header from flash */
	if (qla4_83xx_flash_read_u32(ha, addr, p_buff, tmplt_hdr_def_size)) {
		ql4_err(ha, "%s: Failed to read reset template\n", __func__);
		ret_val = QLA_ERROR;
		goto exit_read_template_error;
	}

	ha->reset_tmplt.hdr =
	 (struct qla4_83xx_reset_template_hdr *) ha->reset_tmplt.buff;

	/* Validate the template header size and signature */
	tmplt_hdr_size = ha->reset_tmplt.hdr->hdr_size/sizeof(uint32_t);
	if ((tmplt_hdr_size != tmplt_hdr_def_size) ||
	    (ha->reset_tmplt.hdr->signature != RESET_TMPLT_HDR_SIGNATURE)) {
		ql4_err(ha, "%s: Template Header size invalid %d "
			"tmplt_hdr_def_size %d!!!\n", __func__, 
			tmplt_hdr_size, tmplt_hdr_def_size);
		ret_val = QLA_ERROR;
		goto exit_read_template_error;
	}

	addr = QLA83XX_RESET_TEMPLATE_ADDR + ha->reset_tmplt.hdr->hdr_size;
	p_buff = ha->reset_tmplt.buff + ha->reset_tmplt.hdr->hdr_size;
	tmplt_hdr_def_size = (ha->reset_tmplt.hdr->size -
			ha->reset_tmplt.hdr->hdr_size)/sizeof(uint32_t);

	DEBUG2(ql4_info(ha, "%s: Read rest of the template size %d\n", 
		__func__, ha->reset_tmplt.hdr->size));

	/* Copy rest of the template */
	if (qla4_83xx_flash_read_u32(ha, addr, p_buff, tmplt_hdr_def_size)) {
		ql4_err(ha, "%s: Failed to read reset tempelate\n", __func__);
		ret_val = QLA_ERROR;
		goto exit_read_template_error;
	}

	/* Integrity check */
	if (qla4_83xx_reset_seq_checksum_test(ha)) {
		ql4_err(ha, "%s: Reset Seq checksum failed!\n", __func__);
		ret_val = QLA_ERROR;
		goto exit_read_template_error;
	}

	DEBUG2(ql4_info(ha, "%s: Reset Seq checksum passed! Get stop, "
			"start and init seq offsets\n", __func__));
	/* Get STOP, START, INIT sequence offsets */
	ha->reset_tmplt.init_offset = ha->reset_tmplt.buff +
			ha->reset_tmplt.hdr->init_seq_offset;

	ha->reset_tmplt.start_offset = ha->reset_tmplt.buff +
			ha->reset_tmplt.hdr->start_seq_offset;

	ha->reset_tmplt.stop_offset = ha->reset_tmplt.buff +
			ha->reset_tmplt.hdr->hdr_size;

	qla4_83xx_dump_reset_seq_hdr(ha);

	goto exit_read_reset_template;

exit_read_template_error:
	vfree(ha->reset_tmplt.buff);

exit_read_reset_template:
	return ret_val;
}

void qla4_83xx_get_flash_info(struct scsi_qla_host *ha)
{
	uint32_t idc_major = 0;
	uint32_t idc_minor = 0;
	uint32_t idc_ctrl = 0;
	uint32_t dev_state = 0;
	uint32_t drv_active = 0;
	uint32_t dev_part1 = 0;
	uint32_t dev_part2 = 0;
	uint32_t idc_params;
	uint32_t lock_id;

	/* TBD: Added to check IDC register initialization by
	 * firmware, will be removed later.
	 */
	idc_major = qla4_83xx_rd_reg(ha, QLA83XX_CRB_IDC_VER_MAJOR);
	idc_minor = qla4_83xx_rd_reg(ha, QLA83XX_CRB_IDC_VER_MINOR);
	idc_ctrl = qla4_83xx_rd_reg(ha, QLA83XX_IDC_DRV_CTRL);
	dev_state = qla4_83xx_rd_reg(ha, QLA83XX_CRB_DEV_STATE);
	drv_active = qla4_83xx_rd_reg(ha, QLA83XX_CRB_DRV_ACTIVE);
	dev_part1 = qla4_83xx_rd_reg(ha, QLA83XX_CRB_DEV_PART_INFO1);
	dev_part2 = qla4_83xx_rd_reg(ha, QLA83XX_CRB_DEV_PART_INFO2);
	lock_id = qla4_83xx_rd_reg(ha, QLA83XX_DRV_LOCKRECOVERY);

	DEBUG2(ql4_info(ha, "%s: IDC Major#=%d, IDC Minor#=%d, IDC drv_ctrl=%d,"
                " IDC dev_state=%d idc_drv_active=%d, dev_part_info1=0x%x, "
                "lock_id=0x%x\n", __func__, idc_major, idc_minor, idc_ctrl,
                dev_state, drv_active, dev_part1, lock_id));

	if (qla4_83xx_flash_read_u32(ha, QLA83XX_IDC_PARAM_ADDR,
				     (uint8_t *)&idc_params, 1)) {
                ql4_err(ha, "%s: Failed to read idc params from flash\n",
                         __func__);
		ha->nx_dev_init_timeout = ROM_DEV_INIT_TIMEOUT;
		ha->nx_reset_timeout = ROM_DRV_RESET_ACK_TIMEOUT;
        } else {
		ha->nx_dev_init_timeout = idc_params & 0xFFFF;
		ha->nx_reset_timeout = (idc_params >> 16) & 0xFFFF;
	}

	DEBUG2(ql4_info(ha, "%s: dev_init_timeout=%d reset_ack_timeout=%d\n",
			__func__, ha->nx_dev_init_timeout,
			ha->nx_reset_timeout));

	/* Copy the firmware defined Reset Sequence from flash */
	if (qla4_83xx_read_reset_template(ha))
		ql4_err(ha, "%s: Read Reset template Error!!!\n", __func__);
	else
		DEBUG2(ql4_info(ha, "%s: Read Init Reset template done!\n",
			__func__));
}

/* 
 * qla4_83xx_read_write_crb_reg - Read from raddr and write value to waddr.
 *
 * @ha : Pointer to adapter structure
 * @raddr : CRB address to read from
 * @waddr : CRB address to write to
 *
 */ 
static void qla4_83xx_read_write_crb_reg(struct scsi_qla_host *ha, 
					 uint32_t raddr, uint32_t waddr)
{
	uint32_t value;

	qla4_83xx_rd_reg_indirect(ha, raddr, &value);
	qla4_83xx_wr_reg_indirect(ha, waddr, value);
}

/* 
 * qla4_83xx_rmw_crb_reg - Read value from raddr, AND with test_mask, 
 * Shift Left,Right/OR/XOR with values RMW header and write value to waddr.
 *
 * @ha : Pointer to adapter structure
 * @raddr : CRB address to read from
 * @waddr : CRB address to write to
 * @p_rmw_hdr : header with shift/or/xor values.
 *
 */ 
static void qla4_83xx_rmw_crb_reg(struct scsi_qla_host *ha, uint32_t raddr, 
                    		  uint32_t waddr, 
				  struct qla4_83xx_rmw *p_rmw_hdr)
{
	uint32_t value;

	if (p_rmw_hdr->index_a)
		value = ha->reset_tmplt.array[p_rmw_hdr->index_a];
	else
		qla4_83xx_rd_reg_indirect(ha, raddr, &value);
		
	value &= p_rmw_hdr->test_mask;
	value <<= p_rmw_hdr->shl;
	value >>= p_rmw_hdr->shr;
	value |= p_rmw_hdr->or_value;
	value ^= p_rmw_hdr->xor_value;

	qla4_83xx_wr_reg_indirect(ha, waddr, value);

	return;
}

/* 
 * qla4_83xx_write_list - Write the value (p_entry->arg2) to address specified 
 * by p_entry->arg1 for all entries in header with delay of p_hdr->delay between
 * entries.
 *
 * @ha : Pointer to adapter structure
 * @p_hdr : reset_entry header for WRITE_LIST opcode.
 *
 */ 
static void qla4_83xx_write_list(struct scsi_qla_host *ha,
		   		 struct qla4_83xx_reset_entry_hdr *p_hdr)
{
	struct qla4_83xx_entry *p_entry;
	uint32_t i;

	p_entry = (struct qla4_83xx_entry *)((char *)p_hdr +
				sizeof(struct qla4_83xx_reset_entry_hdr));

	for (i = 0; i < p_hdr->count; i++, p_entry++) {
		qla4_83xx_wr_reg_indirect(ha, p_entry->arg1,
						p_entry->arg2);
		if (p_hdr->delay)
			udelay((uint32_t)(p_hdr->delay));
	}
}

/* 
 * qla4_83xx_read_write_list - Read from address specified by p_entry->arg1, 
 * write value read to address specified by p_entry->arg2, for all entries in 
 * header with delay of p_hdr->delay between entries.
 *
 * @ha : Pointer to adapter structure
 * @p_hdr : reset_entry header for READ_WRITE_LIST opcode.
 *
 */ 
static void qla4_83xx_read_write_list(struct scsi_qla_host *ha,
				      struct qla4_83xx_reset_entry_hdr *p_hdr)
{
	struct qla4_83xx_entry *p_entry;
	uint32_t i;

	p_entry = (struct qla4_83xx_entry *)((char *)p_hdr +
				sizeof(struct qla4_83xx_reset_entry_hdr));

	for (i = 0; i < p_hdr->count; i++, p_entry++) {
		qla4_83xx_read_write_crb_reg(ha, p_entry->arg1,
						p_entry->arg2);
		if (p_hdr->delay)
			udelay((uint32_t)(p_hdr->delay));
	}
}

/* 
 * qla4_83xx_poll_list - For all entries in the POLL_LIST header, poll read CRB 
 * register specified by p_entry->arg1 and compare (value AND test_mask) with 
 * test_result to validate it. Wait for p_hdr->delay between processing entries.
 *
 * @ha : Pointer to adapter structure
 * @p_hdr : reset_entry header for POLL_LIST opcode.
 *
 */ 
static void qla4_83xx_poll_list(struct scsi_qla_host *ha,
		  		struct qla4_83xx_reset_entry_hdr *p_hdr)
{
	long delay;
	struct qla4_83xx_entry *p_entry;
	struct qla4_83xx_poll *p_poll;
	uint32_t i;
	uint32_t value;

	p_poll = (struct qla4_83xx_poll *)
		((char *)p_hdr + sizeof(struct qla4_83xx_reset_entry_hdr));

	/* Entries start after 8 byte qla4_83xx_poll, poll header contains 
 	 * the test_mask, test_value.
 	 */
	p_entry = (struct qla4_83xx_entry *)((char *)p_poll +
				sizeof(struct qla4_83xx_poll));
	
	delay = (long)p_hdr->delay;

	if (!delay) {
		for (i = 0; i < p_hdr->count; i++, p_entry++) {
			qla4_83xx_poll_reg(ha, p_entry->arg1,
				delay, p_poll->test_mask, p_poll->test_value);
		}
	} else {
		for (i = 0; i < p_hdr->count; i++, p_entry++) {
		    if (delay) {
			if (qla4_83xx_poll_reg(ha,
				       p_entry->arg1, delay,
				       p_poll->test_mask,
				       p_poll->test_value)) {
				/* If (data_read&test_mask != test_value), read
 			 	 * TIMEOUT_ADDR (arg1) and ADDR (arg2) registers
 			 	 */
				qla4_83xx_rd_reg_indirect(ha,
					p_entry->arg1, &value);
				qla4_83xx_rd_reg_indirect(ha,
					p_entry->arg2, &value);
			}
		    }
		}
	}
}

/* 
 * qla4_83xx_poll_write_list - Write dr_value, ar_value to dr_addr/ar_addr,
 * read ar_addr, if (value& test_mask != test_mask) re-read till timeout 
 * expires.
 *
 * @ha : Pointer to adapter structure
 * @p_hdr : reset entry header for POLL_WRITE_LIST opcode.
 *
 */ 
static void qla4_83xx_poll_write_list(struct scsi_qla_host *ha,
				      struct qla4_83xx_reset_entry_hdr *p_hdr)
{
	long delay;
	struct qla4_83xx_quad_entry *p_entry;
	struct qla4_83xx_poll *p_poll;
	uint32_t i;

	p_poll = (struct qla4_83xx_poll *)((char *)p_hdr +
				sizeof(struct qla4_83xx_reset_entry_hdr));

	p_entry = (struct qla4_83xx_quad_entry *)((char *)p_poll +
				sizeof(struct qla4_83xx_poll));

	delay = (long)p_hdr->delay;

	for (i = 0; i < p_hdr->count; i++, p_entry++) {
		qla4_83xx_wr_reg_indirect(ha,
			p_entry->dr_addr, p_entry->dr_value);
		qla4_83xx_wr_reg_indirect(ha,
			p_entry->ar_addr, p_entry->ar_value);
		if (delay) {
			if (qla4_83xx_poll_reg(ha,
				       p_entry->ar_addr, delay,
				       p_poll->test_mask,
				       p_poll->test_value)) {
				DEBUG2(ql4_info(ha, "%s: Timeout Error: poll "
					"list, ", __func__));
				DEBUG2(ql4_info(ha, "item_num %d, "
					"entry_num %d\n", i, 
					ha->reset_tmplt.seq_index));
			}
		}
	}
}

/* 
 * qla4_83xx_read_modify_write - Read value from p_entry->arg1, modify the 
 * value, write value to p_entry->arg2. Process entries with p_hdr->delay 
 * between entries.
 *
 * @ha : Pointer to adapter structure
 * @p_hdr : header with shift/or/xor values.
 *
 */ 
static void qla4_83xx_read_modify_write(struct scsi_qla_host *ha,
			  		struct qla4_83xx_reset_entry_hdr *p_hdr)
{
	struct qla4_83xx_entry *p_entry;
	struct qla4_83xx_rmw *p_rmw_hdr;
	uint32_t i;

	p_rmw_hdr = (struct qla4_83xx_rmw *)((char *)p_hdr +
				sizeof(struct qla4_83xx_reset_entry_hdr));

	p_entry = (struct qla4_83xx_entry *)((char *)p_rmw_hdr +
					sizeof(struct qla4_83xx_rmw));

	for (i = 0; i < p_hdr->count; i++, p_entry++) {
		qla4_83xx_rmw_crb_reg(ha, p_entry->arg1,
					p_entry->arg2, p_rmw_hdr);
		if (p_hdr->delay)
			udelay((uint32_t)(p_hdr->delay));
	}
}

/* 
 * qla4_83xx_pause - Wait for p_hdr->delay msecs, called between processing 
 * two entries of a sequence.
 *
 * @ha : Pointer to adapter structure
 * @p_hdr : Common reset entry header.
 *
 */ 
static void qla4_83xx_pause(struct scsi_qla_host *ha,
	      		    struct qla4_83xx_reset_entry_hdr *p_hdr)
{
	if (p_hdr->delay)
		mdelay((uint32_t)((long)p_hdr->delay));
}

/* 
 * qla4_83xx_poll_read_list - Write ar_value to ar_addr register, read ar_addr, 
 * if (value & test_mask != test_value) re-read till timeout value expires, 
 * read dr_addr register and assign to reset_tmplt.array.
 *
 * @ha : Pointer to adapter structure
 * @p_hdr : Common reset entry header.
 *
 */ 
static void qla4_83xx_poll_read_list(struct scsi_qla_host *ha,
		       		     struct qla4_83xx_reset_entry_hdr *p_hdr)
{
	long delay;
	int index;
	struct qla4_83xx_quad_entry *p_entry;
	struct qla4_83xx_poll *p_poll;
	uint32_t i;
	uint32_t value;

	p_poll = (struct qla4_83xx_poll *)
		((char *)p_hdr + sizeof(struct qla4_83xx_reset_entry_hdr));

	p_entry = (struct qla4_83xx_quad_entry *)
		((char *)p_poll + sizeof(struct qla4_83xx_poll));

	delay = (long)p_hdr->delay;

	for (i = 0; i < p_hdr->count; i++, p_entry++) {
		qla4_83xx_wr_reg_indirect(ha, p_entry->ar_addr,
					 p_entry->ar_value);
		if (delay) {
			if (qla4_83xx_poll_reg(ha, p_entry->ar_addr, delay,
					     p_poll->test_mask, 
					     p_poll->test_value)){
				DEBUG2(ql4_info(ha, "%s: Timeout Error: poll "
					"list, ", __func__));
				DEBUG2(ql4_info(ha, "Item_num %d, "
					"entry_num %d\n", i, 
					ha->reset_tmplt.seq_index));
			} else {
				index = ha->reset_tmplt.array_index;
				qla4_83xx_rd_reg_indirect(ha,
							p_entry->dr_addr,
							&value);
				ha->reset_tmplt.array[index++] = value;

				if (index == QLA83XX_MAX_RESET_SEQ_ENTRIES)
					ha->reset_tmplt.array_index = 1;
			}
		}
	}
}

/* 
 * qla4_83xx_seq_end - Informs driver that processing has reached end of a 
 * subsequence.
 *
 * @ha : Pointer to adapter structure
 * @p_hdr : Common reset entry header.
 *
 */ 
static void qla4_83xx_seq_end(struct scsi_qla_host *ha,
			      struct qla4_83xx_reset_entry_hdr *p_hdr)
{
	ha->reset_tmplt.seq_end = 1;
}

/* 
 * qla4_83xx_template_end - Indicates end of reset sequence processing.
 *
 * @ha : Pointer to adapter structure
 * @p_hdr : Common reset entry header.
 *
 */ 
static void qla4_83xx_template_end(struct scsi_qla_host *ha,
		     		   struct qla4_83xx_reset_entry_hdr *p_hdr)
{
	ha->reset_tmplt.template_end = 1;

	if (ha->reset_tmplt.seq_error == 0) {
		DEBUG2(ql4_info(ha, "%s: Reset sequence completed "
			 "SUCCESSFULLY.\n", __func__));
	} else {
		ql4_err(ha, "%s: Reset sequence completed with "
			 "some timeout errors.\n", __func__);
	}
}

/* 
 * qla4_83xx_process_reset_template - Process all entries in reset template 
 * till entry with SEQ_END opcode, which indicates end of the reset template 
 * processing. Each entry has a Reset Entry header, entry opcode/command, with 
 * size of the entry, number of entries in sub-sequence and delay in microsecs 
 * or timeout in millisecs.
 *
 * @ha : Pointer to adapter structure
 * @p_buff : Common reset entry header.
 *
 */ 
static void qla4_83xx_process_reset_template(struct scsi_qla_host *ha, 
					     char *p_buff)
{
	int index, entries;
	struct qla4_83xx_reset_entry_hdr *p_hdr;
	char *p_entry = p_buff;

	ha->reset_tmplt.seq_end = 0;
	ha->reset_tmplt.template_end = 0;
	entries = ha->reset_tmplt.hdr->entries;
	index = ha->reset_tmplt.seq_index;

	for ( ; (!ha->reset_tmplt.seq_end) && (index  < entries); index++) {

		p_hdr = (struct qla4_83xx_reset_entry_hdr *)p_entry;

		DEBUG2(ql4_info(ha, "ISP83xx_DBG:%s: phdr_cmd = %d\n", __func__, 
				p_hdr->cmd));

		switch (p_hdr->cmd) {
		case OPCODE_NOP:
			break;
		case OPCODE_WRITE_LIST:
			qla4_83xx_write_list(ha, p_hdr);
			break;
		case OPCODE_READ_WRITE_LIST:
			qla4_83xx_read_write_list(ha, p_hdr);
			break;
		case OPCODE_POLL_LIST:
			qla4_83xx_poll_list(ha, p_hdr);
			break;
		case OPCODE_POLL_WRITE_LIST:
			qla4_83xx_poll_write_list(ha, p_hdr);
			break;
		case OPCODE_READ_MODIFY_WRITE:
			qla4_83xx_read_modify_write(ha, p_hdr);
			break;
		case OPCODE_SEQ_PAUSE:
			qla4_83xx_pause(ha, p_hdr);
			break;
		case OPCODE_SEQ_END:
			qla4_83xx_seq_end(ha, p_hdr);
			break;
		case OPCODE_TMPL_END:
			qla4_83xx_template_end(ha, p_hdr);
			break;
		case OPCODE_POLL_READ_LIST:
			qla4_83xx_poll_read_list(ha, p_hdr);
			break;
		default:
			ql4_err(ha, "%s: Unknown command ==> 0x%04x on "
				 "entry = %d\n", __func__, p_hdr->cmd, index);
			break;
		}

		/* Set pointer to next entry in the sequence. */
		p_entry += p_hdr->size;
	}

	ha->reset_tmplt.seq_index = index;
}

static void qla4_83xx_process_stop_seq(struct scsi_qla_host *ha)
{
	DEBUG2(ql4_info(ha, "ISP83xx_DBG: %s!\n", __func__));

	ha->reset_tmplt.seq_index = 0;
	qla4_83xx_process_reset_template(ha,
			ha->reset_tmplt.stop_offset);

	if (ha->reset_tmplt.seq_end != 1)
		ql4_err(ha, "%s: Abrupt STOP Sub-Sequence end.\n", 
			 __func__);
}

static void qla4_83xx_process_start_seq(struct scsi_qla_host *ha)
{
	DEBUG2(ql4_info(ha, "ISP83xx_DBG: %s!\n", __func__));
	qla4_83xx_process_reset_template(ha,
			ha->reset_tmplt.start_offset);

	if (ha->reset_tmplt.template_end != 1)
		ql4_err(ha, "%s: Abrupt START Sub-Sequence end.\n",
			 __func__);
}

static void qla4_83xx_process_init_seq(struct scsi_qla_host *ha)
{
	DEBUG2(ql4_info(ha, "ISP83xx_DBG: %s!\n", __func__));
	qla4_83xx_process_reset_template(ha,
			ha->reset_tmplt.init_offset);

	if (ha->reset_tmplt.seq_end != 1)
		ql4_err(ha, "%s: Abrupt INIT Sub-Sequence end.\n",
			 __func__);
}

int qla4_83xx_restart(struct scsi_qla_host *ha)
{
	int ret_val = QLA_SUCCESS;

	qla4_83xx_process_stop_seq(ha);

	/* TBD: Get Firmware Dump if AF_FW_RECOVERY (AF_FW_RECOVERY is
 	 * set for F/W Hang, PCI ERROR) && IDC_CTRL GRACEFUL_RESET_BIT1 is not 
 	 * set. IDC_CTRL GRACEFUL_RESET_BIT1 should be set by the driver doing 
 	 * the Flash Update.
 	 */
	/* Collect minidump*/
	if (!test_and_clear_bit(AF_83XX_NO_FW_DUMP, &ha->flags))
		qla4_8xxx_get_minidump(ha);

	qla4_83xx_process_init_seq(ha);

	if (qla4_83xx_copy_bootloader(ha)) {
		ql4_err(ha, "%s: Copy bootloader, firmware restart failed!\n", 
			__func__);
		ret_val = QLA_ERROR;
		goto exit_restart;
	}

	/* If FW_IMAGE_VALID is QLA83XX_BOOT_FROM_FLASH,
 	 * Bootloader loads F/W from flash 
 	 */
	qla4_83xx_wr_reg(ha, QLA83XX_FW_IMAGE_VALID, QLA83XX_BOOT_FROM_FLASH);

	qla4_83xx_process_start_seq(ha);

exit_restart:
	return ret_val;
}

int qla4_83xx_start_firmware(struct scsi_qla_host *ha)
{
	int ret_val = QLA_SUCCESS;

	if (qla4_83xx_restart(ha)) {
		ql4_err(ha, "%s: Restart Error!!!, Need Reset!!!\n", 
			__func__);
		ret_val = QLA_ERROR;
		goto exit_start_fw;
	} else
		DEBUG2(ql4_info(ha, "%s: Restart done!\n", __func__));

	ret_val = qla4_83xx_check_cmd_peg_status(ha);
	if (ret_val) {
		ql4_err(ha, "%s: Peg not initialized!\n", __func__);
		ret_val = QLA_ERROR;
	}

exit_start_fw:
	return ret_val;
}

/*----------------------Interrupt Related functions ---------------------*/

void qla4_83xx_disable_intrs(struct scsi_qla_host *ha)
{
	uint32_t mb_int, ret;

	if (test_and_clear_bit(AF_INTERRUPTS_ON, &ha->flags))
		qla4_8xxx_mbx_intr_disable(ha);
	
	ret = readl(&ha->qla4_83xx_reg->mbox_int);
	mb_int = ret & ~INT_ENABLE_FW_MB;
	writel(mb_int, &ha->qla4_83xx_reg->mbox_int);
	writel(1, &ha->qla4_83xx_reg->leg_int_mask);
}

void qla4_83xx_enable_intrs(struct scsi_qla_host *ha)
{
	uint32_t mb_int;

	qla4_8xxx_mbx_intr_enable(ha);
	mb_int = INT_ENABLE_FW_MB;
	writel(mb_int, &ha->qla4_83xx_reg->mbox_int);
	writel(0, &ha->qla4_83xx_reg->leg_int_mask);

	set_bit(AF_INTERRUPTS_ON, &ha->flags);
}

void qla4_83xx_queue_mbox_cmd(struct scsi_qla_host *ha, uint32_t *mbx_cmd,
			      int incount)
{
	int i;

	/* Load all mailbox registers, except mailbox 0. */
	for (i = 1; i < incount; i++)
		writel(mbx_cmd[i], &ha->qla4_83xx_reg->mailbox_in[i]);

	writel(mbx_cmd[0], &ha->qla4_83xx_reg->mailbox_in[0]);

	/* Set Host Interrupt register to 1, to tell the firmware that
 	* a mailbox command is pending. Firmware after reading the
 	* mailbox command, clears the host interrupt register.
 	*/
	writel(HINT_MBX_INT_PENDING, &ha->qla4_83xx_reg->host_intr);
}

void qla4_83xx_process_mbox_intr(struct scsi_qla_host *ha, int outcount)
{
	int intr_status;

	intr_status = readl(&ha->qla4_83xx_reg->risc_intr);
	if (intr_status) {
		ha->mbox_status_count = outcount;
		ha->isp_ops->interrupt_service_routine(ha, intr_status);
	}
	writel(0, &ha->qla4_83xx_reg->risc_intr);
}
