/*
 * QLogic iSCSI HBA Driver
 * Copyright (c)  2003-2011 QLogic Corporation
 *
 * See LICENSE.qla4xxx for copyright and licensing details.
 */

#include "ql4_def.h"
#include "ql4_glbl.h"
#include "ql4_dbg.h"

/* SYSFS Attributes */

static ssize_t
qla4_8xxx_read_from_buffer(void *buf, size_t count, loff_t *ppos,
			   const void *from, size_t size)
{
	loff_t pos = *ppos;

	if (pos < 0)
		return -EINVAL;
	if (pos >= size)
		return 0;
	if (count > size - pos)
		count = size - pos;
	memcpy(buf, from + pos, count);
	*ppos = pos + count;

	return count;
}

static ssize_t
qla4_8xxx_sysfs_read_fw_dump(struct kobject *kobj, char *buf, loff_t off,
			    size_t count)
{
	struct scsi_qla_host *ha = to_qla_host(dev_to_shost(container_of(kobj,
					       struct device, kobj)));

	if (!IS_QLA80XX(ha))
		return -EINVAL;

	if (!test_bit(AF_82XX_DUMP_READING, &ha->flags))
		return 0;

	return qla4_8xxx_read_from_buffer(buf, count, &off, ha->fw_dump,
					  ha->fw_dump_size);
}

static ssize_t
qla4_8xxx_sysfs_write_fw_dump(struct kobject *kobj, char *buf, loff_t off,
			      size_t count)
{
	struct scsi_qla_host *ha = to_qla_host(dev_to_shost(container_of(kobj,
					       struct device, kobj)));
	int reading;
	uint32_t dev_state;

	if (!IS_QLA80XX(ha))
		return -EINVAL;

	if (off != 0)
		return 0;

	reading = simple_strtol(buf, NULL, 10);
	switch (reading) {
	case 0:
		if (test_and_clear_bit(AF_82XX_DUMP_READING, &ha->flags)) {
			clear_bit(AF_82XX_FW_DUMPED, &ha->flags);
			DEBUG2(ql4_info(ha, "Firmware dump cleared on (%d).\n",
					ha->host_no));
		}

		break;
	case 1:
		if (test_bit(AF_82XX_FW_DUMPED, &ha->flags) &&
		    !test_bit(AF_82XX_DUMP_READING, &ha->flags)) {
			set_bit(AF_82XX_DUMP_READING, &ha->flags);
			ql4_info(ha, "Raw firmware dump ready for read "
				 "on (%d).\n", ha->host_no);
		}
		break;
	case 2:
		/* Placeholder case for similarity with FC */
		break;
	case 3:
		ha->isp_ops->idc_lock(ha);
		dev_state = qla4_8xxx_rd_direct(ha, QLA8XXX_CRB_DEV_STATE);
		if (dev_state == QLA8XXX_DEV_READY) {
			ql4_info(ha, "%s: Setting Need reset, RESET_OWNER is "
				 "0x%x.\n", __func__, ha->func_num);
			qla4_8xxx_wr_direct(ha, QLA8XXX_CRB_DEV_STATE,
					QLA8XXX_DEV_NEED_RESET);
			set_bit(AF_8XXX_RST_OWNER, &ha->flags);
		} else
			ql4_info(ha, "%s: Device state is 0x%x\n", __func__,
				 dev_state);

		ha->isp_ops->idc_unlock(ha);
		break;
	default:
		/* do nothing */
		break;
	}

	return count;
}

static struct bin_attribute sysfs_fw_dump_attr = {
	.attr = {
		.name = "fw_dump",
		.mode = S_IRUSR | S_IWUSR,
		.owner = THIS_MODULE,
	},
	.size = 0,
	.read = qla4_8xxx_sysfs_read_fw_dump,
	.write = qla4_8xxx_sysfs_write_fw_dump,
};

static struct sysfs_entry {
	char *name;
	struct bin_attribute *attr;
} bin_file_entries[] = {
	{ "fw_dump", &sysfs_fw_dump_attr },
	{ NULL },
};

void
qla4_8xxx_alloc_sysfs_attr(struct scsi_qla_host *ha)
{
	struct Scsi_Host *host = ha->host;
	struct sysfs_entry *iter;
	int ret;

	for (iter = bin_file_entries; iter->name; iter++) {
		ret = sysfs_create_bin_file(&host->shost_gendev.kobj,
					    iter->attr);
		if (ret)
			ql4_info(ha, "Unable to create sysfs %s binary "
				 "attribute (%d).\n", iter->name, ret);
	}
}

void
qla4_8xxx_free_sysfs_attr(struct scsi_qla_host *ha)
{
	struct Scsi_Host *host = ha->host;
	struct sysfs_entry *iter;

	for (iter = bin_file_entries; iter->name; iter++) {
		sysfs_remove_bin_file(&host->shost_gendev.kobj,
		    iter->attr);
	}
}

uint64_t crbaddr(uint64_t base, uint64_t offset)
{
	return (0x00000000 + (base + offset));
}

int64_t pegreg_addr(u64 peg, u64 offset)
{
	if (peg == 4) {
		return (crbaddr(0x00f00000, offset));
	}
	return (crbaddr((0x01100000 + 0x00100000 * peg), offset));
}

void qla4_8xxx_wrpegreg(struct scsi_qla_host *ha, u64 peg, u64 offset, u32 value)
{
	u64 addr = 0;

	addr = pegreg_addr(peg, offset);
	ql4_info(ha, "peg-halt: peg=0x%llx addr=0x%llx\n",
		 (unsigned long long) peg, (unsigned long long) addr);

	if (IS_QLA8022(ha))
		qla4_82xx_wr_32(ha, (addr + QLA82XX_PCI_CRBSPACE), value);
	else if (IS_QLA8032(ha))
		qla4_83xx_wr_reg_indirect(ha, (addr + 0x32F00000), value);
}

static void qla4_8xxx_peg_halt(struct scsi_qla_host *ha, uint32_t peg)
{
	int iter = 0;

	if ((peg < 0 || peg >= 5) && peg != 0xff) {
		printk("Bad peg number: %d", peg);
		return;
	}

	if (peg == 0xff) {
		for (iter = 0; iter < 5; iter++)
			qla4_8xxx_wrpegreg(ha, iter, 0x3c, 0x1);
	} else
		qla4_8xxx_wrpegreg(ha, peg, 0x3c, 0x1);
}

static ssize_t qla4xxx_peg_hlt_store(struct class_device *dev,
				     const char *buf, size_t count)
{
	struct scsi_qla_host *ha = to_qla_host(class_to_shost(dev));
	int val = 0;

	if (IS_QLA40XX(ha))
		return -EINVAL;

	if (sscanf(buf, "%d", &val) != 1)
		return -EINVAL;

	if (val == 6024) {
		ha->isp_ops->idc_lock(ha);
		qla4_8xxx_peg_halt(ha, 0);
		ha->isp_ops->idc_unlock(ha);
	} else
		return -EINVAL;

	return strlen(buf);
}

static CLASS_DEVICE_ATTR(peg_hlt, S_IWUSR, NULL, qla4xxx_peg_hlt_store);

struct class_device_attribute *qla4xxx_host_attrs[] = {
	&class_device_attr_peg_hlt,
	NULL,
};
