/*
 * QLogic iSCSI HBA Driver
 * Copyright (c)  2003-2012 QLogic Corporation
 *
 * See LICENSE.qla4xxx for copyright and licensing details.
 */

static inline void
qla4xxx_queue_aen(scsi_qla_host_t *ha, uint32_t *mbox_sts)
{
	ha->aen_count++;
	if (ha->aen_count == 0) {
		/* We overflowed the counter */
		ha->aen_report = 0;
	}

	if (ha->aen_count - ha->aen_report == MAX_AEN_ENTRIES) {
		/* We wrapped the queue,
		 * update starting point for SDMAPI copy */
		ha->aen_report++;
	}

	if (ha->aen_in == (MAX_AEN_ENTRIES - 1))
		ha->aen_in = 0;
	else
		ha->aen_in++;

	memcpy(&ha->aen_q[ha->aen_in].mbox_sts[0],
		mbox_sts,
		sizeof(mbox_sts[0]) * MBOX_AEN_REG_COUNT);
}

/**
 * qla4xxx_queue_aen_log - queue AENs to be reported to application layer
 * @ha: Pointer to host adapter structure.
 * @mbox_sts: Pointer to mailbox status structure.
 *
 * Store AENs to be reported to the application layer when requested
 **/
static inline void
qla4xxx_queue_aen_log(struct scsi_qla_host *ha, uint32_t *mbox_sts)
{
	int i;
	if (ha->aen_log.count < MAX_AEN_ENTRIES) {
		for (i = 0; i < MBOX_AEN_REG_COUNT; i++)
			ha->aen_log.entry[ha->aen_log.count].mbox_sts[i] =
				mbox_sts[i];
		ha->aen_log.count++;
	}
}

static inline void qla4xxx_queue_lun_change_aen(scsi_qla_host_t *ha,
						     uint32_t index)
{
	uint32_t mbox_sts[MBOX_REG_COUNT];
	memset(mbox_sts, 0, sizeof(mbox_sts));
	mbox_sts[0] = MBOX_DRVR_ASTS_LUN_STATUS_CHANGE;
	mbox_sts[1] = index;
	qla4xxx_queue_aen_log(ha, &mbox_sts[0]);

	QL4PRINT(QLP7, printk("scsi%d: %s: AEN 0x7003 index[%d]\n",
		ha->host_no, __func__, index));
}

/**************************************************************************
 * qla4xxx_lookup_lun_handle
 *	This routine locates a lun handle given the device handle and lun
 *	number.
 *
 * Input:
 * 	ha - Pointer to host adapter structure.
 *	ddb_entry - Pointer to device database entry
 *	lun - SCSI LUN
 *
 * Returns:
 *	Pointer to corresponding lun_entry structure
 *
 * Context:
 *	Kernel context.
 **************************************************************************/
static inline os_lun_t *
qla4xxx_lookup_lun_handle(scsi_qla_host_t *ha, os_tgt_t *tq, uint16_t lun)
{
	os_lun_t *lq = NULL;

	if (tq && lun < MAX_LUNS)
		lq = tq->olun[lun];
	return lq;
}

/**************************************************************************
 * qla4xxx_lookup_target_by_SCSIID
 *	This routine locates a target handle given the SCSI bus and
 *	target IDs.  If device doesn't exist, returns NULL.
 *
 * Input:
 * 	ha - Pointer to host adapter structure.
 *	bus - SCSI bus number
 *	target - SCSI target ID.
 *
 * Returns:
 *	Pointer to the corresponding internal device database structure
 *
 * Context:
 *	Kernel context.
 **************************************************************************/
static inline os_tgt_t *
qla4xxx_lookup_target_by_SCSIID(scsi_qla_host_t *ha, uint32_t bus,
    uint32_t target)
{
	os_tgt_t *tq = NULL;

	if (target < MAX_TARGETS)
		tq = TGT_Q(ha, target);

	QL4PRINT(QLP3, printk("scsi%d: %s: b%d:t%d, tgt = %p\n",
	    ha->host_no, __func__, bus, target, tq));

	return tq;
}

/**************************************************************************
 * qla4xxx_lookup_target_by_fcport
 *	This routine locates a target handle given the fcport
 *
 * Input:
 * 	ha - Pointer to host adapter structure.
 *	fcport - port handle
 *
 * Returns:
 *
 * Context:
 *	Kernel context.
 **************************************************************************/
static inline os_tgt_t *
qla4xxx_lookup_target_by_fcport(scsi_qla_host_t *ha, fc_port_t  *fcport)
{
	int t;
	os_tgt_t *tq = NULL;

	for (t = 0; t < MAX_TARGETS; t++) {
		if ((tq = TGT_Q(ha, t)) == NULL)
			continue;

		if (fcport == tq->fcport)
			break;
	}

	return tq;
}


/**************************************************************************
 * qla4xxx_lookup_ddb_by_fw_index
 *	This routine locates a device handle given the firmware device
 *	database index.  If device doesn't exist, returns NULL.
 *
 * Input:
 * 	ha - Pointer to host adapter structure.
 *      fw_ddb_index - Firmware's device database index
 *
 * Returns:
 *	Pointer to the corresponding internal device database structure
 *
 * Context:
 *	Kernel context.
 **************************************************************************/
static inline ddb_entry_t *
qla4xxx_lookup_ddb_by_fw_index(scsi_qla_host_t *ha, uint32_t fw_ddb_index)
{
	ddb_entry_t *ddb_entry = NULL;

	if ((fw_ddb_index < MAX_DDB_ENTRIES) &&
	    (ha->fw_ddb_index_map[fw_ddb_index] !=
		(ddb_entry_t *) INVALID_ENTRY)) {
		ddb_entry = ha->fw_ddb_index_map[fw_ddb_index];
	}

	DEBUG3(printk("scsi%d: %s: ddb [%d], ddb_entry = %p\n",
	    ha->host_no, __func__, fw_ddb_index, ddb_entry));

	return ddb_entry;
}

/**************************************************************************
 * qla4xxx_enable_intrs
 *	This routine enables the PCI interrupt request by clearing the
 *	appropriate bit.
 *
 * qla4xxx_disable_intrs
 *	This routine disables the PCI interrupt request by setting the
 *	appropriate bit.
 *
 * Remarks:
 *	The hardware_lock must be unlocked upon entry.
 *
 * Input:
 * 	ha - Pointer to host adapter structure.
 *
 * Returns:
 *	None
 *
 * Context:
 *	Kernel/Interrupt context.
 **************************************************************************/
static inline void __qla4xxx_enable_intrs(scsi_qla_host_t *ha)
{
	if( IS_QLA4010(ha) ) {
		WRT_REG_DWORD(&ha->reg->ctrl_status, SET_RMASK(CSR_SCSI_INTR_ENABLE));
		PCI_POSTING(&ha->reg->ctrl_status);
		QL4PRINT(QLP7, printk("scsi%d: %s: intSET_RMASK = %08x\n",
			      ha->host_no, __func__,
			      RD_REG_DWORD(&ha->reg->ctrl_status)));
	} else {
		WRT_REG_DWORD(&ha->reg->u1.isp4022.intr_mask, SET_RMASK(IMR_SCSI_INTR_ENABLE));
		PCI_POSTING(&ha->reg->u1.isp4022.intr_mask);
	}

	set_bit(AF_INTERRUPTS_ON, &ha->flags);
}

static inline void __qla4xxx_disable_intrs(scsi_qla_host_t *ha)
{

	clear_bit(AF_INTERRUPTS_ON, &ha->flags);

	if( IS_QLA4010(ha) ) {
		WRT_REG_DWORD(&ha->reg->ctrl_status, CLR_RMASK(CSR_SCSI_INTR_ENABLE));
		PCI_POSTING(&ha->reg->ctrl_status);
		QL4PRINT(QLP7, printk("scsi%d: %s: intSET_RMASK = %08x\n",
			      ha->host_no, __func__,
			      RD_REG_DWORD(&ha->reg->ctrl_status)));
	} else {
		WRT_REG_DWORD(&ha->reg->u1.isp4022.intr_mask, CLR_RMASK(IMR_SCSI_INTR_ENABLE));
		PCI_POSTING(&ha->reg->u1.isp4022.intr_mask);
		QL4PRINT(QLP7, printk("scsi%d: %s: intr_mask = %08x\n",
			      ha->host_no, __func__,
			      RD_REG_DWORD(&ha->reg->u1.isp4022.intr_mask)));
	}
}

static inline void qla4xxx_enable_intrs(scsi_qla_host_t *ha)
{
	unsigned long flags = 0;

	spin_lock_irqsave(&ha->hardware_lock, flags);
	__qla4xxx_enable_intrs(ha);
	spin_unlock_irqrestore(&ha->hardware_lock, flags);
}

static inline void qla4xxx_disable_intrs(scsi_qla_host_t *ha)
{
	unsigned long flags = 0;

	spin_lock_irqsave(&ha->hardware_lock, flags);
	__qla4xxx_disable_intrs(ha);
	spin_unlock_irqrestore(&ha->hardware_lock, flags);
}

static inline int
qla4xxx_is_eh_active(struct Scsi_Host *shost)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,15)
 if (shost->eh_active)
 return 1;
#else
 if (shost->shost_state == SHOST_RECOVERY)
 return 1;
#endif
 return 0;
};

static __inline__ void
qla4xxx_suspend_lun(scsi_qla_host_t *, srb_t *sp, os_lun_t *, int, int);
static __inline__ void
qla4xxx_delay_lun(scsi_qla_host_t *, os_lun_t *, int);

static __inline__ void
qla4xxx_suspend_lun(scsi_qla_host_t *ha, srb_t *sp, os_lun_t *lq, int time, int count)
{
	return (__qla4xxx_suspend_lun(ha, sp, lq, time, count, 0));
}

static __inline__ void
qla4xxx_delay_lun(scsi_qla_host_t *ha, os_lun_t *lq, int time)
{
	return (__qla4xxx_suspend_lun(ha, NULL, lq, time, 1, 1));
}

static inline uint8_t ql4_memzero(const char *m, size_t s)
{
	int i, val = 0;

	if (m) {
		for (i = 0; i < s; i++)
			val |= m[i];
	}
	return (val == 0);
}

/**
 * qla4_8xxx_uevent_emit - Send uevent when the firmware dump is ready.
 * @ha: pointer to adapter structure
 */
static inline void
qla4_8xxx_uevent_emit(struct scsi_qla_host *ha, u32 code)
{
	char event_string[40];
	char *envp[] = { event_string, NULL };

	switch (code) {
	case QL4_UEVENT_CODE_FW_DUMP:
		snprintf(event_string, sizeof(event_string), "FW_DUMP=%d",
			ha->host_no);
		break;
	default:
		/*do nothing*/
		break;
	}

	kobject_uevent_env(&(&ha->pdev->dev)->kobj, KOBJ_CHANGE, envp);
}

static inline void qla4_8xxx_get_minidump(struct scsi_qla_host *ha)
{
	if (ql4xenablemd && test_bit(AF_FW_RECOVERY, &ha->flags) &&
	    !test_bit(AF_82XX_FW_DUMPED, &ha->flags)) {
		if (!qla4_8xxx_collect_md_data(ha)) {
			qla4_8xxx_uevent_emit(ha,
					      QL4_UEVENT_CODE_FW_DUMP);
			set_bit(AF_82XX_FW_DUMPED, &ha->flags);
		} else {
			ql4_printk(KERN_WARNING, ha,
				   "%s: Unable to collect minidump\n",
				   __func__);
		}
	}
}

/*
 * Overrides for Emacs so that we almost follow Linus's tabbing style.
 * Emacs will notice this stuff at the end of the file and automatically
 * adjust the settings for this buffer only.  This must remain at the end
 * of the file.
 * ---------------------------------------------------------------------------
 * Local variables:
 * c-indent-level: 2
 * c-brace-imaginary-offset: 0
 * c-brace-offset: -2
 * c-argdecl-indent: 2
 * c-label-offset: -2
 * c-continued-statement-offset: 2
 * c-continued-brace-offset: 0
 * indent-tabs-mode: nil
 * tab-width: 8
 * End:
 */

