/*
 * QLogic iSCSI HBA Driver
 * Copyright (c)  2003-2012 QLogic Corporation
 *
 * See LICENSE.qla4xxx for copyright and licensing details.
 */
#include "ql4im_def.h"
#include "ql4im_ioctl.h"

extern uint8_t ql4im_capture_dump_image(ql4im_scsi_qla_host_t *ql4im_ha, void *dump_image, uint32_t size);
extern void ql4im_build_dump_image_header(ql4im_scsi_qla_host_t *ql4im_ha, void *dump_image_header);

static void qla4xxx_retrieve_probe_dump(scsi_qla_host_t *ha, probe_dump_t *probe_dump);
static void qla4xxx_retrieve_core_dump(scsi_qla_host_t *ha, core_dump_t *core_dump);

uint8_t
ql4im_capture_dump_image(ql4im_scsi_qla_host_t *ql4im_ha,
			   void *dump_image,
			   uint32_t size)
{
	dump_image_t *image = (dump_image_t *)dump_image;

	if (dump_image == NULL) {
		DEBUG2(printk("scsi%d: %s: NULL dump image address\n",
				      ql4im_ha->host_no, __func__));
		return QLA_ERROR;
	}

	if (size >= DUMP_IMAGE_SIZE) {
		ql4im_build_dump_image_header(ql4im_ha, &image->dump_header);
		qla4xxx_retrieve_core_dump(ql4im_ha->drvr_ha, &image->core_dump);
		qla4xxx_retrieve_probe_dump(ql4im_ha->drvr_ha, &image->probe_dump);
		return QLA_SUCCESS;
	}

	return QLA_ERROR;

}
//
// qla4xxx_build_dump_image_header
//

void
ql4im_build_dump_image_header(ql4im_scsi_qla_host_t *ql4im_ha, void *dump_image_header)
{
	dump_image_header_t *header = (dump_image_header_t *) dump_image_header;

	memset(header, 0, sizeof(*header));

	header->cookie = QLGC_COOKIE;
	sprintf((char *)&header->dump_id_string,"%4x Dump", ql4im_ha->drvr_ha->pdev->device);
	header->time_stamp = get_jiffies_64();
	header->total_image_size  = DUMP_IMAGE_SIZE; // image size excluding header
	header->core_dump_offset  = CORE_DUMP_OFFSET;        //
	header->probe_dump_offset = PROBE_DUMP_OFFSET;
	sprintf((char *)&header->driver,"%s_%d v%s",
		    DRIVER_NAME, ql4im_ha->instance, ql4im_ha->drv_verstr);
	sprintf((char *)&header->ioctlmod,"qisioctl_%d %s",
		    ql4im_ha->instance, QL4IM_VERSION);

	DEBUG10(printk("scsi%d: %s: Dump Image Header:\n",
			ql4im_ha->host_no, __func__));
	DEBUG10(ql4im_dump_bytes((uint8_t *)header, sizeof(*header)));
}

/************************************************************
 *
 *                  Core Dump Routines
 *
 ************************************************************/

/*
 * Perform a Write operation via the MADI registers
 */
static uint8_t qla4_write_MADI(scsi_qla_host_t *ha, int addr, int data)
{
   int done = 0;
   int count = 0;
   unsigned long flags;

   spin_lock_irqsave(&ha->hardware_lock, flags);
   if (((RD_REG_DWORD(&ha->reg->arc_madi_cmd) & MADI_STAT_MASK) >> 27) == MADI_STAT_COMMAND_BUSY)
   {
      DEBUG2(printk("scsi%d: %s: MADI not ready for write operation\n",
			    ha->host_no, __func__));
      spin_unlock_irqrestore(&ha->hardware_lock, flags);
      return QLA_ERROR;
   }

   WRT_REG_DWORD(&ha->reg->arc_madi_cmd, addr);
   WRT_REG_DWORD(&ha->reg->arc_madi_data, data);

   while (!done)
   {
      switch ((RD_REG_DWORD(&ha->reg->arc_madi_cmd) & MADI_STAT_MASK) >> 27)
      {
         case MADI_STAT_DATA_VALID:
            done = 1;
            break;

         case MADI_STAT_DATA_INVALID:
	    WRT_REG_DWORD(&ha->reg->arc_madi_cmd, addr);
	    WRT_REG_DWORD(&ha->reg->arc_madi_data, data);
            break;

         case MADI_STAT_COMMAND_BUSY:
         default:
            if (count++ > 10000)
            {
               DEBUG2(printk("scsi%d: %s: "
			"Waiting too long on write (Address 0x%08x)\n",
			ha->host_no, __func__, addr<<2));
	       spin_unlock_irqrestore(&ha->hardware_lock, flags);
               return QLA_ERROR;
            }
            break;
      }
   }

   spin_unlock_irqrestore(&ha->hardware_lock, flags);
   return QLA_SUCCESS;
}

/*
 * Perform a Read operation via the MADI registers
 */
static uint8_t qla4_read_MADI(scsi_qla_host_t *ha, int addr, uint32_t *data)
{
   int done = 0;
   int count = 0;
   unsigned long flags;

   spin_lock_irqsave(&ha->hardware_lock, flags);
   WRT_REG_DWORD(&ha->reg->arc_madi_cmd, MADI_READ_CMD | addr);
   while (!done)
   {
      switch ((RD_REG_DWORD(&ha->reg->arc_madi_cmd) & MADI_STAT_MASK) >> 27)
      {
         case MADI_STAT_DATA_VALID:
            done = 1;
            break;

         case MADI_STAT_DATA_INVALID:
            //badRead++;
	    WRT_REG_DWORD(&ha->reg->arc_madi_cmd, MADI_READ_CMD | addr);
            break;

         case MADI_STAT_COMMAND_BUSY:
         default:
            if (count++ > 10000)
            {
               DEBUG2(printk("scsi%d: %s: "
			"Waiting too long on read (Address 0x%08x)\n",
			ha->host_no, __func__, addr<<2));
	       spin_unlock_irqrestore(&ha->hardware_lock, flags);
               return QLA_ERROR;
            }
            break;

      }
   }

   *data = RD_REG_DWORD(&ha->reg->arc_madi_data);
   spin_unlock_irqrestore(&ha->hardware_lock, flags);
   return QLA_SUCCESS;
}

static void
qla4xxx_retrieve_core_dump(scsi_qla_host_t *ha, core_dump_t *core_dump)
{
	uint32_t             addr, data, rval;
	volatile uint32_t      __iomem *reg_ptr;
	unsigned long        flags;
	uint8_t		     page, num_pci_pages;

	DEBUG12(printk("Entering %s: core_dump = %p\n", __func__, core_dump));

	if (core_dump == NULL) {
		DEBUG2(printk("scsi%d: %s: No core dump buffer available\n",
			ha->host_no, __func__));
		return;
	}

	/* 1 - Select OAP Processor */
	if (qla4_write_MADI(ha, MADI_DEST_ARC_DEBUG, 0x00000000) != QLA_SUCCESS) {
		DEBUG2(printk("scsi%d: %s: 1 - Select OAP Processor Failed1\n",
			ha->host_no, __func__));
		return;
	}

	/* 2 - Halt the OAP Processor */
	if (qla4_write_MADI(ha, MADI_DEST_AUX_REG | 0x00000005, 0x00000002) != QLA_SUCCESS) {
		DEBUG2(printk("scsi%d: %s: 2 - Halt OAP Processor Failed\n",
			ha->host_no, __func__));
		return;
	}

	/* 3 - Disable SRAM Parity */
	if (qla4_write_MADI(ha, MADI_DEST_AUX_REG | 0x00000020, 0x00000001) != QLA_SUCCESS) {
		DEBUG2(printk("scsi%d: %s: 3 - Disable SRAM Parity Failed\n",
			ha->host_no, __func__));
		return;
	}

	/* 4 - Select IAP Processor */
	if (qla4_write_MADI(ha, MADI_DEST_ARC_DEBUG, 0x00000001) != QLA_SUCCESS) {
		DEBUG2(printk("scsi%d: %s: 4 - Select IAP Processor Failed\n",
			ha->host_no, __func__));
		return;
	}

	/* 5 - Halt IAP Processor */
	if (qla4_write_MADI(ha, MADI_DEST_AUX_REG | 0x00000005, 0x00000002) != QLA_SUCCESS) {
		DEBUG2(printk("scsi%d: %s: 5 - Halt IAP Processor Failed\n",
			ha->host_no, __func__));
		return;
	}

	/* 6 - Select OAP Processor */
	if (qla4_write_MADI(ha, MADI_DEST_ARC_DEBUG, 0x00000000) != QLA_SUCCESS) {
		DEBUG2(printk("scsi%d: %s: 6 - Select OAP Processor Failed\n",
			ha->host_no, __func__));
		return;
	}

 	DEBUG10(printk("scsi%d: %s: NOTE: Output truncated for messages file.  "
 				"Extract full dump with SANsurfer or iscli.\n",
 			ha->host_no, __func__));

	/* 7 - PCI Registers from Processor's Perspective */
	for (addr = (PCI_START >> 2); (addr <= (PCI_END >> 2)) ; addr++) {
		rval = qla4_read_MADI(ha, (addr | MADI_READ_CMD) , &data);
		if (rval != QLA_SUCCESS) {
			DEBUG2(printk("scsi%d: %s: 7 - PCI Registers "
				"from Processor's Perspective Failed\n",
				ha->host_no, __func__));
			DEBUG2(printk("scsi%d: %s: PCIReg 0 addr = 0x%08x Failed0\n",
			ha->host_no, __func__, (addr << 2)));
			return;
		}
		core_dump->PCIRegProc[(addr & (0xFFC >> 2))] = data;
	}
	DEBUG10(printk("scsi%d: %s: 7 - PCI Registers from Processor's Perspective:\n",
			ha->host_no, __func__));
	DEBUG10(ql4im_dump_bytes((uint8_t *)core_dump->PCIRegProc, 32));


	/* 8 - SRAM Content */
	for (addr = (RAM_START >> 2); (addr <= (RAM_END >> 2)) ; addr++) {
		rval = qla4_read_MADI(ha, (addr | MADI_READ_CMD) , &data);
		if (rval != QLA_SUCCESS) {
			DEBUG2(printk("scsi%d: %s: 8 - SRAM Content Failed,  addr = 0x%08x\n",
				ha->host_no, __func__, (addr << 2)));
			return;
		}
		core_dump->SRAM[addr] = data;
	}
	DEBUG10(printk("scsi%d: %s: 8 - SRAM Content:\n", ha->host_no, __func__));
	DEBUG10(ql4im_dump_bytes((uint8_t *)core_dump->SRAM, 32));


	/* 9 - OAP Core Registers */
	for (addr = CORE_START; (addr <= CORE_END) ; addr++) {
		rval = qla4_read_MADI(ha, (addr | MADI_READ_CMD | MADI_DEST_CORE_REG) , &data);
		if (rval != QLA_SUCCESS) {
			DEBUG2(printk("scsi%d: %s: 9 - OAP Core Reg Failed, addr = 0x%08x\n",
                                ha->host_no, __func__, (addr)));
			return;
		}
		core_dump->OAPCoreReg[addr] = data;
	}
	DEBUG10(printk("scsi%d: %s: 9 - OAP Core Register:\n", ha->host_no, __func__));
	DEBUG10(ql4im_dump_bytes((uint8_t *)core_dump->OAPCoreReg, 32));


	/* 10 - OAP Auxiliary Registers */
	for (addr = 0; (addr <= 0x309) ; addr++) {
		rval = qla4_read_MADI(ha, (addr | MADI_READ_CMD | MADI_DEST_AUX_REG) , &data);
		if (rval != QLA_SUCCESS) {
			DEBUG2(printk("scsi%d: %s: 10 - OAP Aux Reg Failed, "
				"addr = 0x%08x\n", ha->host_no, __func__, (addr)));
			return;
		}
		core_dump->OAPAuxReg[addr] = data;
	}
	DEBUG10(printk("scsi%d: %s: 10 - OAP Auxiliary Registers:\n",
		ha->host_no, __func__));
	DEBUG10(ql4im_dump_bytes((uint8_t *)core_dump->OAPAuxReg, 32));


	/* 11 - Select IAP Processor */
	if (qla4_write_MADI(ha, MADI_DEST_ARC_DEBUG, 0x00000001) != QLA_SUCCESS) {
		DEBUG2(printk("scsi%d: %s: 11 - Select IAP Processor Failed\n",
			ha->host_no, __func__));
		return;
	}

	/* 12 - IAP Core Registers */
	for (addr = 0; (addr <= 0x3F) ; addr++) {
		rval = qla4_read_MADI(ha, (addr | MADI_READ_CMD | MADI_DEST_CORE_REG), &data);
		if (rval != QLA_SUCCESS) {
			DEBUG2(printk("scsi%d: %s: 12 - IAP Core Reg Failed, "
				"addr = 0x%08x \n", ha->host_no, __func__, addr));
			return;
		}
		core_dump->IAPCoreReg[addr] = data;
	}
	DEBUG10(printk("scsi%d: %s: 12 - IAP Core Registers:\n",
		ha->host_no, __func__));
	DEBUG10(ql4im_dump_bytes((uint8_t *)core_dump->IAPCoreReg, 32));


	/* 13 - IAP Auxiliary Registers */
	for (addr = 0; (addr <= 0x309) ; addr++) {
		rval = qla4_read_MADI(ha, (addr | MADI_READ_CMD | MADI_DEST_AUX_REG), &data);
		if (rval != QLA_SUCCESS) {
			DEBUG2(printk("scsi%d: %s: 13 - IAP Aux Reg Failed, "
				"addr = 0x%08x\n", ha->host_no, __func__, (addr)));
			return;
		}
		core_dump->IAPAuxReg[addr] = data;
	}
	DEBUG10(printk("scsi%d: %s: 13 - IAP Auxiliary Registers:\n",
		ha->host_no, __func__));
	DEBUG10(ql4im_dump_bytes((uint8_t *)core_dump->IAPAuxReg, 32));


	/* 14 - Save IAP load/store RAM */
	for (addr = (LDST_START >> 2); (addr <= (LDST_END >> 2)) ; addr++) {
		rval = qla4_read_MADI(ha, (addr | MADI_READ_CMD), &data);
		if (rval != QLA_SUCCESS) {
			DEBUG2(printk("scsi%d: %s: 14 - IAP SRAM Content Failed, "
				"addr = 0x%08x\n", ha->host_no, __func__, (addr << 2)));
			return;
		}
		core_dump->IAPSRAM[(addr & (0x1FFC >> 2))] = data;
	}
	DEBUG10(printk("scsi%d: %s: 14 - IAP Load/Store RAM\n",
		ha->host_no, __func__));
	DEBUG10(ql4im_dump_bytes((uint8_t *)core_dump->IAPSRAM, 64));


	/* 15 - Select OAP Processor */
	if (qla4_write_MADI(ha, MADI_DEST_ARC_DEBUG, 0x00000000) != QLA_SUCCESS) {
		DEBUG2(printk("scsi%d: %s: 15 - Select OAP Processor Failed3\n",
			ha->host_no, __func__));
		return;
	}

	/* 16 - Save Host PCI Registers */
	if (IS_QLA4010(ha))
                num_pci_pages = 4;
	else
                num_pci_pages = 3;

	for (page = 0; page < num_pci_pages; page++) {
		spin_lock_irqsave(&ha->hardware_lock, flags);
		WRT_REG_DWORD(&ha->reg->ctrl_status, CLR_RMASK(CSR_SCSI_PAGE_SELECT) | page);
		reg_ptr = &ha->reg->mailbox[0];
		for (addr = 0; addr < 64; addr++) {
			data = RD_REG_DWORD(reg_ptr);
			core_dump->HostPCIRegPage[page][addr] = data;
			reg_ptr++;
		}
		spin_unlock_irqrestore(&ha->hardware_lock, flags);
		DEBUG10(printk("scsi%d: %s: 16 - Host PCI Registers Page %d:\n",
			ha->host_no, __func__, page));
		DEBUG10(ql4im_dump_bytes((uint8_t *)core_dump->HostPCIRegPage[page], 64));
	}

	/* 17 - Save statistics registers */
	if (IS_QLA4010(ha)) {
		/* Statistics registers were saved from page 3 registers above */
	}
	else {
		spin_lock_irqsave(&ha->hardware_lock, flags);
		WRT_REG_DWORD(&ha->reg->ctrl_status, CLR_RMASK(CSR_SCSI_PAGE_SELECT));
		WRT_REG_DWORD(&ha->reg->u2.isp4022.p0.stats_index, 0);
		for (addr = 0; addr < 64; addr++) {
			data = RD_REG_DWORD(&ha->reg->u2.isp4022.p0.stats_read_data_inc);
			core_dump->HostPCIRegPage[PROT_STAT_PAGE][addr] = data;
		}
		DEBUG10(printk("scsi%d: %s: 17 - Statistics Registers:\n",
			ha->host_no, __func__));
		DEBUG10(ql4im_dump_bytes((uint8_t *)core_dump->HostPCIRegPage[page], 64));
		spin_unlock_irqrestore(&ha->hardware_lock, flags);
	}

	/* Select OAP Processor */
	if (qla4_write_MADI(ha, MADI_DEST_ARC_DEBUG, 0x00000000) != QLA_SUCCESS) {
		DEBUG2(printk("scsi%d: %s: 18 - Select OAP Processor Failed\n",
			ha->host_no, __func__));
		return;
	}
	/* Enable SRAM Parity */
	if (qla4_write_MADI(ha, MADI_DEST_AUX_REG | 0x00000020, 0x00000000) != QLA_SUCCESS) {
		DEBUG2(printk("scsi%d: %s: 19 - Enable SRAM Parity Failed\n",
			ha->host_no, __func__));
		return;
	}
}


/************************************************************
 *
 *                  Probe Dump Routines
 *
 ************************************************************/
static uint32_t
probeRead(scsi_qla_host_t *ha, uint32_t probeAddr, probeData_t *probeData)
{
	uint32_t   oldPage;
	unsigned long	flags;
	
	spin_lock_irqsave(&ha->hardware_lock, flags);
	oldPage = RD_REG_DWORD(&ha->reg->ctrl_status) & 0x0003;
	WRT_REG_DWORD(&ha->reg->ctrl_status, 0x00030000);  // Set to page 0
	
	WRT_REG_DWORD(ISP_PROBE_MUX_ADDR(ha), (u_long)(probeAddr | PROBE_RE | PROBE_UP));
	probeData->high = (RD_REG_DWORD(ISP_PROBE_MUX_DATA(ha)) >> 24) & 0xff;
	WRT_REG_DWORD(ISP_PROBE_MUX_ADDR(ha), (u_long)(probeAddr | PROBE_RE | PROBE_LO));
	probeData->low = RD_REG_DWORD(ISP_PROBE_MUX_DATA(ha));
	
	WRT_REG_DWORD(&ha->reg->ctrl_status, (u_long)(0x00030000 | oldPage)); // Reset page
	spin_unlock_irqrestore(&ha->hardware_lock, flags);
	return probeData->low;
}
//
// 4010 ProbeMux table
//
static PROBEMUX_INFO  probeModuleInfo4010[] = {
     {"0"        , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"DA"       , CLK_BIT(SYSCLK) | CLK_BIT(PCICLK)    , MUX_SELECT_MAX}
   , {"NRM"      , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"ODE"      , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"SRM"      , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"SCM"      , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"NCM"      , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"PRD"      , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"SDE"      , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"RBM"      , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"IDE"      , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"TDE"      , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"RA"       , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"ERM"      , CLK_BIT(SYSCLK) | CLK_BIT(CPUCLK)    , MUX_SELECT_MAX}
   , {"RMI"      , CLK_BIT(SYSCLK) | CLK_BIT(CPUCLK)    , MUX_SELECT_MAX}
   , {"OAP"      ,                   CLK_BIT(CPUCLK)    , MUX_SELECT_MAX}
   , {"ECM"      , CLK_BIT(SYSCLK) | CLK_BIT(CPUCLK)    , MUX_SELECT_MAX}
   , {"NPF"      , CLK_BIT(SYSCLK) | CLK_BIT(CPUCLK)    , MUX_SELECT_MAX}
   , {"IAP"      , CLK_BIT(SYSCLK) | CLK_BIT(CPUCLK)    , MUX_SELECT_MAX}
   , {"OTP"      , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"TTM"      , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"ITP"      , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"MAM"      , CLK_BIT(SYSCLK) | CLK_BIT(CPUCLK)    , MUX_SELECT_MAX}
   , {"BLM"      , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"ILM"      , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"IFP"      , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"IPV"      , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"OIP"      , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"OFB"      , CLK_BIT(SYSCLK) | CLK_BIT(NRXCLK)    , MUX_SELECT_MAX}
   , {"MAC"      , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"IFB"      , CLK_BIT(SYSCLK) | CLK_BIT(NRXCLK)    , MUX_SELECT_MAX}
   , {"PCORE"    , CLK_BIT(PCICLK)                      , MUX_SELECT_MAX}
   , {"20"       , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"21"       , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"22"       , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"23"       , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"24"       , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"25"       , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"26"       , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"27"       , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"28"       , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"29"       , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"2A"       , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"2B"       , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"2C"       , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"2D"       , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"2E"       , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"2F"       , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"30"       , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"31"       , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"32"       , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"33"       , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"34"       , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"35"       , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"36"       , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"37"       , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"38"       , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"39"       , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"3a"       , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"3b"       , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"3c"       , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"3d"       , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"3e"       , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"3f"       , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
};

//
// 4022/4032 ProbeMux table
//

static PROBEMUX_INFO  probeModuleInfo4022[] = {
     {"0"        , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"DA"       , CLK_BIT(SYSCLK) | CLK_BIT(PCICLK)    , MUX_SELECT_MAX}
   , {"BPM"      , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"ODE"      , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"SRM0"     , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"SRM1"     , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"PMD"      , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"PRD"      , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"SDE"      , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"RMD"      , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"IDE"      , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"TDE"      , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"RA"       , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"REG"      , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"RMI"      , CLK_BIT(SYSCLK) | CLK_BIT(CPUCLK)    , MUX_SELECT_MAX}
   , {"OAP"      ,                   CLK_BIT(CPUCLK)    , MUX_SELECT_MAX}
   , {"ECM"      , CLK_BIT(SYSCLK) | CLK_BIT(CPUCLK)    , MUX_SELECT_MAX}
   , {"NPF"      , CLK_BIT(SYSCLK) | CLK_BIT(CPUCLK)    , MUX_SELECT_MAX}
   , {"IAP"      , CLK_BIT(SYSCLK) | CLK_BIT(CPUCLK)    , MUX_SELECT_MAX}
   , {"OTP"      , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"TTM"      , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"ITP"      , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"MAM"      , CLK_BIT(SYSCLK) | CLK_BIT(CPUCLK)    , MUX_SELECT_MAX}
   , {"BLM"      , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"ILM"      , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"IFP"      , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"IPV"      , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"OIP"      , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"OFB"      , CLK_BIT(SYSCLK) | CLK_BIT(NRXCLK)    , MUX_SELECT_MAX}
   , {"MAC"      , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"IFB"      , CLK_BIT(SYSCLK) | CLK_BIT(NRXCLK)    , MUX_SELECT_MAX}
   , {"PCORE"    , CLK_BIT(PCICLK)                      , MUX_SELECT_MAX}
   , {"NRM0"     , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"NRM1"     , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"SCM0"     , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"SCM1"     , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"NCM0"     , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"NCM1"     , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"RBM0"     , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"RBM1"     , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"RBM2"     , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"RBM3"     , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"ERM0"     , CLK_BIT(SYSCLK) | CLK_BIT(CPUCLK)    , MUX_SELECT_MAX}
   , {"ERM1"     , CLK_BIT(SYSCLK) | CLK_BIT(CPUCLK)    , MUX_SELECT_MAX}
   , {"PERF0"    , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"PERF1"    , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"2E"       , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"2F"       , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"30"       , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"31"       , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"32"       , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"33"       , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"34"       , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"35"       , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"36"       , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"37"       , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"38"       , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"39"       , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"3a"       , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"3b"       , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"3c"       , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"3d"       , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"3e"       , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
   , {"3f"       , CLK_BIT(SYSCLK)                      , MUX_SELECT_MAX}
};

#if defined(QL_DEBUG_LEVEL_10)
static char  *probeClockNames[] = {
     "SYS"
   , "PCI"
   , "NRX"
   , "CPU"
};
#endif

static void
qla4xxx_retrieve_probe_dump(scsi_qla_host_t *ha, probe_dump_t *probe_dump)
{
	uint32_t   firstModule = probe_DA;
	uint32_t   lastModule  = probe_PERF1;
	uint32_t   probeModule;
	
	uint32_t   firstClock = 0;
	uint32_t   lastClock  = 4;
	uint32_t   probeClock;
	
	uint32_t   lastMux;
	uint32_t   muxSelect;
	
	uint32_t   probeAddr;
	probeData_t probeData[MUX_SELECT_MAX];

	PROBEMUX_INFO  *probeModuleInfo;
	
	
	DEBUG12(printk("qla4xxx: Entering %s(): probe_dump = %p\n",
			       __func__, probe_dump));

	probeModuleInfo = (IS_QLA4010(ha)) ? probeModuleInfo4010 : probeModuleInfo4022;
	
	for (probeModule = firstModule; probeModule <= lastModule; probeModule++) {
		for (probeClock = firstClock; probeClock < lastClock; probeClock++) {
			if ((probeModuleInfo[probeModule].clocks & (1 << probeClock))) {
				lastMux = probeModuleInfo[probeModule].maxSelect;
				probeAddr = (probeModule << 8) | (probeClock << 6);
				DEBUG10(printk("%5s %3s:  ModuleSelect:0x%02x   Clock:0x%x  StartingProbeAddr:0x%04x\n",
					probeModuleInfo[probeModule].moduleName,
					probeClockNames[probeClock],
					probeModule, probeClock,
					probeAddr));
				for (muxSelect = 0; muxSelect < lastMux; muxSelect++) {
					probeRead(ha, (uint32_t)(probeAddr | muxSelect), &probeData[muxSelect]);
				}
				
				#if defined(QL_DEBUG_LEVEL_10)
				{
					for (muxSelect = 0; muxSelect < lastMux; muxSelect+=4) {
					DEBUG10(printk("%5s %3s 0x%04x:  %02x_%08x  %02x_%08x  %02x_%08x  %02x_%08x\n",
						probeModuleInfo[probeModule].moduleName,
						probeClockNames[probeClock],
						probeAddr | muxSelect,
						probeData[muxSelect].high,   probeData[muxSelect].low,
						probeData[muxSelect+1].high, probeData[muxSelect+1].low,
						probeData[muxSelect+2].high, probeData[muxSelect+2].low,
						probeData[muxSelect+3].high, probeData[muxSelect+3].low));
					}
					DEBUG10(printk("\n"));
				}
				#endif
			}
		}
	}
}

