/*
 * QLogic iSCSI IOCTL Module
 * Copyright (c)  2003-2012 QLogic Corporation
 *
 * See LICENSE.qla4xxx for copyright and licensing details.
 */

/******************************************************************************
 *             Please see release.txt for revision history.                   *
 *                                                                            *
 ******************************************************************************
 * Function Table of Contents:
 *	qla4intioctl_logout_iscsi
 *	qla4intioctl_ping
 *	qla4intioctl_get_flash
 *      qla4intioctl_get_core_dump
 *	qla4intioctl_get_host_no
 *	qla4intioctl_get_data
 *	qla4intioctl_set_flash
 *	qla4intioctl_set_data
 *	qla4intioctl_hba_reset
 *	qla4intioctl_copy_fw_flash
 *	qla4xxx_iocb_pass_done
 *	qla4intioctl_iocb_passthru
 * 	qla4intioctl_restore_factory_defaults
 ****************************************************************************/
#include "ql4im_def.h"
#include "ql4im_ioctl.h"

/*
 * Sub codes for Get Data.
 * Use in combination with INT_GET_DATA as the ioctl code
 */
#define INT_SC_GET_HOST_NO 		3


ioctl_tbl_row_t IOCTL_SCMD_IGET_DATA_TBL[] =
{
	{INT_SC_GET_FLASH, "INT_SC_GET_FLASH"},
	{INT_SC_GET_CORE_DUMP, "INT_SC_GET_CORE_DUMP"},
	{INT_SC_GET_HOST_NO, "INT_SC_GET_HOST_NO"},
	{0, "UNKNOWN"}
};

ioctl_tbl_row_t IOCTL_SCMD_ISET_DATA_TBL[] =
{
	{INT_SC_SET_FLASH, "INT_SC_SET_FLASH"},
	{0, "UNKNOWN"}
};

/**************************************************************************
 * qla4intioctl_logout_iscsi
 *	This routine requests that the specified device either login or
 *	logout, depending on the option specified.
 *
 * Input:
 *	ql4im_ha    = adapter structure pointer.
 *	ioctl = IOCTL structure pointer.
 *
 * Output:
 *	None
 *
 * Returns:
 *	QLA_SUCCESS = success
 *	QLA_ERROR   = error
 *
 * Context:
 *	Kernel context.
 **************************************************************************/
int
qla4intioctl_logout_iscsi(ql4im_scsi_qla_host_t *ql4im_ha, EXT_IOCTL_ISCSI *ioctl)
{
	int status = 0;
	INT_LOGOUT_ISCSI logout;

	if (ioctl->RequestLen > sizeof(INT_LOGOUT_ISCSI)) {
		DEBUG2(printk("qisioctl%d: %s: memory area too small\n",
		    ql4im_ha->host_no, __func__));

		ioctl->Status = EXT_STATUS_BUFFER_TOO_SMALL;
		goto exit_logout;
	}

	/* --- Copy logout structure from user space --- */
	if (copy_from_user((void *)&logout,
                                     Q64BIT_TO_PTR(ioctl->RequestAdr,
						   ioctl->AddrMode),
                                     sizeof(INT_LOGOUT_ISCSI)) != 0) {
		DEBUG2(printk("qisioctl%d: %s: unable to copy data from "
		    "user's memory area\n", ql4im_ha->host_no, __func__));

		status = (-EFAULT);
		ioctl->Status = EXT_STATUS_COPY_ERR;
		goto exit_logout;
	}

	/* --- Execute command --- */
	if (logout.Options == INT_DEF_CLOSE_SESSION) {
		if (ql4im_logout_device(ql4im_ha->drvr_ha, logout.TargetID,
		    logout.ConnectionID) == QLA_SUCCESS) {
			DEBUG4(printk("qisioctl%d: %s: CLOSE_SESSION SUCCEEDED!, "
			    "DDB[%d]\n", ql4im_ha->host_no, __func__,
			    logout.TargetID));

			ioctl->Status = EXT_STATUS_OK;
		} else {
			DEBUG2(printk("qisioctl%d: %s: CLOSE_SESSION FAILED!, "
			    "DDB[%d]\n", ql4im_ha->host_no, __func__,
			    logout.TargetID));

			ioctl->Status = EXT_STATUS_ERR;
		}

	} else if (logout.Options == INT_DEF_RELOGIN_CONNECTION) {
		if (ql4im_login_device(ql4im_ha->drvr_ha, logout.TargetID,
		    logout.ConnectionID) == QLA_SUCCESS) {
			DEBUG4(printk("qisioctl%d: %s: RELOGIN_CONNECTION "
			    "SUCCEEDED!, DDB[%d]\n",
			    ql4im_ha->host_no, __func__, logout.TargetID));

			ioctl->Status = EXT_STATUS_OK;
		} else {
			DEBUG2(printk("qisioctl%d: %s: RELOGIN_CONNECTION "
			    "FAILED!, target %d\n",
			    ql4im_ha->host_no, __func__, logout.TargetID));

			ioctl->Status = EXT_STATUS_ERR;
		}

	} else if ((logout.Options == INT_DEF_DELETE_DDB) ||
		   (logout.Options == INT_DEF_REINDEX_DDB)) {
		if (ql4im_delete_device(ql4im_ha->drvr_ha, logout.TargetID,
		    logout.ConnectionID) == QLA_SUCCESS) {
			DEBUG4(printk("qisioctl%d: %s: DELETE_DDB "
			    "SUCCEEDED!, DDB[%d]\n",
			    ql4im_ha->host_no, __func__, logout.TargetID));

			ioctl->Status = EXT_STATUS_OK;
		} else {
			DEBUG2(printk("qisioctl%d: %s: DELETE_DDB FAILED!, "
			    "DDB[%d]\n",
			    ql4im_ha->host_no, __func__, logout.TargetID));

			ioctl->Status = EXT_STATUS_ERR;
		}
		
		if ((logout.NewTargetID != (uint16_t)(-1)) &&
		    (logout.NewTargetID != 0) &&
		    (logout.TargetID != logout.NewTargetID)) {
			if (ql4im_reindex_ddb(ql4im_ha->drvr_ha, logout.TargetID,
						logout.NewTargetID) == QLA_SUCCESS) {
				DEBUG4(printk("scsi%d: %s: REINDEX_DDB "
                                              "SUCCEEDED!, DDB[%d->%d]\n",
                                              ql4im_ha->host_no, __func__,
                                              logout.TargetID, logout.NewTargetID));

				ioctl->Status = EXT_STATUS_OK;
			} else {
				DEBUG2(printk("scsi%d: %s: REINDEX_DDB FAILED!, "
                                              "DDB[%d->%d]\n",
                                              ql4im_ha->host_no, __func__,
                                              logout.TargetID, logout.NewTargetID));

				ioctl->Status = EXT_STATUS_ERR;
			}
		}
	}

exit_logout:
	DEBUG12(printk("qisioctl%d: %s: inst %d exiting.\n",
	    ql4im_ha->host_no, __func__, ql4im_ha->instance));
	return(status);
}

/**************************************************************************
 * qla4intioctl_ping
 *	This routine requests that the HBA PING the specified IP Address.
 *
 * Input:
 *	ql4im_ha    = adapter structure pointer.
 *	ioctl = IOCTL structure pointer.
 *
 * Output:
 *	None
 *
 * Returns:
 *	QLA_SUCCESS = success
 *	QLA_ERROR   = error
 *
 * Context:
 *	Kernel context.
 **************************************************************************/
int
qla4intioctl_ping(ql4im_scsi_qla_host_t *ql4im_ha, EXT_IOCTL_ISCSI *ioctl)
{
	int		status = 0;
	INT_PING	ping;
	uint32_t	mbox_cmd[MBOX_REG_COUNT];
	uint32_t	mbox_sts[MBOX_REG_COUNT];

	DEBUG12(printk("qisioctl%d: %s: inst %d entered.\n",
	    ql4im_ha->host_no, __func__, ql4im_ha->instance));

	/*
	 * Copy user's data to local buffer
	 */
	if (copy_from_user((uint8_t *)&ping,
                                     Q64BIT_TO_PTR(ioctl->RequestAdr,
						   ioctl->AddrMode),
				     sizeof(ping)) != 0) {
		DEBUG2(printk("qisioctl%d: %s: unable to copy data from "
		    "user's memory area\n", ql4im_ha->host_no, __func__));
		status = (-EFAULT);
		ioctl->Status = EXT_STATUS_COPY_ERR;
		goto exit_ping;
	}

	/*
	 * Debug Print Statement
	 */
	DEBUG4(printk("qisioctl%d: %s: %d.%d.%d.%d\n",
	    ql4im_ha->host_no, __func__,
	    ping.IPAddr.IPAddress[0],
	    ping.IPAddr.IPAddress[1],
	    ping.IPAddr.IPAddress[2],
	    ping.IPAddr.IPAddress[3]));
	/*
	 * Issue Mailbox Command
	 */
	memset(mbox_cmd, 0, sizeof(mbox_cmd));
	memset(mbox_sts, 0, sizeof(mbox_sts));
	mbox_cmd[0] = MBOX_CMD_PING;
	mbox_cmd[1] = ping.Reserved;  /* bitmap - no endian conversion */
	memcpy(&mbox_cmd[2], &ping.IPAddr.IPAddress[0], 4);
	memcpy(&mbox_cmd[3], &ping.IPAddr.IPAddress[4], 4);
	memcpy(&mbox_cmd[4], &ping.IPAddr.IPAddress[8], 4);
	memcpy(&mbox_cmd[5], &ping.IPAddr.IPAddress[12], 4);
	mbox_cmd[6] = (uint32_t) cpu_to_le16(ping.PacketSize);

	if (ql4im_ha->drvr_ha->ql4_mbx(ql4im_ha->drvr_ha, 7, 7, &mbox_cmd[0], &mbox_sts[0]) ==
	    QLA_ERROR) {
		DEBUG2(printk("qisioctl%d: %s: command failed \n",
		    ql4im_ha->host_no, __func__));

		ioctl->Status = EXT_STATUS_MAILBOX;
		ioctl->VendorSpecificStatus[0] = mbox_sts[0];
		ioctl->VendorSpecificStatus[1] = mbox_sts[1];
		ioctl->VendorSpecificStatus[2] = mbox_sts[2];
		ioctl->VendorSpecificStatus[3] = mbox_sts[3];
		ioctl->VendorSpecificStatus[4] = mbox_sts[4];
		ioctl->VendorSpecificStatus[5] = mbox_sts[5];
		ioctl->VendorSpecificStatus[6] = mbox_sts[6];
		goto exit_ping;
	}

	ioctl->Status = EXT_STATUS_OK;
exit_ping:

	DEBUG12(printk("qisioctl%d: %s: inst %d exiting.\n",
	    ql4im_ha->host_no, __func__, ql4im_ha->instance));
	return(status);
}

/**************************************************************************
 * qla4intioctl_get_flash
 *	This routine reads the requested area of FLASH.
 *
 * Input:
 *	ql4im_ha    = adapter structure pointer.
 *	ioctl = IOCTL structure pointer.
 *
 * Output:
 *	None
 *
 * Returns:
 *	QLA_SUCCESS = success
 *	QLA_ERROR   = error
 *
 * Context:
 *	Kernel context.
 **************************************************************************/
static int
qla4intioctl_get_flash(ql4im_scsi_qla_host_t *ql4im_ha, EXT_IOCTL_ISCSI *ioctl)
{
	int		status = 0;
	void		*local_dma_bufv = NULL;
	dma_addr_t	local_dma_bufp;
	INT_ACCESS_FLASH *paccess_flash = NULL;
	uint32_t	mbox_cmd[MBOX_REG_COUNT];
	uint32_t	mbox_sts[MBOX_REG_COUNT];


	DEBUG12(printk("qisioctl%d: %s: inst %d entered.\n",
	    ql4im_ha->host_no, __func__, ql4im_ha->instance));

	/*
	 * Allocate local flash buffer
	 */
	if (ql4im_get_ioctl_scrap_mem(ql4im_ha, (void **)&paccess_flash,
	    sizeof(INT_ACCESS_FLASH))) {
		/* not enough memory */
		DEBUG2(printk("qisioctl%d: %s: inst=%d scrap not big enough. "
		    "size requested=%ld.\n",
		    ql4im_ha->host_no, __func__, ql4im_ha->instance,
		    (ulong)sizeof(INT_ACCESS_FLASH)));

		status = (-ENOMEM);
		ioctl->Status = EXT_STATUS_NO_MEMORY;
		ioctl->ResponseLen = 0;
		goto exit_get_flash;
	}

	/*
	 * Copy user's data to local flash buffer
	 */
	if (copy_from_user((uint8_t *)paccess_flash,
                                     Q64BIT_TO_PTR(ioctl->RequestAdr,
						   ioctl->AddrMode),
				     sizeof(INT_ACCESS_FLASH)) != 0) {
		DEBUG2(printk("qisioctl%d: %s: unable to copy data from user's "
		    "memory area\n", ql4im_ha->host_no, __func__));

		status = (-EFAULT);
		ioctl->Status = EXT_STATUS_COPY_ERR;
		goto exit_get_flash;
	}

	/*
	 * Allocate DMA memory
	 */
	local_dma_bufv = pci_alloc_consistent(ql4im_ha->drvr_ha->pdev,
					      paccess_flash->DataLen,
                                              &local_dma_bufp);
	if (local_dma_bufv == NULL) {
		DEBUG2(printk("qisioctl%d: %s: unable to allocate dma memory\n",
		    ql4im_ha->host_no, __func__));

		status = (-ENOMEM);
		ioctl->Status = EXT_STATUS_NO_MEMORY;
		ioctl->ResponseLen = 0;
		goto exit_get_flash;
	}

	DEBUG4(printk("qisioctl%d: %s: offset=%08x, len=%08x\n",
	    ql4im_ha->host_no, __func__,
	    paccess_flash->DataOffset, paccess_flash->DataLen));

	/*
	 * Issue Mailbox Command
	 */
	memset(mbox_cmd, 0, sizeof(mbox_cmd));
	memset(mbox_sts, 0, sizeof(mbox_sts));
	mbox_cmd[0] = MBOX_CMD_READ_FLASH;
	mbox_cmd[1] = LSDW(local_dma_bufp);
	mbox_cmd[2] = MSDW(local_dma_bufp);
	mbox_cmd[3] = paccess_flash->DataOffset;
	mbox_cmd[4] = paccess_flash->DataLen;

	if (ql4im_ha->drvr_ha->ql4_mbx(ql4im_ha->drvr_ha, 5, 2, &mbox_cmd[0], &mbox_sts[0]) ==
	    QLA_ERROR) {
		DEBUG2(printk("qisioctl%d: %s: command failed with status=0x%08X\n",
		    ql4im_ha->host_no, __func__, mbox_sts[0]));

		ioctl->Status = EXT_STATUS_MAILBOX;
		ioctl->VendorSpecificStatus[0] = mbox_sts[0];
		ioctl->VendorSpecificStatus[1] = mbox_sts[1];
		goto exit_get_flash;
	}

	memcpy(&paccess_flash->FlashData[0], local_dma_bufv,
	    MIN(paccess_flash->DataLen, sizeof(paccess_flash->FlashData)));

	/*
	 * Copy local DMA buffer to user's response data area
	 */
	if (copy_to_user(Q64BIT_TO_PTR(ioctl->ResponseAdr,
						 ioctl->AddrMode),
                                   paccess_flash,
				   sizeof(*paccess_flash)) != 0) {
		DEBUG2(printk("qisioctl%d: %s: unable to copy data to user's "
		    "memory area\n", ql4im_ha->host_no, __func__));

		status = (-EFAULT);
		ioctl->Status = EXT_STATUS_COPY_ERR;
		goto exit_get_flash;
	}

	ioctl->Status = EXT_STATUS_OK;
	ioctl->ResponseLen = paccess_flash->DataLen;

	DEBUG10(printk("INT_ACCESS_FLASH buffer (1st 60h bytes only):\n"));
	DEBUG10(ql4im_dump_bytes(paccess_flash, 0x60));

exit_get_flash:

	if (local_dma_bufv)
		pci_free_consistent(ql4im_ha->drvr_ha->pdev,
		    paccess_flash->DataLen, local_dma_bufv, local_dma_bufp);

	ql4im_free_ioctl_scrap_mem(ql4im_ha);

	DEBUG12(printk("qisioctl%d: %s: inst %d exiting.\n",
	    ql4im_ha->host_no, __func__, ql4im_ha->instance));
	return(status);
}

/**************************************************************************
 * qla4intioctl_get_core_dump
 *	This routine reads a block of data for a core dump.
 *
 * Input:
 *	ql4im_ha    = adapter structure pointer.
 *	ioctl = IOCTL structure pointer.
 *
 * Output:
 *	None
 *
 * Returns:
 *	QLA_SUCCESS = success
 *	QLA_ERROR   = error
 *
 * Context:
 *	Kernel context.
 **************************************************************************/
static int
qla4intioctl_get_core_dump(ql4im_scsi_qla_host_t *ql4im_ha, EXT_IOCTL_ISCSI *ioctl)
{
	int		status = 0;
	INT_ACCESS_CORE_DUMP *pcore_dump = NULL;
	uint8_t		*dump_image = ql4im_ha->ioctl->core_dump_mem;
	int		dump_image_size = 0;

	/*
	 * Allocate local core dump buffer
	 */
	if (ql4im_get_ioctl_scrap_mem(ql4im_ha, (void **)&pcore_dump,
	    ioctl->RequestLen)) {
		/* not enough memory */
		DEBUG2(printk("scsi%d: %s: inst=%d scrap not big enough. "
		    "size requested=%d.\n",
		    ql4im_ha->host_no, __func__, ql4im_ha->instance,
		    ioctl->RequestLen));

		status = (-ENOMEM);
		ioctl->Status = EXT_STATUS_NO_MEMORY;
		ioctl->ResponseLen = 0;
		goto exit_get_core_dump;
	}

	/*
	 * Copy user's data to local core dump buffer
	 */
	if (copy_from_user((uint8_t *)pcore_dump,
                           Q64BIT_TO_PTR(ioctl->RequestAdr,
                                         ioctl->AddrMode),
			   ioctl->RequestLen) != 0) {
		DEBUG2(printk("scsi%d: %s: unable to copy data from user's "
		    "memory area\n", ql4im_ha->host_no, __func__));

		status = (-EFAULT);
		ioctl->Status = EXT_STATUS_COPY_ERR;
		goto exit_get_core_dump;
	}

	if (IS_QLA80XX(ql4im_ha->drvr_ha)) {
		if (!test_bit(AF_82XX_FW_DUMPED, &ql4im_ha->drvr_ha->flags)) {
			ioctl->Status = EXT_STATUS_ERR;
			printk("scsi%d: %s: FW Dump not available\n",
					ql4im_ha->host_no, __func__);
			goto exit_get_core_dump;
		}

		if ((pcore_dump->Offset == 0) &&
		    pcore_dump->DataLen == DUMP_IMAGE_HEADER_SIZE) {
			dump_image_size = DUMP_IMAGE_HEADER_SIZE;
		} else {
			dump_image_size = ql4im_ha->drvr_ha->fw_dump_size;

			if (dump_image == NULL) {
				/* should never get here, allocated at init */
				printk("scsi%d: %s: unable to allocate "
					"dump image memory area\n",
					ql4im_ha->host_no, __func__);
				status = (-EFAULT);
				ioctl->Status = EXT_STATUS_ERR;
				goto exit_get_core_dump;
			}

			memset (dump_image, 0, dump_image_size);
			dump_image = ql4im_ha->drvr_ha->fw_dump;

			if (pcore_dump->Offset + pcore_dump->DataLen >= dump_image_size) {
				printk("scsi%d: %s: LastBlock\n",
					ql4im_ha->host_no, __func__);
				pcore_dump->LastBlockFlag = 1;
				ioctl->ResponseLen = dump_image_size - pcore_dump->Offset;
				clear_bit(AF_82XX_FW_DUMPED, &ql4im_ha->drvr_ha->flags);
				memcpy(&pcore_dump->Data[0],
					dump_image + pcore_dump->Offset,
					ioctl->ResponseLen);
				goto exit_last_block;
			}
		}
		DEBUG2(printk("scsi%d: %s: offset=%08x, len=%08x, "
			"max=%08lx\n", ql4im_ha->host_no, __func__,
			pcore_dump->Offset, pcore_dump->DataLen,
			(u_long)dump_image_size));
	} else {
		dump_image_size = DUMP_IMAGE_SIZE;
		DEBUG4(printk("scsi%d: %s: offset=%08x, len=%08x, "
			"max=%08lx/%08lx\n", ql4im_ha->host_no, __func__,
			pcore_dump->Offset, pcore_dump->DataLen,
			(u_long)dump_image_size, (u_long)QLA_IOCTL_DUMP_IMAGE_SIZE));

		if (pcore_dump->Offset + pcore_dump->DataLen > dump_image_size) {
			printk("scsi%d: %s: LastBlock\n",
			ql4im_ha->host_no, __func__);
			pcore_dump->LastBlockFlag = 1;
		}

		if (pcore_dump->Offset + pcore_dump->DataLen > QLA_IOCTL_DUMP_IMAGE_SIZE) {
			DEBUG2(printk("scsi%d: %s: data requested beyond buffer "
				"memory area\n", ql4im_ha->host_no, __func__));
			status = (-EFAULT);
			ioctl->Status = EXT_STATUS_DATA_OVERRUN;
			goto exit_get_core_dump;
		}
	}

	/*
	 * Retrieve Core Dump Data
	 */
	if (pcore_dump->Offset == 0) {
		if (pcore_dump->DataLen == DUMP_IMAGE_HEADER_SIZE) {
			DEBUG4(printk("scsi%d: %s: calling ql4im_build_dump_image_header\n",
			    ql4im_ha->host_no, __func__));
			ql4im_build_dump_image_header(ql4im_ha, &pcore_dump->Data[0]);
		}

		if (pcore_dump->DataLen > DUMP_IMAGE_HEADER_SIZE) {
			/*
			 * Retrieve Core Dump Data
			 */
			if (IS_QLA40XX(ql4im_ha->drvr_ha)) {
				if (dump_image == NULL) {
					// should never get here, allocated at init  ...
					DEBUG2(printk("scsi%d: %s: unable to allocate "
					    "dump image memory area\n",
					    ql4im_ha->host_no, __func__));

					status = (-EFAULT);
					ioctl->Status = EXT_STATUS_ERR;
					goto exit_get_core_dump;
				}
				memset (dump_image, 0, QLA_IOCTL_DUMP_IMAGE_SIZE);

				DEBUG4(printk("scsi%d: %s: calling "
					"ql4im_capture_dump_image\n",
					ql4im_ha->host_no, __func__));
				if (ql4im_capture_dump_image(ql4im_ha, dump_image,
					QLA_IOCTL_DUMP_IMAGE_SIZE) == QLA_ERROR) {
					DEBUG2(printk("scsi%d: %s: command failed \n",
					    ql4im_ha->host_no, __func__));

					ioctl->Status = EXT_STATUS_ERR;
					goto exit_get_core_dump;
				}
			}
		}
	}

	if (pcore_dump->DataLen > DUMP_IMAGE_HEADER_SIZE && dump_image) {
		memcpy(&pcore_dump->Data[0],
		       dump_image + pcore_dump->Offset,
		       pcore_dump->DataLen);
	}

exit_last_block:
	/*
	 * Copy local DMA buffer to user's response data area
	 */
	if (copy_to_user(Q64BIT_TO_PTR(ioctl->ResponseAdr,
                                       ioctl->AddrMode),
                         pcore_dump, ioctl->ResponseLen) != 0) {
		DEBUG2(printk("scsi%d: %s: unable to copy data to user's "
		    "memory area\n", ql4im_ha->host_no, __func__));

		status = (-EFAULT);
		ioctl->Status = EXT_STATUS_COPY_ERR;
		goto exit_get_core_dump;
	}

	ioctl->Status = EXT_STATUS_OK;

	if (!pcore_dump->LastBlockFlag)
		ioctl->ResponseLen = pcore_dump->DataLen;

exit_get_core_dump:
	ql4im_free_ioctl_scrap_mem(ql4im_ha);
	return(status);
}

/**
 * qla4intioctl_update_beacon - Enable/Disable beacon
 * @ql4im_ha : Pointer to host adapter structure.
 * @ioctl    : IOCTL structure from application.
 * @mbox_cmd : Outgoing mailbox reg. from application.
 * @mbox_sts : Mailbox status to send back.
 *
 * This routine enables or disables beacon for specified host,
 * based on the mailbox command sent by application.
 **/
void qla4intioctl_update_beacon(ql4im_scsi_qla_host_t *ql4im_ha,
                        EXT_IOCTL_ISCSI *ioctl, uint32_t *mbox_cmd,
			uint32_t *mbox_sts)
{
        ioctl->Status = EXT_STATUS_OK;

        if (ql4im_ha->drvr_ha->ql4_mbx(ql4im_ha->drvr_ha, MBOX_REG_COUNT, 1,
				&mbox_cmd[0], &mbox_sts[0]) == QLA_ERROR) {
                DEBUG2(ql4_info(ha, "%s: MBOX_CMD_UPDATE_BEACON, failed w/ "
                                        "status %04X", __func__, mbox_sts[0]));
                ioctl->Status = EXT_STATUS_ERR;
                ioctl->DetailStatus = mbox_sts[0];
        }
}

/**************************************************************************
 * qla4intioctl_get_host_no
 *	This routine retrieves the host number for the specified adapter
 *	instance.
 *
 * Input:
 *	ql4im_ha    = adapter structure pointer.
 *	ioctl = IOCTL structure pointer.
 *
 * Output:
 *	None
 *
 * Returns:
 *	QLA_SUCCESS = success
 *	QLA_ERROR   = error
 *
 * Context:
 *	Kernel context.
 **************************************************************************/
static int
qla4intioctl_get_host_no(ql4im_scsi_qla_host_t *ql4im_ha, EXT_IOCTL_ISCSI *ioctl)
{
	int status = 0;


	DEBUG12(printk("qisioctl%d: %s: inst %d entered.\n",
	    ql4im_ha->host_no, __func__, ql4im_ha->instance));

	if (copy_to_user(Q64BIT_TO_PTR(ioctl->ResponseAdr,
						 ioctl->AddrMode),
                                   &(ql4im_ha->host_no),
				   sizeof(ql4im_ha->host_no)) != 0) {
		DEBUG2(printk("qisioctl%d: %s: failed to copy data\n",
		    ql4im_ha->host_no, __func__));

		status = (-EFAULT);
		ioctl->Status = EXT_STATUS_COPY_ERR;
	} else {
		ioctl->Status = EXT_STATUS_OK;
	}

	DEBUG12(printk("qisioctl%d: %s: inst %d exiting.\n",
	    ql4im_ha->host_no, __func__, ql4im_ha->instance));
	return(status);
}

/**************************************************************************
 * qla4intioctl_get_data
 *	This routine calls get data IOCTLs based on the IOCTL Sub Code.
 *
 * Input:
 *	ql4im_ha    = adapter structure pointer.
 *	ioctl = IOCTL structure pointer.
 *
 * Output:
 *	None
 *
 * Returns:
 *	QLA_SUCCESS = success
 *	QLA_ERROR   = error
 *    	-EINVAL     = if the command is invalid
 *
 * Context:
 *	Kernel context.
 **************************************************************************/
int
qla4intioctl_get_data(ql4im_scsi_qla_host_t *ql4im_ha, EXT_IOCTL_ISCSI *ioctl)
{
	switch (ioctl->SubCode) {
	case INT_SC_GET_FLASH:
		return(qla4intioctl_get_flash(ql4im_ha, ioctl));
        case INT_SC_GET_CORE_DUMP:
                return(qla4intioctl_get_core_dump(ql4im_ha, ioctl));
	case INT_SC_GET_HOST_NO:
		return(qla4intioctl_get_host_no(ql4im_ha, ioctl));
	default:
		DEBUG2(printk("qisioctl%d: %s: unsupported internal get data "
		    "sub-command code (%X)\n",
		    ql4im_ha->host_no, __func__, ioctl->SubCode));

		ioctl->Status = EXT_STATUS_UNSUPPORTED_SUBCODE;
		return (-EINVAL);
	}
}

/**************************************************************************
 * qla4intioctl_set_flash
 *	This routine writes the requested area of FLASH.
 *
 * Input:
 *	ql4im_ha    = adapter structure pointer.
 *	ioctl = IOCTL structure pointer.
 *
 * Output:
 *	None
 *
 * Returns:
 *	QLA_SUCCESS = success
 *	QLA_ERROR   = error
 *
 * Context:
 *	Kernel context.
 **************************************************************************/
static int
qla4intioctl_set_flash(ql4im_scsi_qla_host_t *ql4im_ha, EXT_IOCTL_ISCSI *ioctl)
{
	int		status = 0;
	INT_ACCESS_FLASH *paccess_flash = NULL;
	uint32_t	mbox_cmd[MBOX_REG_COUNT];
	uint32_t	mbox_sts[MBOX_REG_COUNT];


	DEBUG12(printk("qisioctl%d: %s: inst %d entered.\n",
	    ql4im_ha->host_no, __func__, ql4im_ha->instance));

	/*
	 * Allocate local flash buffer
	 */
	if (ql4im_get_ioctl_scrap_mem(ql4im_ha, (void **)&paccess_flash,
	    sizeof(INT_ACCESS_FLASH))) {
		/* not enough memory */
		DEBUG2(printk("qisioctl%d: %s: inst=%d scrap not big enough. "
		    "size requested=%ld.\n",
		    ql4im_ha->host_no, __func__, ql4im_ha->instance,
		    (ulong)sizeof(INT_ACCESS_FLASH)));

		status = (-ENOMEM);
		ioctl->Status = EXT_STATUS_NO_MEMORY;
		ioctl->ResponseLen = 0;
		goto exit_set_flash;
	}

	/*
	 * Copy user's data to local DMA buffer
	 */
	if (copy_from_user((uint8_t *)paccess_flash,
                                     Q64BIT_TO_PTR(ioctl->RequestAdr,
						   ioctl->AddrMode),
				     sizeof(INT_ACCESS_FLASH)) != 0) {
		DEBUG2(printk("qisioctl%d: %s: unable to copy data from user's "
		    "memory area\n", ql4im_ha->host_no, __func__));

		status = (-EFAULT);
		ioctl->Status = EXT_STATUS_COPY_ERR;
		goto exit_set_flash;
	}

	/*
	 * Resize IOCTL DMA memory, if necesary
	 */
	if ((paccess_flash->DataLen != 0) &&
	    (ql4im_ha->ioctl_dma_buf_len < paccess_flash->DataLen)) {
		if (ql4im_resize_ioctl_dma_buf(ql4im_ha, paccess_flash->DataLen) !=
		    QLA_SUCCESS) {
			DEBUG2(printk("qisioctl%d: %s: unable to allocate memory "
			    "for dma buffer.\n",
			    ql4im_ha->host_no, __func__));

			status = (-ENOMEM);
			ioctl->Status = EXT_STATUS_NO_MEMORY;
			ioctl->ResponseLen = 0;
			goto exit_set_flash;
		}
	}

	memcpy(ql4im_ha->ioctl_dma_bufv, &paccess_flash->FlashData[0],
	    MIN(ql4im_ha->ioctl_dma_buf_len, sizeof(paccess_flash->FlashData)));

	DEBUG4(printk("qisioctl%d: %s: offset=%08x, len=%08x\n",
	    ql4im_ha->host_no, __func__,
	    paccess_flash->DataOffset, paccess_flash->DataLen));

	/*
	 * Issue Mailbox Command
	 */
	memset(mbox_cmd, 0, sizeof(mbox_cmd));
	memset(mbox_sts, 0, sizeof(mbox_sts));
	mbox_cmd[0] = MBOX_CMD_WRITE_FLASH;
	mbox_cmd[1] = LSDW(ql4im_ha->ioctl_dma_bufp);
	mbox_cmd[2] = MSDW(ql4im_ha->ioctl_dma_bufp);
	mbox_cmd[3] = paccess_flash->DataOffset;
	mbox_cmd[4] = paccess_flash->DataLen;
	mbox_cmd[5] = paccess_flash->Options;

	if (ql4im_ha->drvr_ha->ql4_mbx(ql4im_ha->drvr_ha, 6, 2,
				  &mbox_cmd[0], &mbox_sts[0]) == QLA_ERROR) {
		DEBUG2(printk("qisioctl%d: %s: command failed \n",
		    ql4im_ha->host_no, __func__));

		ioctl->Status = EXT_STATUS_MAILBOX;
		ioctl->VendorSpecificStatus[0] = mbox_sts[0];
		ioctl->VendorSpecificStatus[1] = mbox_sts[1];
		goto exit_set_flash;
	}

	ioctl->Status = EXT_STATUS_OK;
	ioctl->ResponseLen = paccess_flash->DataLen;
	DEBUG10(printk("qisioctl%d: INT_ACCESS_FLASH buffer (1st 60h bytes only:\n",
	    ql4im_ha->host_no));
	DEBUG10(ql4im_dump_bytes(ql4im_ha->ioctl_dma_bufv, 0x60));

exit_set_flash:
	/*
	 * Free Memory
	 */
	ql4im_free_ioctl_scrap_mem(ql4im_ha);

	DEBUG12(printk("qisioctl%d: %s: inst %d exiting.\n",
	    ql4im_ha->host_no, __func__, ql4im_ha->instance));
	return(status);
}

/**************************************************************************
 * qla4intioctl_set_data
 *	This routine calls set data IOCTLs based on the IOCTL Sub Code.
 *
 * Input:
 *	ql4im_ha    = adapter structure pointer.
 *	ioctl = IOCTL structure pointer.
 *
 * Output:
 *	None
 *
 * Returns:
 *	QLA_SUCCESS = success
 *	QLA_ERROR   = error
 *    	-EINVAL     = if the command is invalid
 *
 * Context:
 *	Kernel context.
 **************************************************************************/
int
qla4intioctl_set_data(ql4im_scsi_qla_host_t *ql4im_ha, EXT_IOCTL_ISCSI *ioctl)
{
	switch (ioctl->SubCode) {
	case INT_SC_SET_FLASH:
		return(qla4intioctl_set_flash(ql4im_ha, ioctl));
	default:
		DEBUG2(printk("qisioctl%d: %s: unsupported internal set data "
				"sub-command code (%X)\n",
				ql4im_ha->host_no, __func__, ioctl->SubCode));

		ioctl->Status = EXT_STATUS_UNSUPPORTED_SUBCODE;
		return (-EINVAL);
	}
}

/**************************************************************************
 * qla4intioctl_hba_reset
 *	This routine resets the specified HBA.
 *
 * Input:
 *	ql4im_ha    = adapter structure pointer.
 *	ioctl = IOCTL structure pointer.
 *
 * Output:
 *	None
 *
 * Returns:
 *	QLA_SUCCESS = success
 *	QLA_ERROR   = error
 *
 * Context:
 *	Kernel context.
 **************************************************************************/
int
qla4intioctl_hba_reset(ql4im_scsi_qla_host_t *ql4im_ha, EXT_IOCTL_ISCSI *ioctl)
{
	uint8_t         status = 0;
	u_long          wait_time;
	int             dpc_wait_bit;
	INT_HBA_RESET   hba_reset = { 0 };
	scsi_qla_host_t *ha = ql4im_ha->drvr_ha;


	DEBUG12(printk("qisioctl%d: %s: inst %d entered.\n",
	    ql4im_ha->host_no, __func__, ql4im_ha->instance));

	if ((status = copy_from_user((uint8_t *)&hba_reset,
				    Q64BIT_TO_PTR(ioctl->RequestAdr,
						  ioctl->AddrMode),
				    sizeof (INT_HBA_RESET))) != 0) {
	       DEBUG2(printk("qisioctl%d: %s: unable to copy data from user's "
		   "memory area\n", ql4im_ha->host_no, __func__));

	       ioctl->Status = EXT_STATUS_COPY_ERR;
	       goto exit_hba_reset;
	}

	if (hba_reset.ResetFlag == HBA_RESET_FUNCTION_RESET) {
	       set_bit(DPC_RESET_HA_FW_CONTEXT, &ha->dpc_flags);
	       dpc_wait_bit = DPC_RESET_HA_FW_CONTEXT;
	} else if (hba_reset.ResetFlag == HBA_RESET_CHIP_RESET) {
	       set_bit(DPC_RESET_HA, &ha->dpc_flags);
	       dpc_wait_bit = DPC_RESET_HA;
	} else {
	       DEBUG2(printk("qisioctl%d: %s: invalid reset request (%d)\n",
		       ql4im_ha->host_no, __func__, hba_reset.ResetFlag));
	       ioctl->Status = EXT_STATUS_INVALID_REQUEST;
	       goto exit_hba_reset;
	}

	wait_time = jiffies + (180 * HZ);

	while (!time_after_eq(jiffies, wait_time)) {
	       ssleep(1);

	       if (!test_bit(dpc_wait_bit, &ha->dpc_flags) &&
		   test_bit(AF_ONLINE, &ha->flags)) {
		       msleep(30);
		       ioctl->Status = EXT_STATUS_OK;
		       goto exit_hba_reset;
	       }
	}

	ioctl->Status = EXT_STATUS_ERR;
	status= QLA_ERROR;

exit_hba_reset:
	DEBUG12(printk("qisioctl%d: %s: inst %d exiting.\n",
	       ql4im_ha->host_no, __func__, ql4im_ha->instance));
	return(status);
}

/**************************************************************************
 * qla4intioctl_copy_fw_flash
 *	This routine requests copying the FW image in FLASH from primary-to-
 *	secondary or secondary-to-primary.
 *
 * Input:
 *	ql4im_ha    = adapter structure pointer.
 *	ioctl = IOCTL structure pointer.
 *
 * Output:
 *	None
 *
 * Returns:
 *	QLA_SUCCESS = success
 *	QLA_ERROR   = error
 *
 * Context:
 *	Kernel context.
 **************************************************************************/
int
qla4intioctl_copy_fw_flash(ql4im_scsi_qla_host_t *ql4im_ha, EXT_IOCTL_ISCSI *ioctl)
{
	int			status = 0;
	INT_COPY_FW_FLASH	copy_flash;
	uint32_t		mbox_cmd[MBOX_REG_COUNT];
	uint32_t		mbox_sts[MBOX_REG_COUNT];


	DEBUG12(printk("qisioctl%d: %s: inst %d entered.\n",
	    ql4im_ha->host_no, __func__, ql4im_ha->instance));

	if (copy_from_user((uint8_t *)&copy_flash,
                                     Q64BIT_TO_PTR(ioctl->RequestAdr,
						   ioctl->AddrMode),
				     ioctl->RequestLen) != 0) {
		DEBUG2(printk("qisioctl%d: %s: unable to copy data from user's "
		    "memory area\n", ql4im_ha->host_no, __func__));

		status = (-EFAULT);
		ioctl->Status = EXT_STATUS_COPY_ERR;
		goto exit_copy_flash;
	}

	memset(mbox_cmd, 0, sizeof(mbox_cmd));
	memset(mbox_sts, 0, sizeof(mbox_sts));
	mbox_cmd[0] = MBOX_CMD_COPY_FLASH;
	mbox_cmd[1] = copy_flash.Options;

	if (ql4im_ha->drvr_ha->ql4_mbx(ql4im_ha->drvr_ha, 2, 2,
				  &mbox_cmd[0], &mbox_sts[0]) == QLA_SUCCESS) {
		DEBUG4(printk("qisioctl%d: %s: Succeeded\n",
		    ql4im_ha->host_no, __func__));

		ioctl->Status = EXT_STATUS_OK;
	} else {
		DEBUG4(printk("qisioctl%d: %s: FAILED\n",
		    ql4im_ha->host_no, __func__));

		ioctl->Status = EXT_STATUS_MAILBOX;
		ioctl->VendorSpecificStatus[0] = mbox_sts[0];
		ioctl->VendorSpecificStatus[1] = mbox_sts[1];
	}

exit_copy_flash:
	DEBUG12(printk("qisioctl%d: %s: inst %d exiting.\n",
	    ql4im_ha->host_no, __func__, ql4im_ha->instance));
	return(status);
}

/**************************************************************************
 * ql4im_iocb_pass_done
 *	This routine resets the ioctl progress flag and wakes up the ioctl
 * 	completion semaphore.
 *
 * Input:
 *	ql4im_ha    = adapter structure pointer.
 *   	sts_entry - pointer to passthru status buffer.
 *
 * Output:
 *	None
 *
 * Returns:
 *	None
 *
 * Context:
 *	Interrupt context.
 **************************************************************************/
void
ql4im_iocb_pass_done(ql4im_scsi_qla_host_t *ql4im_ha, PASSTHRU_STATUS_ENTRY *sts_entry)
{
	INT_IOCB_PASSTHRU *iocb;

	DEBUG12(printk("qisioctl%d: %s: inst %d entered.\n",
	    ql4im_ha->host_no, __func__, ql4im_ha->instance));

	/* --- Copy passthru status buffer to iocb passthru buffer ---*/
	iocb = (INT_IOCB_PASSTHRU *)(ulong)le32_to_cpu(sts_entry->handle);
	memcpy(iocb->IOCBStatusBuffer, sts_entry,
	    MIN(sizeof(iocb->IOCBStatusBuffer), sizeof(*sts_entry)));

	/* --- Reset IOCTL flags and wakeup semaphore.
	 *     But first check to see if IOCTL has already
	 *     timed out because we don't want to get the
	 *     up/down semaphore counters off.             --- */
	if (ql4im_ha->ioctl->ioctl_iocb_pass_in_progress == 1) {
		ql4im_ha->ioctl->ioctl_iocb_pass_in_progress = 0;
		ql4im_ha->ioctl->ioctl_tov = 0;

		DEBUG4(printk("qisioctl%d: %s: UP count=%d\n",
			      ql4im_ha->host_no, __func__,
                              atomic_read(&ql4im_ha->ioctl->ioctl_cmpl_sem.count)));
		up(&ql4im_ha->ioctl->ioctl_cmpl_sem);
	}

	DEBUG12(printk("qisioctl%d: %s: inst %d exiting.\n",
	    ql4im_ha->host_no, __func__, ql4im_ha->instance));
	return;
}

/**************************************************************************
 * qla4intioctl_iocb_passthru
 *	This routine
 *	
 *
 * Input:
 *	ql4im_ha    = adapter structure pointer.
 *	ioctl = IOCTL structure pointer.
 *
 * Output:
 *	None
 *
 * Returns:
 *	QLA_SUCCESS = success
 *	QLA_ERROR   = error
 *
 * Context:
 *	Kernel context.
 **************************************************************************/
int
qla4intioctl_iocb_passthru(ql4im_scsi_qla_host_t *ql4im_ha, EXT_IOCTL_ISCSI *ioctl)
{
	int			status = 0;
	INT_IOCB_PASSTHRU	*iocb;
	INT_IOCB_PASSTHRU	*iocb_dma;
	PASSTHRU0_ENTRY		*passthru_entry;
	unsigned long		flags;
	DATA_SEG_A64		*data_seg;
	scsi_qla_host_t   	*drvr_ha = ql4im_ha->drvr_ha;


	DEBUG12(printk("qisioctl%d: %s:\n", ql4im_ha->host_no, __func__));

	/* --- Use internal DMA buffer for iocb structure --- */

	if (ql4im_ha->ioctl_dma_buf_len < sizeof(*iocb))
		ql4im_resize_ioctl_dma_buf(ql4im_ha, sizeof(*iocb));

	if (!ql4im_ha->ioctl_dma_bufv || ql4im_ha->ioctl_dma_buf_len < sizeof(*iocb)) {
		DEBUG2(printk("qisioctl%d: %s: dma buffer inaccessible.\n",
		    ql4im_ha->host_no, __func__));
		ioctl->Status = EXT_STATUS_BUFFER_TOO_SMALL;
		goto exit_iocb_passthru;
	}

	iocb = (INT_IOCB_PASSTHRU *) ql4im_ha->ioctl_dma_bufv;
	iocb_dma = (INT_IOCB_PASSTHRU *)(unsigned long)ql4im_ha->ioctl_dma_bufp;

	/* --- Copy IOCB_PASSTHRU structure from user space --- */
	if (copy_from_user((uint8_t *)iocb,
                                     Q64BIT_TO_PTR(ioctl->RequestAdr,
						   ioctl->AddrMode),
				     ioctl->RequestLen) != 0) {
		DEBUG2(printk("qisioctl%d: %s: unable to copy data from user's "
		    "memory area\n", ql4im_ha->host_no, __func__));

		status = (-EFAULT);
		ioctl->Status = EXT_STATUS_COPY_ERR;
		goto exit_iocb_passthru;
	}

	if ((iocb->IOCBCmdBuffer[0x00] == 0x3A) &&
	    (iocb->IOCBCmdBuffer[0x0A] == 0x10) &&
	    (iocb->IOCBCmdBuffer[0x0B] == 0x80) &&
	    (iocb->SendData[0x0C] == 0x81) &&
	    (iocb->SendData[0x0D] == 0x4F)) {
		// ok to process command, proceed ...
	} else {
		DEBUG2(printk("qisioctl%d: %s: unable to process command. "
                                      "Did not pass secure I/O boundary check.\n",
                                      ql4im_ha->host_no, __func__));
		DEBUG2(printk("IOCBCmdBuffer[0x00] = 0x%x, expecting 0x3A\n",
			iocb->IOCBCmdBuffer[0x00]));
		DEBUG2(printk("IOCBCmdBuffer[0x0A] = 0x%x, expecting 0x10\n",
			iocb->IOCBCmdBuffer[0x0A]));
		DEBUG2(printk("IOCBCmdBuffer[0x0B] = 0x%x, expecting 0x80\n",
			iocb->IOCBCmdBuffer[0x0B]));
		DEBUG2(printk("SendData[0x0C] = 0x%x, expecting 0x81\n",
			iocb->SendData[0x0C]));
		DEBUG2(printk("SendData[0x0D] = 0x%x, expecting 0x4F\n",
			iocb->SendData[0x0D]));
		status = (-EFAULT);
		ioctl->Status = EXT_STATUS_INVALID_PARAM;
		goto exit_iocb_passthru;
	}

	/* --- Get pointer to the passthru queue entry --- */
	spin_lock_irqsave(&drvr_ha->hardware_lock, flags);
	if (ql4im_get_req_pkt(drvr_ha, (QUEUE_ENTRY **)
			      &passthru_entry) != QLA_SUCCESS) {
		DEBUG2(printk("qisioctl%d: %s: request queue full, "
			"try again later\n",
			ql4im_ha->host_no, __func__));

		ioctl->Status = EXT_STATUS_HBA_QUEUE_FULL;
		spin_unlock_irqrestore(&drvr_ha->hardware_lock, flags);
		goto exit_iocb_passthru;
	}

	/* --- Fill in passthru queue entry --- */
	if (iocb->SendDMAOffset) {
		data_seg = (DATA_SEG_A64 *)(iocb->IOCBCmdBuffer +
		    iocb->SendDMAOffset);
		data_seg->base.addrHigh =
		    cpu_to_le32(MSDW((ulong)&iocb_dma->SendData[0]));
		data_seg->base.addrLow =
		    cpu_to_le32(LSDW((ulong)&iocb_dma->SendData[0]));
	}

	if (iocb->RspDMAOffset) {
		data_seg =
		    (DATA_SEG_A64 *)(iocb->IOCBCmdBuffer + iocb->RspDMAOffset);
		data_seg->base.addrHigh =
		    cpu_to_le32(MSDW((ulong)&iocb_dma->RspData[0]));
		data_seg->base.addrLow =
		    cpu_to_le32(LSDW((ulong)&iocb_dma->RspData[0]));
	}

	memcpy(passthru_entry, iocb->IOCBCmdBuffer,
	    MIN(sizeof(*passthru_entry), sizeof(iocb->IOCBCmdBuffer)));
	passthru_entry->handle = (uint32_t) (unsigned long) iocb;
	passthru_entry->hdr.systemDefined = SD_PASSTHRU_IOCB;

	if (passthru_entry->hdr.entryType != ET_PASSTHRU0)
		passthru_entry->timeout = MBOX_TOV;

	DEBUG4(printk("qisioctl%d: Passthru0 IOCB type %x count %x In (%x) %p\n",
	    ql4im_ha->host_no, passthru_entry->hdr.entryType,
	    passthru_entry->hdr.entryCount, drvr_ha->request_in, passthru_entry));

	DEBUG4(printk("qisioctl%d: Dump Passthru entry %p: \n",
	    ql4im_ha->host_no, passthru_entry));
	DEBUG4(ql4im_dump_bytes(passthru_entry, sizeof(*passthru_entry)));

	/* ---- Prepare for receiving completion ---- */
	ql4im_ha->ioctl->ioctl_iocb_pass_in_progress = 1;
	ql4im_ha->ioctl->ioctl_tov = passthru_entry->timeout * HZ;
	ql4im_ioctl_sem_init(ql4im_ha);

	/* ---- Send command to adapter ---- */
	ql4im_ha->ioctl->ioctl_cmpl_timer.expires = jiffies + ql4im_ha->ioctl->ioctl_tov;
	add_timer(&ql4im_ha->ioctl->ioctl_cmpl_timer);

	WRT_REG_DWORD(&drvr_ha->reg->req_q_in, drvr_ha->request_in);
	PCI_POSTING(&drvr_ha->reg->req_q_in);
	spin_unlock_irqrestore(&drvr_ha->hardware_lock, flags);

	DEBUG4(printk("qisioctl%d: %s: DOWN count=%d\n",
                      drvr_ha->host_no, __func__,
		      atomic_read(&ql4im_ha->ioctl->ioctl_cmpl_sem.count)));

	down(&ql4im_ha->ioctl->ioctl_cmpl_sem);

	/*******************************************************
	 *						       *
	 *             Passthru Completion                     *
	 *						       *
	 *******************************************************/
	del_timer(&ql4im_ha->ioctl->ioctl_cmpl_timer);

	/* ---- Check for timeout --- */
	if (ql4im_ha->ioctl->ioctl_iocb_pass_in_progress == 1) {
		DEBUG2(printk("qisioctl%d: %s: ERROR = command timeout.\n",
		    ql4im_ha->host_no, __func__));

		ql4im_ha->ioctl->ioctl_iocb_pass_in_progress = 0;
		ioctl->Status = EXT_STATUS_ERR;
		goto exit_iocb_passthru;
	}

	/* ---- Copy IOCB Passthru structure with updated status buffer
	 *      to user space ---- */
	if (copy_to_user(Q64BIT_TO_PTR(ioctl->ResponseAdr,
						 ioctl->AddrMode),
                                   iocb, sizeof(INT_IOCB_PASSTHRU)) != 0) {
		DEBUG2(printk("qisioctl%d: %s: unable to copy passthru struct "
		    "to user's memory area.\n",
		    ql4im_ha->host_no, __func__));

		status = (-EFAULT);
		ioctl->Status = EXT_STATUS_COPY_ERR;
		goto exit_iocb_passthru;
	}

	DEBUG4(printk("Dump iocb structure (OUT)\n"));
	DEBUG4(ql4im_print_iocb_passthru(drvr_ha, iocb));

	DEBUG4(printk("qisioctl%d: %s: Succeeded\n",
	    ql4im_ha->host_no, __func__));

	ioctl->Status = EXT_STATUS_OK;

exit_iocb_passthru:
	DEBUG12(printk("qisioctl%d: %s: inst %d exiting.\n",
	    ql4im_ha->host_no, __func__, ql4im_ha->instance));
	return(status);
}

/**************************************************************************
 * qla4intioctl_restore_factory_defaults
 *	This routine restores factory defaults of the adapter.
 *
 * Input:
 *	ql4im_ha    = adapter structure pointer.
 *	ioctl = IOCTL structure pointer.
 *
 * Output:
 *	None
 *
 * Returns:
 *	QLA_SUCCESS = success
 *	QLA_ERROR   = error
 *
 * Context:
 *	Kernel context.
 **************************************************************************/
int
qla4intioctl_restore_factory_defaults(ql4im_scsi_qla_host_t *ql4im_ha, EXT_IOCTL_ISCSI *ioctl)
{
	int status = 0;
	INT_RESTORE_FACTORY_DEFAULTS defaults;

	if (ioctl->RequestLen > sizeof(INT_RESTORE_FACTORY_DEFAULTS)) {
		DEBUG2(printk("qisioctl%d: %s: memory area too small\n",
		    ql4im_ha->host_no, __func__));

		status = (-ENOMEM);
		ioctl->Status = EXT_STATUS_BUFFER_TOO_SMALL;
		ioctl->ResponseLen = 0;
		goto exit_defaults;
	}

	/* --- Copy logout structure from user space --- */
	if (copy_from_user((void *)&defaults,
                                     Q64BIT_TO_PTR(ioctl->RequestAdr,
						   ioctl->AddrMode),
				     sizeof(INT_RESTORE_FACTORY_DEFAULTS)) != 0) {
		DEBUG2(printk("qisioctl%d: %s: unable to copy data from "
		    "user's memory area\n", ql4im_ha->host_no, __func__));

		status = (-EFAULT);
		ioctl->Status = EXT_STATUS_COPY_ERR;
		ioctl->ResponseLen = 0;
		goto exit_defaults;
	}

	/* --- Execute command --- */
	if (ql4im_mbx_restore_factory_defaults(ql4im_ha->drvr_ha, defaults.BlockMask,
	    defaults.IFCBMask1, defaults.IFCBMask2) == QLA_SUCCESS) {
		DEBUG4(printk("qisioctl%d: %s: RESTORE_FACTORY_DEFAULTS SUCCEEDED!\n",
			   ql4im_ha->host_no, __func__));

		ioctl->Status = EXT_STATUS_OK;
	} else {
		DEBUG2(printk("qisioctl%d: %s: RESTORE_FACTORY_DEFAULTS FAILED\n",
		    ql4im_ha->host_no, __func__));

		ioctl->Status = EXT_STATUS_ERR;
		ioctl->ResponseLen = 0;
	}

exit_defaults:
	return(status);
}



/*
 * Overrides for Emacs so that we almost follow Linus's tabbing style.
 * Emacs will notice this stuff at the end of the file and automatically
 * adjust the settings for this buffer only.  This must remain at the end
 * of the file.
 * ---------------------------------------------------------------------------
 * Local variables:
 * c-indent-level: 2
 * c-brace-imaginary-offset: 0
 * c-brace-offset: -2
 * c-argdecl-indent: 2
 * c-label-offset: -2
 * c-continued-statement-offset: 2
 * c-continued-brace-offset: 0
 * indent-tabs-mode: nil
 * tab-width: 8
 * End:
 */


