/*
 * QLogic iSCSI IOCTL Module
 * Copyright (c)  2003-2012 QLogic Corporation
 *
 * See LICENSE.qla4xxx for copyright and licensing details.
 */

/******************************************************************************
 *             Please see release.txt for revision history.                   *
 *                                                                            *
 ******************************************************************************
 * Function Table of Contents:
 *      ql4im_get_req_pkt
 *      ql4im_send_marker_iocb
 ****************************************************************************/

#include "ql4im_def.h"
#include <scsi/scsi_tcq.h>

/**************************************************************************
 * ql4im_get_req_pkt
 *	This routine performs the following tasks:
 *	- returns the current request_in pointer (if queue not full)
 *	- advances the request_in pointer
 *	- checks for queue full
 *
 * Input:
 * 	ha - Pointer to host adapter structure.
 *	queue_entry - Pointer to pointer to queue entry structure
 *
 * Output:
 *	queue_entry - Return pointer to next available request packet
 *
 * Returns:
 *	QLA_SUCCESS - Successfully retrieved request packet
 *	QLA_ERROR   - Failed to retrieve request packet
 *
 * Context:
 *	Kernel context.
 **************************************************************************/
uint8_t
ql4im_get_req_pkt(scsi_qla_host_t *ha, QUEUE_ENTRY **queue_entry)
{
	uint8_t   status = QLA_ERROR;
	uint16_t        cnt;
	uint16_t        req_cnt = 1;
	uint32_t        timer;

	*queue_entry = NULL;

	/* Wait 1 second for slot. */
	for (timer = HZ; timer; timer--) {
		if ((req_cnt + 2) >= ha->req_q_count) {
			/* Calculate number of free request entries. */
			cnt = (uint16_t) le32_to_cpu(ha->shadow_regs->req_q_out);
			if (ha->request_in < cnt) {
				ha->req_q_count = cnt - ha->request_in;
			}
			else {
				ha->req_q_count = ha->maxcmds -
					(ha->request_in - cnt);
			}
		}

		if ((req_cnt + 2) < ha->req_q_count) {
		/* Check if room for request in request ring. */
			*queue_entry = ha->request_ptr;
			memset(*queue_entry, 0, sizeof(**queue_entry));

			/* Advance request queue pointer */
			ha->request_in++;
			if (ha->request_in == ha->maxcmds) {
				ha->request_in = 0;
				ha->request_ptr = ha->request_ring;
			} else {
				ha->request_ptr++;
			}
			ha->req_q_count -= req_cnt;
			status = QLA_SUCCESS;
			break;
		}

		/* Release ring specific lock */
		spin_unlock(&ha->hardware_lock);
		udelay(2);   /* 2 us */
		spin_lock_irq(&ha->hardware_lock);
	}

	if (status == QLA_ERROR) {
		DEBUG2(printk("scsi%d: %s: request queue is full, "
		    "iocb_cnt=%d, req_q_count=%d\n", ha->host_no, __func__,
		    ha->iocb_cnt, ha->req_q_count));
	}
	return (status);
}

/**************************************************************************
 * ql4im_send_marker_iocb
 *	This routine issues a marker IOCB.
 *
 * Input:
 * 	ha - Pointer to host adapter structure.
 *	ddb_entry - Pointer to device database entry
 *	lun - SCSI LUN
 *	marker_type - marker identifier
 *
 * Returns:
 *	QLA_SUCCESS - Successfully sent marker IOCB
 *	QLA_ERROR   - Failed to send marker IOCB
 *
 * Context:
 *	Kernel context.
 **************************************************************************/
static uint8_t
ql4im_send_marker_iocb(scsi_qla_host_t *ha, ddb_entry_t *ddb_entry,
    fc_lun_t *lun_entry)
{
	MARKER_ENTRY *marker_entry;
	unsigned long flags = 0;
	uint8_t       status = QLA_SUCCESS;

	/* Acquire hardware specific lock */
	spin_lock_irqsave(&ha->hardware_lock, flags);

	/* Get pointer to the queue entry for the marker */
	if (ql4im_get_req_pkt(ha, (QUEUE_ENTRY **) &marker_entry)
	    != QLA_SUCCESS) {
		DEBUG2(printk("qisioctl%d: %s: request queue full, try "
		    "again later\n", ha->host_no, __func__));

		status = QLA_ERROR;
		goto exit_send_marker;
	}

	/* Put the marker in the request queue */
	marker_entry->hdr.entryType = ET_MARKER;
	marker_entry->hdr.entryCount = 1;
	marker_entry->target = cpu_to_le16(ddb_entry->fw_ddb_index);
	marker_entry->modifier = cpu_to_le16(MM_LUN_RESET);
	marker_entry->lun[1] = LSB(lun_entry->lun);	 /*SAMII compliant lun*/
	marker_entry->lun[2] = MSB(lun_entry->lun);
	wmb();

	DEBUG3(printk("qisioctl%d:%d:%d:%d: LUN_RESET Marker sent\n", ha->host_no,
	    ddb_entry->bus, ddb_entry->target, lun_entry->lun));

	/* Tell ISP it's got a new I/O request */
	WRT_REG_DWORD(&ha->reg->req_q_in, ha->request_in);
	PCI_POSTING(&ha->reg->req_q_in);

exit_send_marker:
	spin_unlock_irqrestore(&ha->hardware_lock, flags);

	return (status);
}

static CONTINUE_ENTRY *
ql4im_alloc_cont_entry(scsi_qla_host_t *ha)
{
        CONTINUE_ENTRY *cont_entry;

        cont_entry = (CONTINUE_ENTRY *)ha->request_ptr;

        /* Advance request queue pointer */
        if (ha->request_in == (ha->maxcmds - 1)) {
                ha->request_in = 0;
                ha->request_ptr = ha->request_ring;
                DEBUG10(printk("qisioctl%d: %s: wraparound -- new "
                    "request_in = %04x, new request_ptr = %p\n", ha->host_no,
                    __func__, ha->request_in, ha->request_ptr));
        } else {
		ha->request_in++;
                ha->request_ptr++;
                DEBUG10(printk("qisioctl%d: %s: new request_in = %04x, new "
                    "request_ptr = %p\n", ha->host_no, __func__, ha->request_in,
                    ha->request_ptr));
        }

        /* Load packet defaults */
        cont_entry->hdr.entryType = ET_CONTINUE;
        cont_entry->hdr.entryCount = 1;
        cont_entry->hdr.systemDefined =
        (uint8_t) cpu_to_le16(ha->request_in);

        return(cont_entry);
}

static void
ql4im_build_scsi_iocbs(srb_t *srb, COMMAND_ENTRY *cmd_entry, uint16_t tot_dsds)
{
	scsi_qla_host_t	*ha;
	uint16_t	avail_dsds;
	DATA_SEG_A64 *cur_dsd;
	struct scsi_cmnd *cmd;

	cmd = srb->cmd;
	ha = srb->ha;

	if (cmd->request_bufflen == 0 ||
	    cmd->sc_data_direction == DMA_NONE) {
		/* No data being transferred */
		DEBUG5( printk("qisioctl%d:%d:%d:%d: %s: No data xfer\n",
		    ha->host_no, cmd->device->channel, cmd->device->id,
		    cmd->device->lun, __func__));

		cmd_entry->ttlByteCnt = __constant_cpu_to_le32(0);
		return;
	}

	avail_dsds = COMMAND_SEG;
	cur_dsd = (DATA_SEG_A64 *) &(cmd_entry->dataseg[0]);

	/* Load data segments */
	if (cmd->use_sg) {
		struct  scatterlist *cur_seg;
		struct  scatterlist *end_seg;

		/* Data transfer with Scatter/Gather
		 *
		 * We must build an SG list in adapter format, as the kernel's
		 * SG list cannot be used directly because of data field size
		 * (__alpha__) differences and the kernel SG list uses virtual
		 * addresses where we need physical addresses.
		 */
		cur_seg = (struct scatterlist *) cmd->request_buffer;
		end_seg = cur_seg + tot_dsds;

		while (cur_seg < end_seg) {
			dma_addr_t sle_dma;

			/* Allocate additional continuation packets? */
			if (avail_dsds == 0) {
				CONTINUE_ENTRY *cont_entry;

				cont_entry = ql4im_alloc_cont_entry(ha);
				cur_dsd = (DATA_SEG_A64 *) &cont_entry->dataseg[0];
				avail_dsds = CONTINUE_SEG;
			}

			sle_dma = sg_dma_address(cur_seg);
			cur_dsd->base.addrLow = cpu_to_le32(LSDW(sle_dma));
			cur_dsd->base.addrHigh = cpu_to_le32(MSDW(sle_dma));
			cur_dsd->count = cpu_to_le32(sg_dma_len(cur_seg));
			avail_dsds--;

			DEBUG(printk("qisioctl%d:%d:%d:%d: %s: S/G "
			    "DSD %p phys_addr=%x:%08x, len=0x%x, tot_dsd=0x%x, "
			    "avail_dsd=0x%x\n", ha->host_no,
			    cmd->device->channel, cmd->device->id,
			    cmd->device->lun, __func__, cur_dsd,
			    cur_dsd->base.addrHigh, cur_dsd->base.addrLow,
			    cur_dsd->count, tot_dsds, avail_dsds));

			cur_dsd++;
			cur_seg++;
	}
	} else {
		/* Data transfer without SG entries. */
		dma_addr_t	req_dma;
		struct page	*page;
		unsigned long	offset;

		page = virt_to_page(cmd->request_buffer);
		offset = ((unsigned long) cmd->request_buffer & ~PAGE_MASK);
		req_dma = pci_map_page(ha->pdev, page, offset,
		    cmd->request_bufflen, cmd->sc_data_direction);
		srb->saved_dma_handle = req_dma;

		cur_dsd->base.addrLow = cpu_to_le32(LSDW(req_dma));
		cur_dsd->base.addrHigh = cpu_to_le32(MSDW(req_dma));
		cur_dsd->count = cpu_to_le32(cmd->request_bufflen);

		DEBUG5( printk("qisioctl%d:%d:%d:%d: %s: No S/G transfer, "
		    "DSD=%p cmd=%p dma_addr=%x:%08x, len=%x, tot_dsd=0x%x, "
		    "avail_dsd=0x%x\n", ha->host_no, cmd->device->channel,
		    cmd->device->id, cmd->device->lun, __func__, cur_dsd, cmd,
		    cur_dsd->base.addrHigh, cur_dsd->base.addrLow,
		    cur_dsd->count, tot_dsds, avail_dsds));

		cur_dsd++;
	}
}

static uint16_t
ql4im_calc_request_entries(uint16_t dsds)
{
	uint16_t iocbs;/* number of request queue entries */
				/* (commamd + continue) */
	iocbs = 1;
	if (dsds > COMMAND_SEG) {
		iocbs += (dsds - COMMAND_SEG) / CONTINUE_SEG;
		if ((dsds - COMMAND_SEG) % CONTINUE_SEG)
			iocbs++;
	}
	return (iocbs);
}


/**************************************************************************
 * ql4im_send_command_to_isp
 *      This routine is called  to build an ISP
 *      command and pass it to the ISP for execution.
 *
 * Input:
 *      ha - Pointer to host adapter structure.
 *      srb - pointer to SCSI Request Block to be sent to ISP
 *
 * Output:
 *      None
 *
 * Remarks:
 *      None
 *
 * Returns:
 *      QLA_SUCCESS - Successfully sent command to ISP
 *      QLA_ERROR   - Failed to send command to ISP
 *
 * Context:
 *      Kernel context.
 **************************************************************************/
uint8_t
ql4im_send_command_to_isp(scsi_qla_host_t *os_ha, srb_t *srb)
{
	struct scsi_cmnd     *cmd = srb->cmd;
	ddb_entry_t   *ddb_entry;
	os_lun_t   *lun_entry;
	COMMAND_ENTRY *cmd_entry;
	struct scatterlist *sg = NULL;

	uint16_t      tot_dsds;	/* number of data segments */
				/* (sg entries, if sg request) */
	uint16_t	req_cnt;	/* number of request queue entries */

	unsigned long flags;
	uint16_t	cnt;
	uint16_t      i;
	uint32_t      index;
	fc_lun_t                *fclun;
	scsi_qla_host_t *ha;
	char		tag[2];
	uint32_t	timeout;

	/* Get real lun and adapter */
	fclun = srb->lun_queue->fclun;
	ha = fclun->fcport->ha;

	cmd = srb->cmd;
	ddb_entry = fclun->fcport->ddbptr;
	lun_entry = srb->lun_queue;

	/* Send marker(s) if needed. */
	if (ha->marker_needed == 1) {
		if (ql4im_send_marker_iocb(ha, ddb_entry, fclun) !=
		    QLA_SUCCESS) {
			return(QLA_ERROR);
		}
		ha->marker_needed = 0;
	}

	/* Acquire hardware specific lock */
	spin_lock_irqsave(&ha->hardware_lock, flags);

	/* Check for room in active srb array */
	index = ha->current_active_index;
	for (i = 0; i < MAX_SRBS; i++) {
		index++;
		if (index == MAX_SRBS)
			index = 1;
		if (ha->active_srb_array[index] == 0) {
			ha->current_active_index = index;
			break;
		}
	}
	if (i >= MAX_SRBS) {
		DEBUG5(printk("qisioctl%d: %s: NO more SRB entries"
		    "used iocbs=%d, \n reqs remaining=%d\n", ha->host_no,
		    __func__, ha->iocb_cnt, ha->req_q_count));
		goto queuing_error;
	}


	/* Calculate the number of request entries needed. */
	tot_dsds = 0;
	if (cmd->use_sg) {
		sg = (struct scatterlist *) cmd->request_buffer;
		tot_dsds = pci_map_sg(ha->pdev, sg, cmd->use_sg,
		    cmd->sc_data_direction);
		if (tot_dsds == 0)
			goto queuing_error;
	} else if (cmd->request_bufflen) {
	    tot_dsds++;
	}
	req_cnt = ql4im_calc_request_entries(tot_dsds);

	if (ha->req_q_count < (req_cnt + 2)) {
		// cnt = RD_REG_WORD_RELAXED(ISP_REQ_Q_OUT(ha));
	    	cnt = (uint16_t) le32_to_cpu(ha->shadow_regs->req_q_out);
		if (ha->request_in < cnt) {
			ha->req_q_count = cnt - ha->request_in;
		}
		else {
			ha->req_q_count = ha->maxcmds - (ha->request_in - cnt);
		}

		DEBUG5(printk("qisioctl%d: %s: new request queue count -"
		    "calc req_q_count=%d, reqs needed=%d, shadow reqs=%d, req_in =%d\n", ha->host_no,
		    __func__, ha->req_q_count, req_cnt, cnt, ha->request_in));
	}

	if (ha->req_q_count < (req_cnt + 2)) {
		/* free the resources */
		if (cmd->use_sg)
			pci_unmap_sg(ha->pdev, sg, cmd->use_sg, cmd->sc_data_direction);

		DEBUG5(printk("qisioctl%d: %s: No more request queues, "
		    "used iocbs=%d, \n reqs remaining=%d, reqs needed=%d\n", ha->host_no,
		    __func__, ha->iocb_cnt, ha->req_q_count, req_cnt));
		goto queuing_error;
	}

	/* total iocbs active */
	if ((ha->iocb_cnt + req_cnt)  >= ha->maxcmds ) {
		if (cmd->use_sg)
			pci_unmap_sg(ha->pdev, sg, cmd->use_sg, cmd->sc_data_direction);
		goto queuing_error;
	}

	/* Build command packet */
	cmd_entry = (COMMAND_ENTRY *) ha->request_ptr;
	memset(cmd_entry, 0, sizeof(COMMAND_ENTRY));
	cmd_entry->hdr.entryType = ET_COMMAND;
	cmd_entry->handle = cpu_to_le32(index);
	cmd_entry->target = cpu_to_le16(ddb_entry->fw_ddb_index);
	cmd_entry->connection_id = cpu_to_le16(ddb_entry->connection_id);
	
	cmd_entry->lun[1] = LSB(cmd->device->lun);	/* SAMII compliant. */
	cmd_entry->lun[2] = MSB(cmd->device->lun);
	cmd_entry->cmdSeqNum = cpu_to_le32(ddb_entry->CmdSn);
	cmd_entry->ttlByteCnt = cpu_to_le32(cmd->request_bufflen);
	memcpy(cmd_entry->cdb, cmd->cmnd, MIN(MAX_COMMAND_SIZE, cmd->cmd_len));
	cmd_entry->dataSegCnt = cpu_to_le16(tot_dsds);
	cmd_entry->hdr.entryCount = req_cnt;

	/* We want the firmware to time out the command first.*/
	timeout = (uint32_t)(cmd->timeout_per_command / HZ);
	if (timeout > 65535)
                cmd_entry->timeout = __constant_cpu_to_le16(0);
	else if (timeout > MIN_CMD_TOV)
		cmd_entry->timeout = cpu_to_le16((uint16_t) timeout -
					 (QLA_CMD_TIMER_DELTA+1));
	else
		cmd_entry->timeout = cpu_to_le16((uint16_t)timeout);

	srb->iocb_tov = le16_to_cpu(cmd_entry->timeout);
	srb->os_tov = timeout;

	/* Set data transfer direction control flags
	 * NOTE: Look at data_direction bits iff there is data to be
	 *       transferred, as the data direction bit is sometimed filled
	 *       in when there is no data to be transferred */
	cmd_entry->control_flags = CF_NO_DATA;
	if (cmd->request_bufflen) {
		if (cmd->sc_data_direction == DMA_TO_DEVICE)
			cmd_entry->control_flags = CF_WRITE;
		else if (cmd->sc_data_direction == DMA_FROM_DEVICE)
			cmd_entry->control_flags = CF_READ;
	}

	/* Set tagged queueing control flags */
	cmd_entry->control_flags |= CF_SIMPLE_TAG;
	if (scsi_populate_tag_msg(cmd, tag)) {
		switch (tag[0]) {
		case MSG_HEAD_TAG:
			cmd_entry->control_flags |= CF_HEAD_TAG;
			break;
		case MSG_ORDERED_TAG:
			cmd_entry->control_flags |= CF_ORDERED_TAG;
			break;
		}
	}

	/* Advance request queue pointer */
	ha->request_in++;
        if (ha->request_in == ha->maxcmds ) {
                ha->request_in = 0;
                ha->request_ptr = ha->request_ring;
        } else {
                ha->request_ptr++;
        }

	ql4im_build_scsi_iocbs(srb, cmd_entry, tot_dsds);

	wmb();

	/*
	 * Check to see if adapter is online before placing request on
	 * request queue.  If a reset occurs and a request is in the queue,
	 * the firmware will still attempt to process the request, retrieving
	 * garbage for pointers.
	 */
	if (!test_bit(AF_ONLINE, &ha->flags)) {
		DEBUG2(printk("qisioctl%d: %s: Adapter OFFLINE! "
				      "Do not issue command.\n",
                                      ha->host_no, __func__));
		goto queuing_error;
	}

	/* put command in active array */
	ha->active_srb_array[index] = srb;
	srb->cmd->host_scribble = (unsigned char *)(unsigned long)index;

	/* update counters */
	ha->active_srb_count++;
	lun_entry->out_count++;
	lun_entry->tot_io_count++;
	srb->state = SRB_ACTIVE_STATE;
	srb->flags |= SRB_DMA_VALID;

	/* Track IOCB used */
	ha->iocb_cnt += req_cnt;
	srb->entry_count = req_cnt;
	ha->req_q_count -= req_cnt;

	/* Debug print statements */
#ifdef QL_DEBUG_LEVEL_3
	printk("qisioctl%d:%d:%d:%d: %s: CDB = ", ha->host_no,
	    cmd->device->channel, cmd->device->id, cmd->device->lun,
	    __func__);
	for (i = 0; i < cmd->cmd_len; i++)
		printk("%02x ", cmd->cmnd[i]);
	printk("\n");
#endif
	srb->u_start = jiffies;

	/* Tell ISP it's got a new I/O request */
	ha->isp_ops->queue_iocb(ha);
/*	if (IS_QLA8032(ha)) {
		ha->qla4_83xx_queue_iocb(ha);
	} else if(IS_QLA8022(ha)) {
		ha->qla4_82xx_queue_iocb(ha);
	} else {
	       WRT_REG_DWORD(&ha->reg->req_q_in, ha->request_in);
	       PCI_POSTING(&ha->reg->req_q_in);
	}
*/
	spin_unlock_irqrestore(&ha->hardware_lock, flags);

	return(QLA_SUCCESS);

queuing_error:
	/* Release hardware specific lock */
	spin_unlock_irqrestore(&ha->hardware_lock, flags);
	return(QLA_ERROR);
}

/*
 * Overrides for Emacs so that we almost follow Linus's tabbing style.
 * Emacs will notice this stuff at the end of the file and automatically
 * adjust the settings for this buffer only.  This must remain at the end
 * of the file.
 * ---------------------------------------------------------------------------
 * Local variables:
 * c-indent-level: 2
 * c-brace-imaginary-offset: 0
 * c-brace-offset: -2
 * c-argdecl-indent: 2
 * c-label-offset: -2
 * c-continued-statement-offset: 2
 * c-continued-brace-offset: 0
 * indent-tabs-mode: nil
 * tab-width: 8
 * End:
 */

