/*
 * QLogic iSCSI IOCTL Module
 * Copyright (c)  2003-2012 QLogic Corporation
 *
 * See LICENSE.qla4xxx for copyright and licensing details.
 */

/******************************************************************************
 *             Please see release.txt for revision history.                   *
 *                                                                            *
 ******************************************************************************
 * Function Table of Contents:
 *	ql4im_mailbox_command
 *	ql4im_mbx_mbx_test
 *	ql4im_mbx_send_noop
 *	ql4im_mbx_conn_close_sess_logout
 *	ql4im_mbx_clear_database_entry
 *	ql4im_mbx_initialize_fw_cb
 *	ql4im_mbx_get_fwddb_entry
 *	ql4im_mbx_set_ddb_entry
 *	ql4im_mbx_reset_lun
 *	ql4im_mbx_isns_enable
 *	ql4im_mbx_isns_disable
 *	ql4im_mbx_isns_status
 *	ql4im_mbx_get_flash
 ****************************************************************************/

#include "ql4im_def.h"

#include <linux/delay.h>

/**
 * ql4im_qla4_82xx_interrupt_service_routine - isr
 * @ha: pointer to host adapter structure.
 *
 * This is the main interrupt service routine.
 * hardware_lock locked upon entry. runs in interrupt context.
 **/
void ql4im_qla4_82xx_interrupt_service_routine(struct scsi_qla_host * ha,
                                               uint32_t intr_status)
{
        if (intr_status & HSRX_RISC_MB_INT) {
               uint32_t mbox_status = RD_REG_DWORD(&ha->qla4_82xx_reg->mailbox_out[0]);

               if ((mbox_status == MBOX_STS_BUSY) ||
                    (mbox_status == MBOX_STS_INTERMEDIATE_COMPLETION) ||
                    (mbox_status >>12 == MBOX_COMPLETION_STATUS)) {
                        ha->mbox_status[0] = mbox_status;

                        if (test_bit(AF_MBOX_COMMAND, &ha->flags)) {
                                /*
                                 * Copy all mailbox registers to a temporary
                                 * location and set mailbox command done flag
                                 */
                                uint8_t i;

                                for (i = 0; i < ha->mbox_status_count; i++) {
                                        ha->mbox_status[i] = RD_REG_DWORD(&ha->qla4_82xx_reg->mailbox_out[i]);
                                }

                                DEBUG10(printk("qisioctl%d: %s: mailbox cmd done!\n",
                                                ha->host_no, __func__));

                                ha->f_end = jiffies;
                                set_bit(AF_MBOX_COMMAND_DONE, &ha->flags);

				if (test_bit(AF_MBOX_COMMAND_NOPOLL, &ha->flags))
					complete(&ha->mbx_intr_comp);
					
                                wake_up(&ha->mailbox_wait_queue);
                        }
               }
        }

	/* clear the interrupt */
        writel(0, &ha->qla4_82xx_reg->host_int);
        readl(&ha->qla4_82xx_reg->host_int);
}

/**************************************************************************
 * ql4im_interrupt_service_routine
 *	This routine services the interrupt
 *
 * Input:
 * 	ha - Pointer to host adapter structure.
 *
 * Remarks:
 *      hardware_lock locked upon entry
 *
 * Returns:
 *      QLA_SUCCESS - success, An interrupt was found and processed
 *	QLA_ERROR - failure, The adapter was not interrupting
 *
 * Context:
 *	Interrupt context.
 **************************************************************************/
static void
ql4im_interrupt_service_routine(scsi_qla_host_t *ha, uint32_t intr_status)
{
	/*
	 * Process mailbox interrupt.
	 * Asynch events are processed in the driver.
	 */
	if (intr_status & CSR_SCSI_PROCESSOR_INTR) {
		uint32_t mbox_status = RD_REG_DWORD(&ha->reg->mailbox[0]);

		if ((mbox_status == MBOX_STS_BUSY) ||
		    (mbox_status == MBOX_STS_INTERMEDIATE_COMPLETION) ||
		    (mbox_status >>12 == MBOX_COMPLETION_STATUS)) {
			ha->mbox_status[0] = mbox_status;

			if (test_bit(AF_MBOX_COMMAND, &ha->flags)) {
				/*
				 * Copy all mailbox registers to a temporary
				 * location and set mailbox command done flag
				 */
				uint8_t i;

				for (i = 1; i < ha->mbox_status_count; i++) {
					ha->mbox_status[i] =
					    RD_REG_DWORD(&ha->reg->mailbox[i]);
				}

				DEBUG10(printk("qisioctl%d: %s: mailbox cmd done!\n",
						ha->host_no, __func__));

				ha->f_end = jiffies;
				set_bit(AF_MBOX_COMMAND_DONE, &ha->flags);

				if (test_bit(AF_MBOX_COMMAND_NOPOLL, &ha->flags))
					complete(&ha->mbx_intr_comp);

				wake_up(&ha->mailbox_wait_queue);
			}

			/* Clear Mailbox Interrupt */
			WRT_REG_DWORD(&ha->reg->ctrl_status,
			    SET_RMASK(CSR_SCSI_PROCESSOR_INTR));
			PCI_POSTING(&ha->reg->ctrl_status);
		}
	}
}

/**************************************************************************
 * ql4im_mailbox_command
 *	This routine sssue mailbox commands and waits for completion.
 *
 * Input:
 * 	ha - Pointer to host adapter structure.
 *	inCount	 - number of mailbox registers to load.
 *      outCount - number of mailbox registers to return.
 *      mbx_cmd  - data pointer for mailbox in registers.
 *      mbx_sts  - data pointer for mailbox out registers.
 *
 * Output:
 *      mbx_sts - returned mailbox out data.
 *
 * Remarks:
 *	If outCount is 0, this routine completes successfully WITHOUT waiting
 *	for the mailbox command to complete.
 *
 * Returns:
 *	QLA_SUCCESS - Mailbox command completed successfully
 *	QLA_ERROR   - Mailbox command competed in error.
 *
 * Context:
 *	Kernel context.
 **************************************************************************/
uint8_t
ql4im_mailbox_command(scsi_qla_host_t *ha,
			uint8_t inCount,
			uint8_t outCount,
			uint32_t *mbx_cmd,
			uint32_t *mbx_sts)
{
	uint8_t      status = QLA_ERROR;
	uint8_t      i;
	uint8_t      force_polling = 0;
	u_long     wait_count;
	uint32_t     intr_status;
	unsigned long flags = 0;
	DECLARE_WAITQUEUE(wait, current);

	down(&ha->mbox_sem);
	set_bit(AF_MBOX_COMMAND, &ha->flags);

	if (IS_QLA8022(ha) &&
	    test_bit(AF_FW_RECOVERY, &ha->flags)) {
		DEBUG2(printk("qisioctl%d: %s: prematurely completing mbx cmd as "
				"firmware hang detected\n",
				ha->host_no, __func__));
		goto mbox_exit;
	}

	/* Make sure that pointers are valid */
	if (!mbx_cmd || !mbx_sts) {
		DEBUG2(printk("qisioctl%d: %s: Invalid mbx_cmd or mbx_sts pointer\n",
				ha->host_no, __func__));

		goto mbox_exit;
	}

	/* To prevent overwriting mailbox registers for a command that has
	 * not yet been serviced, check to see if an active command
	 * (AEN, IOCB, etc.) is interrupting, then service it.
	 * -----------------------------------------------------------------
	 */
	spin_lock_irqsave(&ha->hardware_lock, flags);

	if (!IS_QLA8022(ha)) {
		intr_status = RD_REG_DWORD(&ha->reg->ctrl_status);
		DEBUG5(printk("qisioctl%d: %s: INTR_STATUS = 0x%X\n",
				ha->host_no, __func__, intr_status));
		if (intr_status & CSR_SCSI_PROCESSOR_INTR) {
			DEBUG2(printk("qisioctl%d: %s: Trying to execute a "
					"mailbox request, while another one "
					"is interrupting. Service existing "
					"interrupt first\n",
					ha->host_no, __func__));
			/* Service existing interrupt first */
			ql4im_interrupt_service_routine(ha, intr_status);
			clear_bit(AF_MBOX_COMMAND_DONE, &ha->flags);
		}
	}

	/* Send the mailbox command to the firmware
	 */
	ha->f_start = jiffies;
	ha->mbox_status_count = outCount;
	for (i=0; i < MBOX_REG_COUNT; i++) {
		ha->mbox_status[i] = 0;
	}

	for (i=0; i<inCount; i++) {
		DEBUG5(printk("qisioctl%d: %s: Mailbox In[%d]  0x%08X\n",
				       ha->host_no, __func__, i, mbx_cmd[i]));
	}

	if (IS_QLA8022(ha)) {
		/* Load all mailbox registers, except mailbox 0. */
		for (i = 1; i < inCount; i++)
			writel(mbx_cmd[i], &ha->qla4_82xx_reg->mailbox_in[i]);
		for (i = inCount; i < MBOX_REG_COUNT; i++)
			writel(0, &ha->qla4_82xx_reg->mailbox_in[i]);

		writel(mbx_cmd[0], &ha->qla4_82xx_reg->mailbox_in[0]);
		readl(&ha->qla4_82xx_reg->mailbox_in[0]);
		writel(HINT_MBX_INT_PENDING, &ha->qla4_82xx_reg->hint);
	} else {
		/* Load all mailbox registers, except mailbox 0.*/
		for (i = 1; i < inCount; i++) {
			WRT_REG_DWORD(&ha->reg->mailbox[i], mbx_cmd[i]);
		}
		for (i = inCount; i < MBOX_REG_COUNT; i++) {
			WRT_REG_DWORD(&ha->reg->mailbox[i], 0);
		}

		/* Write Mailbox 0 to alert the firmware that the mailbox registers
		 * contain a command to be processed.  NOTE: We could be interrupted
		 * here if system interrupts are enabled */
		WRT_REG_DWORD(&ha->reg->mailbox[0], mbx_cmd[0]);
		PCI_POSTING(&ha->reg->mailbox[0]);
		WRT_REG_DWORD(&ha->reg->ctrl_status, SET_RMASK(CSR_INTR_RISC));
		PCI_POSTING(&ha->reg->ctrl_status);
	}
	spin_unlock_irqrestore(&ha->hardware_lock, flags);

	add_wait_queue(&ha->mailbox_wait_queue,&wait);

	/*
	 * If we don't want status, don't wait for the mailbox command to
	 * complete.  For example, MBOX_CMD_RESET_FW doesn't return status,
	 * you must poll the inbound Interrupt Mask for completion.
	 */
	if (outCount == 0) {
		status = QLA_SUCCESS;
		remove_wait_queue(&ha->mailbox_wait_queue,&wait);
		ha->f_end = jiffies;
		goto mbox_exit;
	}

	/* Always poll for mailbox completion when calling the Disable
	   Interrupts mailbox command, as the interrupts will be disabled
	   upon completion */
	if (IS_QLA8022(ha) &&
	    (mbx_cmd[0] == MBOX_CMD_ENABLE_INTRS) &&
	    (mbx_cmd[1] == INTR_DISABLE))
		force_polling = 1;

	/*
	 * Wait for command to complete
	 * -----------------------------
	 */
	if (test_bit(AF_IRQ_ATTACHED, &ha->flags) &&
	    test_bit(AF_INTERRUPTS_ON, &ha->flags) &&
	    test_bit(AF_ONLINE, &ha->flags) &&
	    !test_bit(AF_HA_REMOVAL, &ha->flags) &&
	    !force_polling) {
                /* Do not poll for completion.  Use completion queue */
                set_bit(AF_MBOX_COMMAND_NOPOLL, &ha->flags);
                wait_for_completion_timeout(&ha->mbx_intr_comp, (MBOX_TOV * HZ));
                clear_bit(AF_MBOX_COMMAND_NOPOLL, &ha->flags);
        } else {
		/* Poll for command to complete */
		set_current_state(TASK_UNINTERRUPTIBLE);
		wait_count = jiffies + (MBOX_TOV * HZ);
		while (test_bit(AF_MBOX_COMMAND_DONE, &ha->flags) == 0) {
			if (time_after_eq(jiffies, wait_count)) {
				break;
			}
			/*
			 * Service the interrupt.
			 * The ISR will save the mailbox status registers
			 * to a temporary storage location in the adapter
			 * structure.
			 */
			spin_lock_irqsave(&ha->hardware_lock, flags);
			if (IS_QLA8022(ha)) {
				intr_status = readl(&ha->qla4_82xx_reg->host_int);
				if (intr_status & ISRX_QLA82XX_RISC_INT) {
					ha->mbox_status_count = outCount;
					intr_status = readl(&ha->qla4_82xx_reg->host_status);
					ql4im_qla4_82xx_interrupt_service_routine(ha, intr_status);
				}
			} else {
				intr_status = RD_REG_DWORD(&ha->reg->ctrl_status);
				DEBUG5(printk("qisioctl%d: %s: INTR_STATUS = 0x%X\n",
					       ha->host_no, __func__, intr_status));
				if (intr_status & INTR_PENDING) {
					ha->mbox_status_count = outCount;
					ql4im_interrupt_service_routine(ha, intr_status);
				}
			}
			spin_unlock_irqrestore(&ha->hardware_lock, flags);

			/*
			 * Delay for 10 microseconds
			 * NOTE: Interrupt_handler may be called here,
			 *       if interrupts are enabled
			 */
			msleep(10);
		} /* wait loop */
	}

	remove_wait_queue(&ha->mailbox_wait_queue,&wait);

	/*
	 * Check for mailbox timeout
	 */
	if (!test_bit(AF_MBOX_COMMAND_DONE, &ha->flags)) {
		if (IS_QLA8022(ha) &&
			test_bit(AF_FW_RECOVERY, &ha->flags)) {
			DEBUG2(printk("qisioctl%d: %s: prematurely completing mbx cmd as "
				"firmware hang detected\n",
				ha->host_no, __func__));
			goto mbox_exit;
		}

		if (!test_bit(DPC_RESET_HA, &ha->dpc_flags)) {
			DEBUG2(printk("qisioctl%d: Mailbox Cmd 0x%08X timed out ...,"
					" Scheduling Adapter Reset\n",
					ha->host_no, mbx_cmd[0]));

			ha->mailbox_timeout_count++;
			mbx_sts[0] = (-1);

			if (IS_QLA8022(ha)) {
				set_bit(AF_FW_RECOVERY, &ha->flags);
				DEBUG2(printk("qisioctl%d: %s: set FW RECOVERY!\n",
					ha->host_no, __func__));
				if (test_bit(AF_MBOX_COMMAND, &ha->flags)) {
					if (test_bit(AF_MBOX_COMMAND_NOPOLL, &ha->flags)) {
						printk("qisioctl%d: %s: Due to fw "
						"recovery, doing premature completion of "
						"mbx cmd\n", ha->host_no, __func__);
						complete(&ha->mbx_intr_comp);
					} else {
						printk("qisioctl%d: %s: Due to fw "
						"recovery, doing premature completion of "
						"polling mbx cmd\n", ha->host_no, __func__);
						set_bit(AF_MBOX_COMMAND_DONE, &ha->flags);
					}
				}
			}
			set_bit(DPC_RESET_HA, &ha->dpc_flags);
		}
		goto mbox_exit;
	}

	DEBUG5(printk("qisioctl%d: %s: mailbox cmd done!\n",
			ha->host_no, __func__));

	/*
	 * Copy the mailbox out registers to the caller's mailbox in/out
	 * structure.
	 */
	spin_lock_irqsave(&ha->hardware_lock, flags);
	for (i=0; i < outCount; i++) {
		mbx_sts[i] = ha->mbox_status[i];
		DEBUG5(printk("qisioctl%d: %s: Mailbox Status[%d]  0x%08X\n",
				ha->host_no, __func__, i, mbx_sts[i]));
	}

	/*
	 * Set return status and error flags (if applicable)
	 */
	switch (ha->mbox_status[0]) {
	
	case MBOX_STS_COMMAND_COMPLETE:
		status = QLA_SUCCESS;
		break;

	case MBOX_STS_INTERMEDIATE_COMPLETION:
		status = QLA_SUCCESS;
		DEBUG5(printk("qisioctl%d: %s: Cmd = %08X, Intermediate completion\n",
				ha->host_no, __func__, mbx_cmd[0]));
		break;

	case MBOX_STS_BUSY:
		DEBUG2(printk("qisioctl%d: %s: Cmd = %08X, ISP BUSY\n",
				      ha->host_no, __func__, mbx_cmd[0]));

		ha->mailbox_timeout_count++;
		break;

	case MBOX_STS_COMMAND_PARAMETER_ERROR:
		DEBUG2(printk("qisioctl%d: %s: Cmd = %08X, Command Parameter Error\n",
					ha->host_no, __func__, mbx_cmd[0]));
		break;

	case MBOX_STS_INVALID_COMMAND:
	case MBOX_STS_HOST_INTERFACE_ERROR:
	case MBOX_STS_TEST_FAILED:
	case MBOX_STS_COMMAND_ERROR:
	default:
		DEBUG2(printk("qisioctl%d: %s: **** FAILED, cmd = %08X, "
				"sts = %08X ****\n",
				ha->host_no, __func__, mbx_cmd[0], mbx_sts[0]));


		for (i = 0; i < MBOX_REG_COUNT; i++) {
			DEBUG2(printk("mbox_cmd[%d] = 0x%08X    ", i, mbx_cmd[i]));
			DEBUG2(printk("mbox_sts[%d] = 0x%08X\n", i, ha->mbox_status[i]));
		}
		break;
	} /* switch mbox status */
	spin_unlock_irqrestore(&ha->hardware_lock, flags);

	mbox_exit:
	clear_bit(AF_MBOX_COMMAND, &ha->flags);
	clear_bit(AF_MBOX_COMMAND_DONE, &ha->flags);
	up(&ha->mbox_sem);

	return(status);
}

uint8_t
ql4im_mbx_clear_database_entry(scsi_qla_host_t *ha, uint16_t fw_ddb_index)
{
	uint32_t    mbox_cmd[MBOX_REG_COUNT];
	uint32_t    mbox_sts[MBOX_REG_COUNT];

	memset(&mbox_cmd, 0, sizeof(mbox_cmd));
	memset(&mbox_sts, 0, sizeof(mbox_sts));
	mbox_cmd[0] = MBOX_CMD_CLEAR_DATABASE_ENTRY;
	mbox_cmd[1] = fw_ddb_index;

	if (ha->ql4_mbx(ha, 2, 5, &mbox_cmd[0], &mbox_sts[0])
	    != QLA_SUCCESS) {
		DEBUG2(printk("qisioctl%d: %s: MBOX_CMD_CLEAR_DATABASE_ENTRY "
		    "failed sts %04X index [%d], state %04x\n",
		    ha->host_no, __func__, mbox_sts[0], fw_ddb_index,
		    mbox_sts[4]));
		return(QLA_ERROR);
	}

	return(QLA_SUCCESS);
}

uint8_t
ql4im_mbx_conn_close_sess_logout(scsi_qla_host_t *ha, uint16_t fw_ddb_index,
			       uint16_t connection_id, uint16_t option)
{
	uint32_t    mbox_cmd[MBOX_REG_COUNT];
	uint32_t    mbox_sts[MBOX_REG_COUNT];

	memset(&mbox_cmd, 0, sizeof(mbox_cmd));
	memset(&mbox_sts, 0, sizeof(mbox_sts));
	mbox_cmd[0] = MBOX_CMD_CONN_CLOSE_SESS_LOGOUT;
	mbox_cmd[1] = fw_ddb_index;
	mbox_cmd[2] = connection_id;
	mbox_cmd[3] = option;

	if (ha->ql4_mbx(ha, 4, 2, &mbox_cmd[0], &mbox_sts[0])
	    != QLA_SUCCESS) {
		DEBUG2(printk("qisioctl%d: %s: MBOX_CMD_CONN_CLOSE_SESS_LOGOUT "
				"option %04x failed sts %04X %04X",
				ha->host_no, __func__, option,
				mbox_sts[0], mbox_sts[1]));

		if (mbox_sts[0] == MBOX_STS_COMMAND_ERROR) {
			DEBUG2(printk(", reason %04X\n", mbox_sts[1]));
		}
		else {
			DEBUG2(printk("\n"));
		}
	}

	return(QLA_SUCCESS);
}


/**************************************************************************
 * ql4im_mbx_get_fwddb_entry
 *	This routine retrieves the firmware's device database entry.
 *
 * Input:
 * 	ha - Pointer to host adapter structure.
 *      fw_ddb_index - Firmware's device database index
 *      fw_ddb_entry - Pointer to firmware's device database entry structure
 *      num_valid_ddb_entries - Pointer to number of valid ddb entries
 *      next_ddb_index - Pointer to next valid device database index
 *      fw_ddb_device_state - Pointer to device state
 *
 * Output:
 *      fw_ddb_entry - Fills in structure if pointer is supplied
 *      num_valid_ddb_entries - Fills in if pointer is supplied
 *      next_ddb_index - Fills in if pointer is supplied
 *      fw_ddb_device_state - Fills in if pointer is supplied
 *
 * Returns:
 *	QLA_SUCCESS - Successfully retrieved ddb info from firmware
 *	QLA_ERROR   - Failed to retrieve ddb info from firmware
 *
 * Context:
 *	Kernel context.
 **************************************************************************/
uint8_t
ql4im_mbx_get_fwddb_entry(scsi_qla_host_t *ha,
		      uint16_t        fw_ddb_index,
		      DEV_DB_ENTRY    *fw_ddb_entry,
		      dma_addr_t      fw_ddb_entry_dma,
		      uint32_t        *num_valid_ddb_entries,
		      uint32_t        *next_ddb_index,
		      uint32_t        *fw_ddb_device_state,
		      uint32_t        *conn_err_detail,
		      uint16_t        *tcp_source_port_num,
		      uint16_t        *connection_id)
{
	uint8_t         status = QLA_ERROR;
	uint32_t	mbox_cmd[MBOX_REG_COUNT];
	uint32_t	mbox_sts[MBOX_REG_COUNT];

	/* Make sure the device index is valid */
	if (fw_ddb_index >= MAX_DDB_ENTRIES) {
		DEBUG2( printk("qisioctl%d: %s: index [%d] out of range.\n",
				ha->host_no, __func__, fw_ddb_index));
		goto exit_get_fwddb;
	}

	memset(&mbox_cmd, 0, sizeof(mbox_cmd));
	memset(&mbox_sts, 0, sizeof(mbox_sts));
	mbox_cmd[0] = MBOX_CMD_GET_DATABASE_ENTRY;
	mbox_cmd[1] = (uint32_t) fw_ddb_index;
	mbox_cmd[2] = LSDW(fw_ddb_entry_dma);
	mbox_cmd[3] = MSDW(fw_ddb_entry_dma);

	if (ha->ql4_mbx(ha, 4, 7, &mbox_cmd[0], &mbox_sts[0])
	    == QLA_ERROR) {
		DEBUG2(printk("qisioctl%d: %s: MBOX_CMD_GET_DATABASE_ENTRY failed "
		    "with status 0x%04X\n",
		    ha->host_no, __func__, mbox_sts[0]));
		goto exit_get_fwddb;
	}

	if (fw_ddb_index != mbox_sts[1]) {
		DEBUG2(printk("qisioctl%d: %s: index mismatch [%d] != [%d].\n",
		    ha->host_no, __func__, fw_ddb_index,
		    mbox_sts[1]));
		goto exit_get_fwddb;
	}

	if (fw_ddb_entry) {
		DEBUG4(printk("qisioctl%d: %s: "
		    "DDB[%d] MB0 %04x Tot %d Next %d "
		    "State %04x ConnErr %08x %d.%d.%d.%d:%04d \"%s\"\n",
		    ha->host_no, __func__, fw_ddb_index,
		    mbox_sts[0], mbox_sts[2], mbox_sts[3], mbox_sts[4], mbox_sts[5],
		    fw_ddb_entry->RemoteIPAddr[0],
		    fw_ddb_entry->RemoteIPAddr[1],
		    fw_ddb_entry->RemoteIPAddr[2],
		    fw_ddb_entry->RemoteIPAddr[3],
		    le16_to_cpu(fw_ddb_entry->RemoteTCPPortNumber),
		    fw_ddb_entry->iscsiName));
	}

	if (num_valid_ddb_entries)
		*num_valid_ddb_entries = mbox_sts[2];

	if (next_ddb_index)
		*next_ddb_index = mbox_sts[3];

	if (fw_ddb_device_state)
		*fw_ddb_device_state = mbox_sts[4];
	/* RA: This mailbox has been changed to pass connection error and details.
	 * Its true for ISP4010 as per Version E - Not sure when it was changed.
	 * Get the time2wait from the fw_dd_entry field : default_time2wait which
 	 * we call it as minTime2Wait DEV_DB_ENTRY struct.
	 */
	if (conn_err_detail)
		*conn_err_detail = mbox_sts[5];

	if (tcp_source_port_num)
		*tcp_source_port_num = (uint16_t) mbox_sts[6] >> 16;

	if (connection_id)
		*connection_id = (uint16_t) mbox_sts[6] & 0x00FF;

	status = QLA_SUCCESS;

exit_get_fwddb:
	return(status);
}


/**************************************************************************
 * ql4im_mbx_set_fwddb_entry
 *	This routine initializes or updates the adapter's device database
 *	entry for the specified device.
 *
 * Input:
 * 	ha - Pointer to host adapter structure.
 *      fw_ddb_index - Firmware's device database index
 *      fw_ddb_entry - Pointer to firmware's device database entry
 *		       structure, or NULL.
 *
 * Output:
 *	None
 *
 * Remarks:
 *	This routine also triggers a login for the specified device.
 *	Therefore, it may also be used as a secondary login routine when
 *	a NULL pointer is specified for the fw_ddb_entry.
 *
 * Returns:
 *	QLA_SUCCESS - Successfully set ddb_entry in firmware
 *	QLA_ERROR   - Failed to set ddb_entry in firmware
 *
 * Context:
 *	Kernel context.
 **************************************************************************/
uint8_t
ql4im_mbx_set_ddb_entry(scsi_qla_host_t *ha,
		      uint16_t        fw_ddb_index,
		      DEV_DB_ENTRY    *fw_ddb_entry,
		      dma_addr_t      fw_ddb_entry_dma)
{
	uint8_t         status = QLA_ERROR;
	uint32_t mbox_cmd[MBOX_REG_COUNT];
	uint32_t mbox_sts[MBOX_REG_COUNT];

	DEBUG4(printk("qisioctl%d: %s: index [%d]\n",
			      ha->host_no, __func__, fw_ddb_index));

	/* Do not wait for completion. The firmware will send us an
	 * ASTS_DATABASE_CHANGED (0x8014) to notify us of the login status.
	 */
	memset(&mbox_cmd, 0, sizeof(mbox_cmd));
	memset(&mbox_sts, 0, sizeof(mbox_sts));
	mbox_cmd[0] = MBOX_CMD_SET_DATABASE_ENTRY;
	mbox_cmd[1] = (uint32_t) fw_ddb_index;
	mbox_cmd[2] = LSDW(fw_ddb_entry_dma);
	mbox_cmd[3] = MSDW(fw_ddb_entry_dma);

	status = ha->ql4_mbx(ha, 4, 1, &mbox_cmd[0], &mbox_sts[0]);
	if (status == QLA_SUCCESS) {
		DEBUG4(printk("qisioctl%d: %s: mbx[0] = 0x%04x\n",
				      ha->host_no, __func__, mbox_sts[0]));
	}

	return(status);
}

#if ISNS_SUPPORTED
uint8_t
ql4im_mbx_isns_enable(scsi_qla_host_t *ha,
		    uint32_t isns_ip_addr,
		    uint16_t isns_server_port_num)
{
	uint32_t mbox_cmd[MBOX_REG_COUNT];
	uint32_t mbox_sts[MBOX_REG_COUNT];

	DEBUG4(printk("qisioctl%d: %s: isns_ip_addr %08x\n",
			       ha->host_no, __func__, isns_ip_addr));

	ql4im_isns_build_entity_id(ha);

	memset(&mbox_cmd, 0, sizeof(mbox_cmd));
	memset(&mbox_sts, 0, sizeof(mbox_sts));
	mbox_cmd[0] = MBOX_CMD_SET_ISNS_SERVICE;
	mbox_cmd[1] = ISNSv4_ENABLE;
	mbox_cmd[2] = isns_ip_addr;
	mbox_cmd[3] = isns_server_port_num;

	if (ha->ql4_mbx(ha, 4, 6, &mbox_cmd[0], &mbox_sts[0])
	    != QLA_SUCCESS) {
		DEBUG2(printk("qisioctl%d: %s: MBOX_CMD_SET_ISNS_SERVICE failed "
				"w/ status %04X %04X\n",
				ha->host_no, __func__, mbox_sts[0], mbox_sts[1]));
		return(QLA_ERROR);
	}

	DEBUG4(printk("qisioctl%d: Start iSNS Service "
				    "%d.%d.%d.%d Port %04d . . .\n", ha->host_no,
				    (isns_ip_addr & 0x000000FF),
				    (isns_ip_addr & 0x0000FF00) >> 8,
				    (isns_ip_addr & 0x00FF0000) >> 16,
				    (isns_ip_addr & 0xFF000000) >> 24,
				    isns_server_port_num));

	return(QLA_SUCCESS);
}

uint8_t
ql4im_mbx_isns_disable(scsi_qla_host_t *ha)
{
	uint32_t mbox_cmd[MBOX_REG_COUNT];
	uint32_t mbox_sts[MBOX_REG_COUNT];

	if (atomic_read(&ha->isns.state) == ISNS_STATE_TCP_CONNECTED) {
		memset(&mbox_cmd, 0, sizeof(mbox_cmd));
		memset(&mbox_sts, 0, sizeof(mbox_sts));
		mbox_cmd[0] = MBOX_CMD_SET_ISNS_SERVICE;
		mbox_cmd[1] = ISNS_DISABLE;

		if (ha->ql4_mbx(ha, 2, 2, &mbox_cmd[0], &mbox_sts[0])
		    != QLA_SUCCESS) {
			DEBUG2(printk("qisioctl%d: %s: MBOX_CMD_SET_ISNS_SERVICE failed "
					"w/ status %04X %04X\n",
					ha->host_no, __func__, mbox_sts[0], mbox_sts[1]));
			return(QLA_ERROR);
		}
	}

	/* Trigger DPC to poll for iSNS connection */
	atomic_set(&ha->isns.state, ISNS_STATE_TCP_DISCONNECTED);
	memset(ha->isns.entity_id, 0, sizeof(ha->isns.entity_id));
	return(QLA_SUCCESS);
}
#endif

uint8_t
ql4im_mbx_get_flash(scsi_qla_host_t *ha, dma_addr_t dma_addr, uint32_t offset, uint32_t len)
{
	uint32_t mbox_cmd[MBOX_REG_COUNT];
	uint32_t mbox_sts[MBOX_REG_COUNT];

	memset(&mbox_cmd, 0, sizeof(mbox_cmd));
	memset(&mbox_sts, 0, sizeof(mbox_sts));

	mbox_cmd[0] = MBOX_CMD_READ_FLASH;
	mbox_cmd[1] = LSDW(dma_addr);
	mbox_cmd[2] = MSDW(dma_addr);
	mbox_cmd[3] = offset;
	mbox_cmd[4] = len;

	if (ha->ql4_mbx(ha, 5, 2, &mbox_cmd[0], &mbox_sts[0])
	    != QLA_SUCCESS) {
		DEBUG2(printk("qisioctl%d: %s: MBOX_CMD_READ_FLASH, failed w/ "
				      "status %04X %04X, offset %08x, len %08x\n",
				      ha->host_no, __func__, mbox_sts[0], mbox_sts[1],
				      offset, len));
		return(QLA_ERROR);
	}
	return(QLA_SUCCESS);
}

/**************************************************************************
 * ql4im_mbx_restore_factory_defaults
 *	This routine restores the factory defaults for the specified adapter.
 *
 * Input:
 * 	ha - Pointer to host adapter structure.
 *
 * Output:
 *	None
 *
 * Returns:
 *	QLA_SUCCESS - Successfully restored defaults
 *	QLA_ERROR   - Failed to restore defaults
 *
 * Context:
 *	Kernel context.
 **************************************************************************/
uint8_t
ql4im_mbx_restore_factory_defaults(scsi_qla_host_t *ha,
				   uint32_t block_mask,
				   uint32_t ifcb_mask1,
				   uint32_t ifcb_mask2)
{
	uint32_t mbox_cmd[MBOX_REG_COUNT];
	uint32_t mbox_sts[MBOX_REG_COUNT];

	memset(&mbox_cmd, 0, sizeof(mbox_cmd));
	memset(&mbox_sts, 0, sizeof(mbox_sts));
	mbox_cmd[0] = MBOX_CMD_RESTORE_FACTORY_DEFAULTS;
	mbox_cmd[3] = block_mask;
        mbox_cmd[4] = ifcb_mask1;
        mbox_cmd[5] = ifcb_mask2;

	if (ha->ql4_mbx(ha, 6, 1, &mbox_cmd[0], &mbox_sts[0])
	    != QLA_SUCCESS) {
		DEBUG2(printk("qisioctl%d: %s: MBOX_CMD_RESTORE_FACTORY_DEFAULTS"
				      " failed w/ status %04X\n",
				      ha->host_no, __func__, mbox_sts[0]));
		return(QLA_ERROR);
	}

	return(QLA_SUCCESS);
}

/*
 * Overrides for Emacs so that we almost follow Linus's tabbing style.
 * Emacs will notice this stuff at the end of the file and automatically
 * adjust the settings for this buffer only.  This must remain at the end
 * of the file.
 * ---------------------------------------------------------------------------
 * Local variables:
 * c-indent-level: 2
 * c-brace-imaginary-offset: 0
 * c-brace-offset: -2
 * c-argdecl-indent: 2
 * c-label-offset: -2
 * c-continued-statement-offset: 2
 * c-continued-brace-offset: 0
 * indent-tabs-mode: nil
 * tab-width: 8
 * End:
 */

