/*
 * QLogic iSCSI IOCTL Module
 * Copyright (c)  2003-2012 QLogic Corporation
 *
 * See LICENSE.qla4xxx for copyright and licensing details.
 */

/******************************************************************************
 *             Please see release.txt for revision history.                   *
 *                                                                            *
 ******************************************************************************
 * Function Table of Contents:
 ****************************************************************************/

#include "ql4im_def.h"
#include <linux/version.h>
#include <linux/vmalloc.h>

/* Restrict compilation to 2.6.0 or greater */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
#error "This module does not support kernel versions earlier than 2.6.0"
#endif

extern struct list_head *qla4xxx_hostlist_ptr;
extern rwlock_t *qla4xxx_hostlist_lock_ptr;
extern rwlock_t qla4xxx_hostlist_lock;
extern void qla4xxx_hostlist;

LIST_HEAD(ql4im_ha_list);
rwlock_t ql4im_ha_list_lock = RW_LOCK_UNLOCKED;

struct list_head **drvr_ha_list_ptr = NULL;
rwlock_t **drvr_ha_list_lock_ptr = NULL;

extern char *qla4xxx_version_str;
uint8_t drvr_major = 5;
uint8_t drvr_minor = 0;
uint8_t drvr_patch = 5;
uint8_t drvr_beta = 9;
char drvr_ver[40];
#define DEFAULT_VER     "5.00.05b9-k"

/* extern functions */
extern int qla4xxx_send_ioctl(struct scsi_device *, int, void __user *);
extern int ql4im_alloc_ioctl_mem(struct ql4im_scsi_qla_host *);
extern void ql4im_free_ioctl_mem(struct ql4im_scsi_qla_host *);
extern ql4im_scsi_qla_host_t *ql4im_get_adapter_handle(uint16_t instance);


/**************************************************************************
 * ql4im_get_hba_count
 *      This routine returns the number of host adapters present.
 *
 * Input:
 *      None
 *
 * Returns:
 *    ql4im_hba_count - Number of host adapters present.
 *
 * Context:
 *      Kernel context.
 **************************************************************************/
uint16_t
ql4im_get_hba_count(void)
{
	struct list_head        *ioctll;
	uint8_t			hba_count = 0;

	read_lock(&ql4im_ha_list_lock);
	list_for_each(ioctll, &ql4im_ha_list) {
		hba_count++;
	}
	read_unlock(&ql4im_ha_list_lock);

	DEBUG4(printk("qisioctl: %s: hba_count %d\n",
		      __func__, hba_count));
	return(hba_count);
}

/**************************************************************************
 * ql4im_get_adapter_handle
 *      This routine returns the adapter handle that corresponds to the
 *      specified instance number.
 *
 * Input:
 *      instance - Instance number of the desired host adapter.
 *
 * Returns:
 *      Pointer to host adapter structure
 *
 * Context:
 *      Interrupt context.
 **************************************************************************/
ql4im_scsi_qla_host_t *ql4im_get_adapter_handle(uint16_t instance)
{
	ql4im_scsi_qla_host_t 	*ql4im_ha;
	struct list_head        *ioctll;

	//FIXME: Need a quicker lookup method.  This routine gets called often.

	read_lock(&ql4im_ha_list_lock);
	list_for_each(ioctll, &ql4im_ha_list) {
		ql4im_ha = list_entry(ioctll, struct ql4im_scsi_qla_host, list);
		if (ql4im_ha->instance != instance) {
			continue;
		}

		DEBUG4(printk("qisioctl%d: %s: handle (%p) for instance %d\n",
			      ql4im_ha->host_no, __func__, ql4im_ha, instance));

		read_unlock(&ql4im_ha_list_lock);
		return(ql4im_ha);
	}

	read_unlock(&ql4im_ha_list_lock);
	DEBUG2(printk("qisioctl: : %s: instance %d not found\n",
		      __func__, instance));

	return(NULL);
}

int
ql4im_down_timeout(struct semaphore *sema, unsigned long timeout)
{
	const unsigned int step = HZ/10;

	do {
		if (!down_trylock(sema))
			return 0;
		
		DEBUG2(printk("%s(): down_trylock failed\n", __func__);)
		set_current_state(TASK_INTERRUPTIBLE);
		if (schedule_timeout(step))
			break;
	} while ((timeout -= step) > 0);

	return -ETIMEDOUT;
}


static void
ql4im_copy_drvr_data(struct scsi_qla_host *drvr_ha,
                     struct ql4im_scsi_qla_host *ql4im_ha)
{
	ql4im_ha->host_no = drvr_ha->host_no;
	ql4im_ha->instance = drvr_ha->instance;
	ql4im_ha->drvr_ha = drvr_ha;

	strcpy(ql4im_ha->drv_verstr, drvr_ha->driver_verstr);
	ql4im_ha->drv_major = drvr_ha->driver_version[0];
	ql4im_ha->drv_minor = drvr_ha->driver_version[1];
	ql4im_ha->drv_patch = drvr_ha->driver_version[2];
	ql4im_ha->drv_beta  = drvr_ha->driver_version[3];

	ql4im_ha->firmware_version[0] = drvr_ha->firmware_version[0];
	ql4im_ha->firmware_version[1] = drvr_ha->firmware_version[1];
	ql4im_ha->patch_number = drvr_ha->patch_number;
	ql4im_ha->build_number = drvr_ha->build_number;
	ql4im_ha->acb_version  = drvr_ha->acb_version;
	ql4im_ha->ifcb_size    = drvr_ha->ifcb_size;

#if ISNS_SUPPORTED
	ql4im_ha->tcp_options = drvr_ha->tcp_options;
	ql4im_ha->isns_server_port_number = drvr_ha->isns.server_port;

	memcpy(ql4im_ha->isns_server_ip_addr, drvr_ha->isns.server_ip,
	       MIN(sizeof(ql4im_ha->isns_server_ip_addr),
		   sizeof(drvr_ha->isns.server_ip)));
#endif
	memcpy(ql4im_ha->ip_address, drvr_ha->ip_address,
	       MIN(sizeof(ql4im_ha->ip_address),
		   sizeof(drvr_ha->ip_address)));
}

static void
ql4im_ioctl_exit(void)
{
	struct list_head	*hal;
	struct ql4im_scsi_qla_host	*tmp_ha_ioctl;

	DEBUG12(printk("qisioctl: %s: entered.\n", __func__));
	write_lock(&ql4im_ha_list_lock);
	list_for_each(hal, &ql4im_ha_list) {
		tmp_ha_ioctl = list_entry(hal, struct ql4im_scsi_qla_host, list);
		ql4im_free_ioctl_mem(tmp_ha_ioctl);
	}
	write_unlock(&ql4im_ha_list_lock);

	qla4xxx_ioctl_exit();
	DEBUG12(printk("qisioctl: %s: exiting.\n", __func__));

}

int
ql4im_init(void)
{
	struct list_head		*hal = NULL;
	struct ql4im_scsi_qla_host	*ql4im_ha = NULL;
	struct scsi_qla_host		*drvr_ha;
	int retval = -ENODEV;

	DEBUG12(printk("qisioctl: %s: entered.\n", __func__));

	read_lock(&qla4xxx_hostlist_lock);
	DEBUG7(printk("qisioctl: %s: got hostlist lock.\n", __func__));

	/* Allocate our ql4im_ha list */
	write_lock(&ql4im_ha_list_lock);
	list_for_each(hal, (struct list_head *)&qla4xxx_hostlist) {
		drvr_ha = list_entry(hal, struct scsi_qla_host, list);
		if (drvr_ha) {
			ql4im_ha = vmalloc(sizeof(struct ql4im_scsi_qla_host));
			if (ql4im_ha) {
				DEBUG7(printk("qisioctl: %s: got ql4im_ha=%p.\n",
					__func__, ql4im_ha));
				memset(ql4im_ha, 0, sizeof(struct ql4im_scsi_qla_host));
				list_add_tail(&ql4im_ha->list, &ql4im_ha_list);
				ql4im_copy_drvr_data(drvr_ha, ql4im_ha);
			}
		}
	}
	write_unlock(&ql4im_ha_list_lock);
	read_unlock(&qla4xxx_hostlist_lock);

	qla4xxx_ioctl_init();

	list_for_each(hal, &ql4im_ha_list) {
		ql4im_ha = list_entry(hal, struct ql4im_scsi_qla_host, list);
		ql4im_alloc_ioctl_mem(ql4im_ha);
	}

	retval = 0;

	return retval;
}

void
ql4im_exit(void)
{
	struct list_head	*hal = NULL;
	struct ql4im_scsi_qla_host	*ql4im_ha;

	DEBUG12(printk("qisioctl: %s: entered.\n", __func__));

	ql4im_ioctl_exit();

	DEBUG7(printk("qisioctl: %s: done ioctl_exit.\n", __func__));

	/* remove/free the list */
	write_lock(&ql4im_ha_list_lock);
	/* original list_for_each code:
	for (pos = (head)->next; prefetch(pos->next), pos != (head);
	pos = pos->next)
	*/
	for (hal = (&ql4im_ha_list)->next; hal !=
	    (&ql4im_ha_list); ) {
		DEBUG7(printk("qisioctl: %s: going to get next ql4im_ha.\n", __func__));
		ql4im_ha = list_entry(hal, struct ql4im_scsi_qla_host, list);
		
		DEBUG7(printk("qisioctl: %s: going to del &ql4im_ha->list=%p.\n",
		    __func__, &ql4im_ha->list));
		hal = hal->next;
		list_del(&ql4im_ha->list);
		DEBUG7(printk("qisioctl: %s: going to free ql4im_ha=%p.\n",
		    __func__, ql4im_ha));
		vfree(ql4im_ha);
		DEBUG7(printk("qisioctl: %s: freed ql4im_ha.\n", __func__));
	}
	write_unlock(&ql4im_ha_list_lock);

	DEBUG12(printk("qisioctl: %s: exiting.\n", __func__));
}
