/*
 * QLogic iSCSI IOCTL Module
 * Copyright (c)  2003-2012 QLogic Corporation
 *
 * See LICENSE.qla4xxx for copyright and licensing details.
 */

/******************************************************************************
 *             Please see release.txt for revision history.                   *
 *                                                                            *
 ******************************************************************************
 * Function Table of Contents:
 ****************************************************************************/

#include "ql4im_def.h"
#include <linux/delay.h>



/**************************************************************************
 * ql4im_lookup_ddb_by_fw_index
 *	This routine locates a device handle given the firmware device
 *	database index.  If device doesn't exist, returns NULL.
 *
 * Input:
 * 	ha - Pointer to host adapter structure.
 *      fw_ddb_index - Firmware's device database index
 *
 * Returns:
 *	Pointer to the corresponding internal device database structure
 *
 * Context:
 *	Kernel context.
 **************************************************************************/
ddb_entry_t *
ql4im_lookup_ddb_by_fw_index(scsi_qla_host_t *ha, uint32_t fw_ddb_index)
{
	ddb_entry_t *ddb_entry = NULL;

	if ((fw_ddb_index < MAX_DDB_ENTRIES) &&
	    (ha->fw_ddb_index_map[fw_ddb_index] !=
		(ddb_entry_t *) INVALID_ENTRY)) {
		ddb_entry = ha->fw_ddb_index_map[fw_ddb_index];
	}

	DEBUG3(printk("qisioctl%d: %s: index [%d], ddb_entry = %p\n",
	    ha->host_no, __func__, fw_ddb_index, ddb_entry));

	return ddb_entry;
}

/**************************************************************************
 * qla4xxx_lookup_target_by_fcport
 *	This routine locates a target handle given the fcport
 *
 * Input:
 * 	ha - Pointer to host adapter structure.
 *	fcport - port handle
 *
 * Returns:
 *
 * Context:
 *	Kernel context.
 **************************************************************************/
os_tgt_t *
ql4im_lookup_target_by_fcport(scsi_qla_host_t *ha, fc_port_t  *fcport)
{
	int t;
	os_tgt_t *tq = NULL;

	for (t = 0; t < MAX_TARGETS; t++) {
		if ((tq = TGT_Q(ha, t)) == NULL)
			continue;

		if (fcport == tq->fcport)
			break;
	}

	return tq;
}

/**************************************************************************
 * ql4im_lookup_lun_handle
 *	This routine locates a lun handle given the device handle and lun
 *	number.
 *
 * Input:
 * 	ha - Pointer to host adapter structure.
 *	ddb_entry - Pointer to device database entry
 *	lun - SCSI LUN
 *
 * Returns:
 *	Pointer to corresponding lun_entry structure
 *
 * Context:
 *	Kernel context.
 **************************************************************************/
os_lun_t *
ql4im_lookup_lun_handle(scsi_qla_host_t *ha, os_tgt_t *tq, uint16_t lun)
{
	os_lun_t *lq = NULL;

	if (tq && lun < MAX_LUNS)
		lq = tq->olun[lun];
	return lq;
}

/**************************************************************************
 * ql4im_lookup_target_by_SCSIID
 *	This routine locates a target handle given the SCSI bus and
 *	target IDs.  If device doesn't exist, returns NULL.
 *
 * Input:
 * 	ha - Pointer to host adapter structure.
 *	bus - SCSI bus number
 *	target - SCSI target ID.
 *
 * Returns:
 *	Pointer to the corresponding internal device database structure
 *
 * Context:
 *	Kernel context.
 **************************************************************************/
os_tgt_t *
ql4im_lookup_target_by_SCSIID(scsi_qla_host_t *ha, uint32_t bus,
    uint32_t target)
{
	os_tgt_t *tq = NULL;

	if (target < MAX_TARGETS)
		tq = TGT_Q(ha, target);

	DEBUG3(printk("qisioctl%d: %s: b%d:t%d, tgt = %p\n",
	    ha->host_no, __func__, bus, target, tq));

	return tq;
}

srb_t *ql4im_del_from_free_srb_q_head(scsi_qla_host_t *ha)
{
	unsigned long flags;
	struct list_head *ptr;
	srb_t *srb = NULL;

	spin_lock_irqsave(&ha->list_lock, flags);
	if (!list_empty(&ha->free_srb_q)) {
		/* Remove list entry from head of queue */
		ptr = ha->free_srb_q.next;
		list_del_init(ptr);

		/* Return pointer to srb structure */
		srb = list_entry(ptr, srb_t, list_entry);
		atomic_set(&srb->ref_count, 1);
		srb->state = SRB_NO_QUEUE_STATE;
		ha->free_srb_q_count--;
	}
	spin_unlock_irqrestore(&ha->list_lock, flags);

	DEBUG5(printk("qisioctl%d: %s: instance %d, srb = %p\n",
			      ha->host_no, __func__, ha->instance,
			      srb ));

	return(srb);
}

void
ql4im_add_to_free_srb_q(scsi_qla_host_t *ha, srb_t *srb)
{
	unsigned long flags;

	DEBUG5(printk("qisioctl%d: %s: instance %d, srb = %p\n",
			      ha->host_no, __func__, ha->instance,
			      srb ));

	spin_lock_irqsave(&ha->list_lock, flags);
	atomic_set(&srb->ref_count, 0);
	memset(srb, 0, sizeof(srb_t));
	list_add_tail(&srb->list_entry, &ha->free_srb_q);
	spin_unlock_irqrestore(&ha->list_lock, flags);

	ha->free_srb_q_count++;
	srb->state = SRB_FREE_STATE;
}

static void
ql4im_add_to_done_srb_q(scsi_qla_host_t *ha, srb_t *srb)
{
	unsigned long flags;

	spin_lock_irqsave(&ha->list_lock, flags);
	DEBUG5(printk("qisioctl%d: %s: ha %d, srb = %p\n",
			      ha->host_no, __func__, ha->instance, srb));
	list_add_tail(&srb->list_entry, &ha->done_srb_q);
	srb->state = SRB_DONE_STATE;
	ha->done_srb_q_count++;
	srb->ha = ha;
	spin_unlock_irqrestore(&ha->list_lock, flags);
}

/**************************************************************************
 * del_from_active_array
 *      This routine removes and returns the srb at the specified index
 *
 * Input:
 *      ha - Pointer to host adapter structure.
 *      index - index into to the active_array
 *
 * Returns:
 *      Pointer to corresponding SCSI Request Block
 *
 * Context:
 *      Kernel/Interrupt context.
 **************************************************************************/
static srb_t *
ql4im_del_from_active_array(scsi_qla_host_t *ha, uint32_t index)
{
	srb_t *srb = NULL;

	/* validate handle and remove from active array */
	if (index < MAX_SRBS) {
		srb = ha->active_srb_array[index];
		ha->active_srb_array[index] = NULL;

		if (srb) {
			os_lun_t *lun_entry = srb->lun_queue;

			/* update counters */
			if ((srb->flags & SRB_DMA_VALID) != 0) {
				ha->req_q_count += srb->entry_count;
				ha->iocb_cnt -= srb->entry_count;
				if (ha->active_srb_count)
					ha->active_srb_count--;
				// if (ddb_entry) ddb_entry->out_count--;
				if (lun_entry)
					lun_entry->out_count--;
				if (srb->cmd)
					srb->cmd->host_scribble = NULL;
			}
		}
		else
			DEBUG2(printk("qisioctl%d: %s: array_index=%d "
					"already completed.\n",
					ha->host_no, __func__, index));
	}
	else
		DEBUG2(printk("qisioctl%d: %s: array_index=%d "
				      "exceeded max index of %d\n",
				      ha->host_no, __func__, index, MAX_SRBS));

	return(srb);
}

/**************************************************************************
 * ql4im_os_cmd_timeout
 *
 * Description:
 *       Handles the command if it times out in any state.
 *
 * Input:
 *     sp - pointer to validate
 *
 * Returns:
 * None.
 **************************************************************************/
static void
ql4im_os_cmd_timeout(srb_t *sp)
{
	int t, l;
	int processed;
	scsi_qla_host_t *dest_ha;
	struct scsi_cmnd *cmd;
	ulong      flags;
	ulong      cpu_flags;
	fc_port_t       *fcport;

	cmd = sp->cmd;

	t = cmd->device->id;
	l = cmd->device->lun;
	fcport = sp->fclun->fcport;
	dest_ha = sp->ha;

	DEBUG2(printk("qisioctl%d: %s: sp->state = %x\n",
		      dest_ha->host_no, __func__, sp->state);)


	/*
	 * If IO is found either in retry Queue
	 *    OR in Lun Queue
	 * Return this IO back to host
	 */
	processed = 0;

	spin_lock_irqsave(&dest_ha->list_lock, cpu_flags);
	if (sp->state == SRB_SUSPENDED_STATE) {
		DEBUG2(printk("qisioctl%d: %s: Found SP %p in suspended state  "
			      "- pid %ld:\n",
			      dest_ha->host_no,__func__, sp,
			      sp->cmd->serial_number);)
	}
	else if (sp->state == SRB_ACTIVE_STATE) {
		/*
		 * IO is with ISP find the command in our active list.
		 */
		spin_unlock_irqrestore(&dest_ha->list_lock, cpu_flags);	/* 01/03 */
		spin_lock_irqsave(&dest_ha->hardware_lock, flags);
		if (sp == dest_ha->active_srb_array
		    [(unsigned long)sp->cmd->host_scribble]) {
			sp->state = SRB_ACTIVE_TIMEOUT_STATE;
			spin_unlock_irqrestore(&dest_ha->hardware_lock, flags);
		}
		else {
			spin_unlock_irqrestore(&dest_ha->hardware_lock, flags);
			DEBUG2(printk(
			       "qisioctl%d: %s: State indicates it is with "
			       "ISP, But not in active array\n",
			       dest_ha->host_no, __func__));
		}
		spin_lock_irqsave(&dest_ha->list_lock, cpu_flags);
	}
	else if (sp->state == SRB_ACTIVE_TIMEOUT_STATE) {
		/* double timeout */
	}
	else {
		/* EMPTY */
		DEBUG3(printk("qisioctl%d: %s: LOST command state = "
			      "0x%x, sp=%p\n",
			      dest_ha->host_no, __func__, sp->state, sp);)

		printk(KERN_INFO "qisioctl%d: %s: LOST command state = 0x%x\n",
			dest_ha->host_no, __func__, sp->state);
	}
	spin_unlock_irqrestore(&dest_ha->list_lock, cpu_flags);
}

/**************************************************************************
 * ql4im_add_timer_to_cmd
 *      This routine creates a timer for the specified command. The timeout
 *      is usually the command time from kernel minus 2 secs.
 *
 * Input:
 *      srb - Pointer to SCSI Request Block
 *      timeout - Number of seconds to extend command timeout.
 *
 * Returns:
 *      None
 *
 * Context:
 *      Kernel context.
 **************************************************************************/
void
ql4im_add_timer_to_cmd(srb_t *srb, int timeout)
{
	init_timer(&srb->timer);
	srb->timer.expires = jiffies + timeout * HZ;
	srb->timer.data = (unsigned long) srb;
	srb->timer.function = (void (*) (unsigned long))ql4im_os_cmd_timeout;
	add_timer(&srb->timer);
	atomic_inc(&srb->ref_count);
	DEBUG3(printk("%s: srb %p, timeout %d\n",
			      __func__, srb, timeout));
}

/**************************************************************************
 * qla4xxx_delete_timer_from_cmd
 *      This routine deletes the timer for the specified command.
 *
 * Input:
 *      srb - Pointer to SCSI Request Block
 *
 * Returns:
 *      None
 *
 * Context:
 *      Kernel/Interrupt context.
 **************************************************************************/
void
ql4im_delete_timer_from_cmd(srb_t *srb )
{
	if (srb->flags & SRB_NO_TIMER)
		return;

	if (del_timer(&srb->timer)) {
		atomic_dec(&srb->ref_count);
		srb->timer.function =  NULL;
		srb->timer.data = (unsigned long) NULL;
	}
}

uint8_t
ql4im_free_srb_dma(scsi_qla_host_t *ha, srb_t *srb)
{
        struct scsi_cmnd *cmd;

        cmd = srb->cmd;
        if (cmd == NULL) {
                DEBUG2(printk("scsi%d: %s: ERROR: NULL cmd pointer in "
                                "srb=%p\n", ha->host_no, __func__, srb));
                return QLA_ERROR;
        }
        /* Release memory used for this I/O */
        if ( (srb->flags & SRB_DMA_VALID) ) {
                srb->flags &= ~SRB_DMA_VALID;

                /* Release memory used for this I/O */
                if (cmd->use_sg) {
                        pci_unmap_sg(ha->pdev,
                                     cmd->request_buffer,
                                     cmd->use_sg,
                                     cmd->sc_data_direction);
                }
                else if (cmd->request_bufflen) {
                        pci_unmap_page(ha->pdev,
                                       srb->saved_dma_handle,
                                       cmd->request_bufflen,
                                       srb->cmd->sc_data_direction);
                }
                ha->total_mbytes_xferred += cmd->request_bufflen / 1024;
        }
        return QLA_SUCCESS;
}


/**************************************************************************
 * ql4im_request_cleanup
 *      This routine frees resources for a command that
 *      didn't get completed.
 *
 * Input:
 *      ha - Pointer to host adapter structure.
 *      srb - Pointer to SCSI Request Block
 *
 * Remarks:
 *    The srb pointer should be guaranteed to be nonzero before calling
 *    this function.  The caller should also ensure that the list_lock is
 *    released before calling this function.
 *
 * Returns:
 *
 * Context:
 *      Kernel/Interrupt context.
 **************************************************************************/
void
ql4im_request_cleanup(scsi_qla_host_t *ha, srb_t *srb)
{
        ql4im_delete_timer_from_cmd(srb);
        ql4im_free_srb_dma(ha, srb);

        /*  Let abort handler know we are completing the command */
        CMD_SP(srb->cmd) = NULL;

        atomic_dec(&srb->ref_count);
        srb->cmd = NULL;
        ql4im_add_to_free_srb_q(ha, srb);
}

static void
qla4xxx_strtolower(uint8_t *str)
{
	uint8_t *tmp;
	for (tmp = str; *tmp != '\0'; tmp++) {
		if (*tmp >= 'A' && *tmp <= 'Z')
			*tmp += 'a' - 'A';
	}
}

void
ql4im_isns_build_entity_id(scsi_qla_host_t *ha)
{
	sprintf(ha->isns.entity_id, "%s.%d", ha->serial_number, ha->func_num);
	qla4xxx_strtolower(ha->isns.entity_id);
}


void
ql4im_flush_all_srbs(scsi_qla_host_t *ha, ddb_entry_t *ddb_entry,
		       os_lun_t *lun_entry)
{
	int         i;
	unsigned long flags;
	srb_t       *srb;

	if (lun_entry == NULL || ddb_entry == NULL)
		return;

	/* free active commands */
	spin_lock_irqsave(&ha->hardware_lock, flags);
	if (lun_entry->out_count != 0) {
		for (i = 1; i < MAX_SRBS; i++) {
			srb = ha->active_srb_array[i];
			if (!srb)
				continue;

			DEBUG4(printk("qisioctl%d:%d:%d:%d: %s: found srb "
                                      "%p in active_q\n",
				      ha->host_no, ddb_entry->bus,
				      ddb_entry->target, lun_entry->lun,
				      __func__, srb));

			if (srb->lun_queue != lun_entry)
				continue;

			ql4im_del_from_active_array(ha, i);
			srb->cmd->result = DID_NO_CONNECT << 16;
			ql4im_add_to_done_srb_q(ha,srb);
		}
	}
	spin_unlock_irqrestore(&ha->hardware_lock, flags);

}

/**************************************************************************
 * ql4im_login_device
 *	This routine is called by the login IOCTL to log in the specified
 *	device.
 *
 * Input:
 * 	ha - Pointer to host adapter structure.
 * 	fw_ddb_index - Index of the device to login
 * 	connection_id - Connection ID of the device to login
 *
 * Returns:
 *	QLA_SUCCESS - Successfully logged in device
 *	QLA_ERROR   - Failed to login device
 *
 * Context:
 *	Kernel context.
 **************************************************************************/
uint8_t
ql4im_login_device(scsi_qla_host_t *ha, uint16_t fw_ddb_index,
		     uint16_t connection_id)
{
	ddb_entry_t *ddb_entry;
	uint8_t status = QLA_ERROR;

	DEBUG3(printk("qisioctl%d: %s: Login index [%d]\n", ha->host_no,
			      __func__, fw_ddb_index));

	ddb_entry = ql4im_lookup_ddb_by_fw_index(ha, fw_ddb_index);
	if (ddb_entry == NULL) {
		DEBUG2(printk("qisioctl%d: %s: Invalid index [%d]\n",
                        ha->host_no, __func__, fw_ddb_index));
		goto exit_login_device;
	}

	if (ql4im_mbx_get_fwddb_entry(ha, fw_ddb_index, NULL, 0, NULL, NULL,
				      &ddb_entry->fw_ddb_device_state,
				      NULL, NULL, NULL) == QLA_ERROR) {

		DEBUG2(printk("qisioctl%d: %s: 1st get ddb entry failed\n",
			ha->host_no, __func__));
		goto exit_login_device;
	}

	if (ddb_entry->fw_ddb_device_state == DDB_DS_SESSION_ACTIVE) {
		DEBUG3(printk("qisioctl%d: %s: login successful for index [%d]\n",
			ha->host_no, __func__, ddb_entry->fw_ddb_index));

		status = QLA_SUCCESS;

		goto exit_login_device;
	}

	if (ql4im_mbx_conn_close_sess_logout(ha, fw_ddb_index, connection_id,
					   LOGOUT_OPTION_RELOGIN) != QLA_SUCCESS) {
		goto exit_login_device;
	}

	status = QLA_SUCCESS;

exit_login_device:
	return(status);
}

/**************************************************************************
 * qla4xxx_delete_device
 *	This routine is called by the logout IOCTL to delete the specified
 *      device.	 Send the LOGOUT and DELETE_DDB commands for the specified
 *      target, even if it's not in our internal database.
 *
 * Input:
 * 	ha - Pointer to host adapter structure.
 * 	fw_ddb_index - Index of the device to delete
 * 	connection_id - Connection ID of the device to delete
 *
 * Returns:
 *	QLA_SUCCESS - Successfully deleted device
 *	QLA_ERROR   - Failed to delete device
 *
 * Context:
 *	Kernel context.
 **************************************************************************/
uint8_t
ql4im_delete_device(scsi_qla_host_t *ha, uint16_t fw_ddb_index,
		      uint16_t connection_id)
{
	uint8_t     status = QLA_ERROR;
	uint32_t    fw_ddb_device_state = 0xFFFF;
	u_long      wait_count;
	ddb_entry_t *ddb_entry;

	/* In order to preserve SCSI target ID assignments to DDB indexes,
	 * we do not delete the actual ddb_entry when this call is made.
	 * Instead, we mark it DEAD.  A ddb index may be reused for a newly
	 * added target, so precautions should be taken.
	 */

	/* If the device is in our internal tables, set the NO_RELOGIN bit. */
	ddb_entry = ql4im_lookup_ddb_by_fw_index(ha, fw_ddb_index);
	if (ddb_entry != NULL) {
		/* Do not attempt relogins until after a reset */
		DEBUG4(printk("qisioctl%d:%d:%d: %s:  setting NO_RELOGIN flag\n",
                        ha->host_no, ddb_entry->bus, ddb_entry->target, __func__));

		set_bit(DF_NO_RELOGIN, &ddb_entry->flags);
	}

	/*
	 * If the device state is already one that we can delete, bypass the
	 * logout command.
	 */
	ql4im_mbx_get_fwddb_entry(ha, fw_ddb_index, NULL, 0, NULL, NULL,
				&fw_ddb_device_state, NULL, NULL, NULL);

	DEBUG4(printk("scsi%d: %s: index [%d], state=%d \n",
			      ha->host_no, __func__,
                              fw_ddb_index, fw_ddb_device_state));

	if (fw_ddb_device_state == DDB_DS_UNASSIGNED ||
	    fw_ddb_device_state == DDB_DS_NO_CONNECTION_ACTIVE ||
	    fw_ddb_device_state == DDB_DS_SESSION_FAILED)
		goto delete_ddb;

	/* First logout index */
	if (ql4im_mbx_conn_close_sess_logout(ha, fw_ddb_index, connection_id,
					   LOGOUT_OPTION_CLOSE_SESSION)
	    != QLA_SUCCESS) {
		DEBUG2(printk("qisioctl%d: %s: LOGOUT_OPTION_CLOSE_SESSION "
				"failed index [%d]\n", ha->host_no, __func__,
				fw_ddb_index));
		goto exit_delete_ddb;
	}

	/* Wait enough time to complete logout */
	wait_count = jiffies + LOGOUT_TOV * HZ;
	while (ql4im_mbx_get_fwddb_entry(ha, fw_ddb_index, NULL, 0, NULL, NULL,
					 &fw_ddb_device_state,
					 NULL, NULL, NULL) == QLA_SUCCESS) {

		if (time_after_eq(jiffies, wait_count))
			goto exit_delete_ddb;

		if (fw_ddb_device_state == DDB_DS_UNASSIGNED ||
		    fw_ddb_device_state == DDB_DS_NO_CONNECTION_ACTIVE ||
		    fw_ddb_device_state == DDB_DS_SESSION_FAILED)
			break;

		udelay(50);
	}

delete_ddb:
	if (fw_ddb_device_state == DDB_DS_UNASSIGNED) {
		/* After a ddb has been deleted via SS, a ddb_ds can go from
		 * NO_CONNECTION_ACTIVE to UNASSIGNED without the driver
		 * receiving an AEN, so update it here as we key off of the
		 * UNASSIGNED old ddb_ds to determine whether a DDB index is
		 * being reused.
		 */
		if (ddb_entry != NULL)
			ddb_entry->fw_ddb_device_state = fw_ddb_device_state;
		goto exit_delete_ddb;
	}

	if (ql4im_mbx_clear_database_entry(ha, fw_ddb_index) == QLA_SUCCESS) {
		uint16_t lun;
		os_lun_t *lun_entry;
		os_tgt_t *tgt_entry;

		status = QLA_SUCCESS;
		if (!ddb_entry)
			goto exit_delete_ddb;

		DEBUG4(printk("scsi%d: %s: mark index %d DELETED.\n",
				ha->host_no, __func__, fw_ddb_index));

		atomic_set(&ddb_entry->state, DEV_STATE_DELETED);
		atomic_set(&ddb_entry->fcport->state, FCS_DEVICE_DELETED);

		tgt_entry = ql4im_lookup_target_by_fcport(ha,
							    ddb_entry->fcport);
		if (tgt_entry) {
			for (lun = 0; lun < MAX_LUNS; lun++) {
				lun_entry = tgt_entry->olun[lun];
				if (lun_entry != NULL) {
					unsigned long cpu_flags;

					DEBUG4(printk(
					    "qisioctl%d:%d:%d:%d: %s: flushing "
					    "srbs, pendq_cnt=%d, retryq_cnt="
					    "%d, activeq_cnt=%d\n", ha->host_no,
					    ddb_entry->bus, tgt_entry->id, lun,
					    __func__, 0 ,
					    0,
					    ha->active_srb_count));

					ql4im_flush_all_srbs(ha, ddb_entry,
							       lun_entry);

					spin_lock_irqsave(&lun_entry->lun_lock,
							  cpu_flags);

					if (lun_entry->lun_state ==
					    LS_LUN_SUSPENDED) {
						lun_entry->lun_state =
						LS_LUN_READY;
					}

					spin_unlock_irqrestore(&lun_entry->lun_lock,
							       cpu_flags);
				}
			}
		}
	}

exit_delete_ddb:
	return(status);
}

/**************************************************************************
 * ql4im_logout_device
 *	This support routine is called by the logout IOCTL to log out
 *	the specified device.
 *
 * Input:
 * 	ha - Pointer to host adapter structure.
 * 	fw_ddb_index - Index of the device to logout
 * 	connection_id - Connection ID of the device to logout
 *
 * Returns:
 *	QLA_SUCCESS - Successfully logged out device
 *	QLA_ERROR   - Failed to logout device
 *
 * Context:
 *	Kernel context.
 **************************************************************************/
uint8_t
ql4im_logout_device(scsi_qla_host_t *ha, uint16_t fw_ddb_index,
		      uint16_t connection_id)
{
	uint8_t     status = QLA_ERROR;
	ddb_entry_t *ddb_entry;
	uint32_t    old_fw_ddb_device_state;

	DEBUG4(printk("qisioctl%d: %s: inst %d entered. index=%d.\n",
		ha->host_no, __func__, ha->instance, fw_ddb_index));

	ddb_entry = ql4im_lookup_ddb_by_fw_index(ha, fw_ddb_index);
	if (ddb_entry == NULL) {
		DEBUG4(printk("qisioctl%d: %s: Invalid index [%d]\n",
			ha->host_no, __func__, fw_ddb_index));
		goto exit_logout_device;
	}

	if (ql4im_mbx_get_fwddb_entry(ha, fw_ddb_index, NULL, 0, NULL, NULL,
                                      &old_fw_ddb_device_state,
				      NULL, NULL, NULL) != QLA_SUCCESS) {
		DEBUG4(printk("qisioctl%d: %s: get_ddb_entry failed\n",
			ha->host_no, __func__));
		goto exit_logout_device;
	}

	set_bit(DF_NO_RELOGIN, &ddb_entry->flags);

	if (ql4im_mbx_conn_close_sess_logout(ha, fw_ddb_index, connection_id,
                                             LOGOUT_OPTION_CLOSE_SESSION)
	    != QLA_SUCCESS) {
		goto exit_logout_device;
	}

	status = QLA_SUCCESS;

exit_logout_device:
	DEBUG4(printk("qisioctl%d: %s: inst %d exiting.\n",
			ha->host_no, __func__, ha->instance));
	return(status);
}

uint8_t ql4im_reindex_ddb(scsi_qla_host_t *ha, uint32_t old_fw_ddb_index, uint32_t new_fw_ddb_index)
{
	ddb_entry_t *ddb_entry = NULL;
	unsigned long flags = 0;

	spin_lock_irqsave(ha->host->host_lock, flags);
	if (ha->fw_ddb_index_map[new_fw_ddb_index] !=
	    (ddb_entry_t *) INVALID_ENTRY) {
		if ((atomic_read(&ha->fw_ddb_index_map[new_fw_ddb_index]->state)
		    != DEV_STATE_DEAD) &&
		    (atomic_read(&ha->fw_ddb_index_map[new_fw_ddb_index]->state)
		    != DEV_STATE_DELETED)) {
			DEBUG2(printk("scsi%d: %s: ERROR: new DDB index [%d] "
					      "already present.\n", ha->host_no, __func__,
					      new_fw_ddb_index));
			spin_unlock_irqrestore(ha->host->host_lock, flags);
			return (QLA_ERROR);
		}
	}

	if (ha->fw_ddb_index_map[old_fw_ddb_index] ==
	    (ddb_entry_t *) INVALID_ENTRY) {
		DEBUG2(printk("scsi%d: %s: ERROR: old DDB index [%d] "
				      "not present\n", ha->host_no, __func__,
				      old_fw_ddb_index));
		spin_unlock_irqrestore(ha->host->host_lock, flags);
		return (QLA_ERROR);
	}

	ddb_entry = ha->fw_ddb_index_map[old_fw_ddb_index];
	ha->fw_ddb_index_map[old_fw_ddb_index] = (ddb_entry_t *) INVALID_ENTRY;
	ddb_entry->fw_ddb_index = new_fw_ddb_index;
	ha->fw_ddb_index_map[new_fw_ddb_index] = ddb_entry;

	DEBUG7(printk("scsi%d: %s: DDB index [%d] "
			      "remapped to DDB index [%d]\n",
			      ha->host_no, __func__,
			      old_fw_ddb_index, new_fw_ddb_index));
	spin_unlock_irqrestore(ha->host->host_lock, flags);
	return (QLA_SUCCESS);
}


/**************************************************************************
 * qla4xxx_wait_for_hba_online
 *      This routine
 *
 * Input:
 *      ha - Pointer to host adapter structure
 *
 * Remarks:
 *
 * Returns:
 *      SUCCESS - Adapter is ONLINE
 *      FAILED  - Adapter is DEAD
 *
 * Context:
 *      Kernel context.  Assume io_request_lock LOCKED upon entry
 **************************************************************************/
uint8_t
ql4im_wait_for_hba_online(scsi_qla_host_t *ha)
{
	unsigned long wait_online;
	#define ADAPTER_ONLINE_TOV 30

	wait_online = jiffies + (ADAPTER_ONLINE_TOV * HZ);
	while (time_before(jiffies, wait_online)) {
		if (test_bit(AF_ONLINE, &ha->flags))
			return QLA_SUCCESS;

		set_current_state(TASK_UNINTERRUPTIBLE);
		schedule_timeout(2 * HZ);
	}

	return QLA_ERROR;
}

void
ql4im_update_local_ifcb(scsi_qla_host_t *ha, ADDRESS_CTRL_BLK  *init_fw_cb)
{
	uint8_t ip_addr_str[IP_ADDR_STR_SIZE];

	/*
	 * Save some info in adapter structure
	 */
	ha->firmware_options = le16_to_cpu(init_fw_cb->FwOptions);
	ha->ip_options = le16_to_cpu(init_fw_cb->IPOptions);
	ha->tcp_options = le16_to_cpu(init_fw_cb->TCPOptions);
	ha->heartbeat_interval = init_fw_cb->HeartbeatInterval;
	ha->isns_server_port_number =
	    le16_to_cpu(init_fw_cb->isns_svr_port);
	ha->acb_version = init_fw_cb->ACBVersion;

	memcpy(ha->ip_address, init_fw_cb->IPAddr,
	    MIN(sizeof(ha->ip_address), sizeof(init_fw_cb->IPAddr)));
	memcpy(ha->subnet_mask, init_fw_cb->SubnetMask,
	    MIN(sizeof(ha->subnet_mask), sizeof(init_fw_cb->SubnetMask)));
	memcpy(ha->gateway, init_fw_cb->GatewayIPAddr,
	    MIN(sizeof(ha->gateway), sizeof(init_fw_cb->GatewayIPAddr)));
	memcpy(ha->isns_server_ip_addr, init_fw_cb->ipv4_isns_svr_ip,
	    MIN(sizeof(ha->isns_server_ip_addr), sizeof(init_fw_cb->ipv4_isns_svr_ip)));
	memcpy(ha->name_string, init_fw_cb->iSCSINameString,
	    MIN(sizeof(ha->name_string), sizeof(init_fw_cb->iSCSINameString)));
	memcpy(ha->alias, init_fw_cb->iSCSIAlias,
	    MIN(sizeof(ha->alias), sizeof(init_fw_cb->iSCSIAlias)));

	if (ha->acb_version == ACB_SUPPORTED) {
                ha->ipv6_options = init_fw_cb->IPv6Options;
	}

	if ((ha->acb_version == ACB_NOT_SUPPORTED) || is_ipv4_enabled(ha)) {
		/* --- IP v4 --- */
		IPv4Addr2Str(init_fw_cb->IPAddr, &ip_addr_str[0]);
		DEBUG7(printk("scsi%d: %s: "
		    "IP Address            %s\n",
		    ha->host_no, __func__, &ip_addr_str[0]));

		IPv4Addr2Str(init_fw_cb->SubnetMask, &ip_addr_str[0]);
		DEBUG7(printk("scsi%d: %s: "
		    "Subnet Mask           %s\n",
		    ha->host_no, __func__, &ip_addr_str[0]));

		IPv4Addr2Str(init_fw_cb->GatewayIPAddr, &ip_addr_str[0]);
		DEBUG7(printk("scsi%d: %s: "
		    "Default Gateway       %s\n",
		    ha->host_no, __func__, &ip_addr_str[0]));

		DEBUG7(printk("scsi%d: %s: "
		    "Auto-Negotiate        %s\n", ha->host_no, __func__,
		    ((le16_to_cpu(init_fw_cb->AddFwOptions) & 0x10) != 0) ?
		    "ON" : "OFF"));
		DEBUG7(printk("scsi%d: %s: "
		    "SLP Use DA Enable     %s\n", ha->host_no, __func__,
		    ((ha->tcp_options & TOPT_SLP_USE_DA_ENABLE) != 0) ?
		    "ON" : "OFF"));
		DEBUG7(printk("scsi%d: %s: "
		    "SLP UA Enable         %s\n", ha->host_no, __func__,
		    ((ha->tcp_options & TOPT_SLP_UA_ENABLE) != 0) ?
		    "ON" : "OFF"));
		DEBUG7(printk("scsi%d: %s: "
		    "DHCP Enable           %s\n", ha->host_no, __func__,
		    ((ha->tcp_options & TOPT_DHCP_ENABLE) != 0) ?
		    "ON" : "OFF"));
		DEBUG7(printk("scsi%d: %s: "
		    "DNS via DHCP Enable   %s\n", ha->host_no, __func__,
		    ((ha->tcp_options & TOPT_GET_DNS_VIA_DHCP_ENABLE) != 0) ?
		    "ON" : "OFF"));
		DEBUG7(printk("scsi%d: %s: "
		    "SLP via DHCP Enable   %s\n", ha->host_no, __func__,
		    ((ha->tcp_options & TOPT_GET_SLP_VIA_DHCP_ENABLE) != 0) ?
		    "ON" : "OFF"));
		DEBUG7(printk("scsi%d: %s: "
		    "Auto Discovery Enable %s\n", ha->host_no, __func__,
		    ((ha->tcp_options & TOPT_AUTO_DISCOVERY_ENABLE) != 0) ?
		    "ON" : "OFF"));
		DEBUG7(printk("scsi%d: %s: "
		    "iSNS Enable           %s\n", ha->host_no, __func__,
		    (is_isnsv4_enabled(ha) ? "ON" : "OFF")));
#if ISNS_SUPPORTED
		if (is_isnsv4_enabled(ha)) {
			set_bit(ISNS_FLAG_ISNS_ENABLED_IN_ISP, &ha->isns.flags);

			IPv4Addr2Str(ha->isns.server_ip, &ip_addr_str[0]);
			DEBUG7(printk("scsi%d: %s: "
			    "iSNSv4 IP Address     %s\n",
			    ha->host_no, __func__, &ip_addr_str[0]));
			DEBUG7(printk("scsi%d: %s: "
			    "iSNS Server Port Number %d\n", ha->host_no,
			    __func__, ha->isns.server_port));
		}
#endif
		DEBUG7(printk("scsi%d: %s: "
		    "Heartbeat Enable      %s\n", ha->host_no, __func__,
		    ((ha->firmware_options & FWOPT_HEARTBEAT_ENABLE) != 0) ?
		    "ON" : "OFF"));
		if (ha->firmware_options & FWOPT_HEARTBEAT_ENABLE)
			DEBUG7(printk("scsi%d: %s: "
			    "Heartbeat Interval    %d\n", ha->host_no, __func__,
			    ha->heartbeat_interval));

		DEBUG7(printk("scsi%d: %s: "
		    "Execution Throttle    %d\n", ha->host_no, __func__,
		    le16_to_cpu(init_fw_cb->ExecThrottle)));

		DEBUG7(printk("scsi%d: %s: "
		    "ACB Supported  (%d)    %s\n", ha->host_no, __func__,
		    ha->acb_version,
		    (ha->acb_version == ACB_SUPPORTED) ? "ON" : "OFF"));
	}

	if (is_ipv6_enabled(ha)){
		/* --- IP v6 --- */

		ha->ipv6_addl_options = init_fw_cb->IPv6AddOptions;
		ha->ipv6_tcp_options = init_fw_cb->IPv6TCPOptions;
		ha->ipv6_link_local_addr[0] = 0xFE;
		ha->ipv6_link_local_addr[1] = 0x80;
		memcpy(&ha->ipv6_link_local_addr[8], init_fw_cb->IPv6InterfaceID,
		    MIN(sizeof(ha->ipv6_link_local_addr)/2, sizeof(init_fw_cb->IPv6InterfaceID)));
		memcpy(ha->ipv6_addr0, init_fw_cb->IPv6Addr0,
		    MIN(sizeof(ha->ipv6_addr0), sizeof(init_fw_cb->IPv6Addr0)));
		memcpy(ha->ipv6_addr1, init_fw_cb->IPv6Addr1,
		    MIN(sizeof(ha->ipv6_addr1), sizeof(init_fw_cb->IPv6Addr1)));
		memcpy(ha->ipv6_default_router_addr, init_fw_cb->IPv6DefaultRouterAddr,
		    MIN(sizeof(ha->ipv6_default_router_addr), sizeof(init_fw_cb->IPv6DefaultRouterAddr)));

		IPv6Addr2Str(ha->ipv6_link_local_addr, &ip_addr_str[0]);
		DEBUG7(printk("scsi%d: %s: "
			"IPv6 Link Local       %s\n",
			ha->host_no, __func__, &ip_addr_str[0]));

		IPv6Addr2Str(ha->ipv6_addr0, &ip_addr_str[0]);
		DEBUG7(printk("scsi%d: %s: "
			"IPv6 IP Address0      %s\n",
			ha->host_no, __func__, &ip_addr_str[0]));

		IPv6Addr2Str(ha->ipv6_addr1, &ip_addr_str[0]);
		DEBUG7(printk("scsi%d: %s: "
			"IPv6 IP Address1      %s\n",
			ha->host_no, __func__, &ip_addr_str[0]));

		IPv6Addr2Str(ha->ipv6_default_router_addr, &ip_addr_str[0]);
		DEBUG7(printk("scsi%d: %s: "
			"IPv6 Default Router   %s\n",
			ha->host_no, __func__, &ip_addr_str[0]));

#if ISNS_SUPPORTED
		/* In IPv6 mode, iSNS may either be IPv4 or IPv6, not both */
		if (is_isnsv4_enabled(ha)) {
			set_bit(ISNS_FLAG_ISNS_ENABLED_IN_ISP, &ha->isns.flags);

			memcpy(ha->isns.server_ip, init_fw_cb->ipv4_isns_svr_ip,
			    MIN(sizeof(ha->isns.server_ip), sizeof(init_fw_cb->ipv4_isns_svr_ip)));

			IPv4Addr2Str(ha->isns.server_ip, &ip_addr_str[0]);
			DEBUG7(printk("scsi%d: %s: "
				"IPv4 iSNS IP Address      %s\n",
				ha->host_no, __func__, &ip_addr_str[0]));
			DEBUG7(printk("scsi%d: %s: "
				"iSNS Server Port Number   %d\n", ha->host_no,
				__func__, ha->isns.server_port));
		}
		else if (is_isnsv6_enabled(ha)) {
			set_bit(ISNS_FLAG_ISNS_ENABLED_IN_ISP, &ha->isns.flags);

			memcpy(ha->isns.server_ip, init_fw_cb->ipv6_isns_svr_ip,
			    MIN(sizeof(ha->isns.server_ip), sizeof(init_fw_cb->ipv6_isns_svr_ip)));

			IPv6Addr2Str(ha->isns.server_ip, &ip_addr_str[0]);
			DEBUG7(printk("scsi%d: %s: "
				"IPv6 iSNS IP Address  %s\n",
				ha->host_no, __func__, &ip_addr_str[0]));
			DEBUG7(printk("scsi%d: %s: "
			    "iSNS Server Port Number   %d\n", ha->host_no,
			    __func__, ha->isns.server_port));
		}
#endif
		DEBUG7(printk("scsi%d: %s: "
		    "IPv6 Protocol Enable      %s\n", ha->host_no, __func__,
		    (IS_IPv6_ENABLED(ha)) ? "ON" : "OFF"));

		DEBUG7(printk("scsi%d: %s: "
		    "IPv6 Autoconfig LinkLocal %s\n", ha->host_no, __func__,
		    ((ha->ipv6_addl_options & IPV6_ADDOPT_AUTOCONFIG_LINK_LOCAL_ADDR) != 0) ?
		    "ON" : "OFF"));

#if ISNS_SUPPORTED
		DEBUG7(printk("scsi%d: %s: "
		    "IPv6 iSNS Enabled         %s\n", ha->host_no, __func__,
		    ((ha->ipv6_tcp_options & IPV6_TCPOPT_ISNSv6_ENABLE) != 0) ?
		    "ON" : "OFF"));
#endif
		DEBUG7(printk("scsi%d: %s: "
		    "IPv6 Options              0x%04x\n", ha->host_no, __func__,
		    ha->ipv6_options));

		DEBUG7(printk("scsi%d: %s: "
		    "IPv6 Additional Options   0x%04x\n", ha->host_no, __func__,
		    ha->ipv6_addl_options));

		DEBUG7(printk("scsi%d: %s: "
		    "IPv6 TCP Options          0x%04x\n", ha->host_no, __func__,
		    ha->ipv6_tcp_options));

		DEBUG7(printk("scsi%d: %s: "
		    "ACB Supported             %s\n", ha->host_no, __func__,
		    (ha->acb_version == ACB_SUPPORTED) ?
		    "ON" : "OFF"));
	}
}

